Type: Package
Package: dvir
Title: Disaster Victim Identification
Version: 3.4.2
Authors@R: c(
    person("Magnus Dehli", "Vigeland", , "m.d.vigeland@medisin.uio.no", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-9134-4962")),
    person("Thore", "Egeland", , "thore.egeland@nmbu.no", role = "aut",
           comment = c(ORCID = "0000-0002-3465-8885"))
  )
Description: Joint DNA-based disaster victim identification (DVI), as
    described in Vigeland and Egeland (2021)
    <doi:10.21203/rs.3.rs-296414/v1>. Identification is performed by
    optimising the joint likelihood of all victim samples and reference
    individuals. Individual identification probabilities, conditional on
    all available information, are derived from the joint solution in the
    form of posterior pairing probabilities. 'dvir' is part of the
    'pedsuite' collection of packages for pedigree analysis.
License: GPL-3
URL: https://github.com/magnusdv/dvir
BugReports: https://github.com/magnusdv/dvir/issues
Depends: pedtools (>= 2.11.0), R (>= 4.2.0)
Imports: forrel (>= 1.9.0), mirai, pedFamilias, pedprobr (>= 1.1.0),
        ribd, verbalisr
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Encoding: UTF-8
Language: en-GB
LazyData: true
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-30 13:44:09 UTC; magnu
Author: Magnus Dehli Vigeland [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9134-4962>),
  Thore Egeland [aut] (ORCID: <https://orcid.org/0000-0002-3465-8885>)
Maintainer: Magnus Dehli Vigeland <m.d.vigeland@medisin.uio.no>
Repository: CRAN
Date/Publication: 2026-06-30 14:50:02 UTC
Built: R 4.6.0; ; 2026-06-30 17:43:51 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,dviData)
export(Bmarginal)
export(addVictims)
export(amDrivenDVI)
export(checkDVI)
export(compactJointRes)
export(directMatch)
export(dviCompare)
export(dviData)
export(dviGridSize)
export(dviJoint)
export(dviSim)
export(dviSolve)
export(excludePairing)
export(exclusionMatrix)
export(expand.grid.nodup)
export(familias2dvir)
export(findExcluded)
export(findNonidentifiable)
export(findUndisputed)
export(formatSummary)
export(generatePairings)
export(getDatabase)
export(getFamily)
export(getSimpleFams)
export(jointDVI)
export(mergePM)
export(ncomb)
export(pairwiseLR)
export(plotDVI)
export(plotSolution)
export(plotUndisputed)
export(relabelDVI)
export(sequentialDVI)
export(setPairing)
export(subsetDVI)
export(swapOrientation)
import(pedtools)
importFrom(forrel,findExclusions)
importFrom(forrel,profileSim)
importFrom(grDevices,dev.new)
importFrom(graphics,grconvertX)
importFrom(graphics,grconvertY)
importFrom(graphics,layout)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,rect)
importFrom(pedprobr,likelihood)
importFrom(stats,aggregate)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
importFrom(verbalisr,verbalise)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/dvir)](https://CRAN.R-project.org/package=dvir)
[![](https://cranlogs.r-pkg.org/badges/grand-total/dvir?color=yellow)](https://cran.r-project.org/package=dvir)
[![](https://cranlogs.r-pkg.org/badges/last-month/dvir?color=yellow)](https://cran.r-project.org/package=dvir)
<!-- badges: end -->

# Disaster Victim Identification in R

The purpose of **dvir** is to implement state-of-the-art algorithms for
DNA-based disaster victim identification (DVI). In particular, **dvir**
performs *joint* identification of multiple victims.

The methodology and algorithms of **dvir** are described in

- Egeland & Vigeland (2025): [Kinship cases with partially specified
  hypotheses](https://doi.org/10.1016/j.fsigen.2025.103270)
- Vigeland & Egeland (2021): [Joint DNA-based Disaster Victim
  Identification](https://www.nature.com/articles/s41598-021-93071-5)

The **dvir** package is part of the
[pedsuite](https://magnusdv.github.io/pedsuite/), a collection of R
packages for pedigree analysis. Much of the machinery behind **dvir** is
imported from other pedsuite packages, especially **pedtools** for
handling pedigree data, and **forrel** for the calculation of likelihood
ratios. A comprehensive presentation of these packages, and much more,
can be found in the book [Pedigree Analysis in
R](https://shop.elsevier.com/books/pedigree-analysis-in-r/vigeland/978-0-12-824430-2).

## Installation

To get the current official version of **dvir**, install from CRAN as
follows:

``` r
install.packages("dvir")
```

Alternatively, the latest development version may be obtained from
GitHub:

``` r
# install.packages("remotes")
remotes::install_github("magnusdv/dvir")
```

## Joint DVI analysis

``` r
library(dvir)
```

We illustrate joint DVI analysis using a toy example from Vigeland &
Egeland (2021), shown below. Three victim samples (V1, V2, V3) are to be
matched against three missing persons (M1, M2, M3) from two families.

<img src="man/figures/README-example2-plot-1.png" alt="" width="85%" style="display: block; margin: auto;" />

The hatched symbols indicate genotyped individuals. This simple example
uses a single marker with 10 equifrequent alleles denoted 1, 2, …, 10.
The observed genotypes are shown in the figure.

DNA profiles from victims are referred to as *post mortem* (PM) data,
while the *ante mortem* (AM) data contain profiles from the reference
individuals, R1 and R2.

The function `dviJoint()` performs a joint likelihood search over all
valid assignments.

``` r
joint = dviJoint(example2)
#> DVI dataset: 
#>  3 victims (2M/1F): V1, V2, V3
#>  3 missing (2M/1F): M1, M2, M3
#>  2 reference individuals: R1, R2
#>  2 pedigrees: F1, F2
#> Markers: 1 
#> Number of assignments: 14 (exact)
```

The output contains the log-likelihood of each of the 14 assignments,
and likelihood ratios (LRs) comparing the most likely solution with each
of the others. Thus the first row has `LR = 1`, while larger values
indicate less likely assignments:

|     | V1  | V2  | V3  |  loglik |  LR | M1  | M2  | M3  |
|:----|:----|:----|:----|--------:|----:|:----|:----|:----|
| 1   | M1  | M2  | M3  | -16.118 |   1 | V1  | V2  | V3  |
| 2   | M1  | M2  | \*  | -17.728 |   5 | V1  | V2  | \*  |
| 3   | \*  | M2  | M3  | -18.421 |  10 | \*  | V2  | V3  |
| 4   | M1  | \*  | M3  | -20.030 |  50 | V1  | \*  | V3  |
| 5   | \*  | M1  | M3  | -20.030 |  50 | V2  | \*  | V3  |
| 6   | \*  | M2  | \*  | -20.030 |  50 | \*  | V2  | \*  |
| 7   | \*  | \*  | M3  | -20.030 |  50 | \*  | \*  | V3  |
| 8   | \*  | M1  | \*  | -21.640 | 250 | V2  | \*  | \*  |
| 9   | M1  | \*  | \*  | -21.640 | 250 | V1  | \*  | \*  |
| 10  | \*  | \*  | \*  | -21.640 | 250 | \*  | \*  | \*  |
| 11  | M2  | M1  | M3  |    -Inf | Inf | V2  | V1  | V3  |
| 12  | M2  | M1  | \*  |    -Inf | Inf | V2  | V1  | \*  |
| 13  | M2  | \*  | M3  |    -Inf | Inf | \*  | V1  | V3  |
| 14  | M2  | \*  | \*  |    -Inf | Inf | \*  | V1  | \*  |

The columns V1, V2 and V3 give the PM-oriented assignment: each victim
is either paired with a missing person or remains unassigned (`"*"`).
The columns M1, M2 and M3 give the AM-oriented view of the same
assignments.

We see that the most likely joint solution is V1 = M1, V2 = M2 and V3 =
M3.

## Automated DVI workflow

For complete DVI analysis, the main function is `dviSolve()`. It
combines several steps, including exclusions, undisputed
identifications, pairwise LR calculations, joint analysis of unresolved
families, and conclusions based on generalised likelihood ratios (GLRs).

We illustrate this with the built-in dataset `fire`, which includes
three victim samples and three missing persons from the same family. The
function `plotDVI()` gives an overview:

``` r
plotDVI(fire, style = 2)
```

<img src="man/figures/README-fire-plot-1.png" alt="" width="85%" style="display: block; margin: auto;" />

Here is the complete workflow using LR threshold 10 000:

``` r
res = dviSolve(fire, detailedOutput = TRUE, threshold = 1e4, verbose = TRUE)
#> 
#> ------ Checking dataset  [+0.0s | 0.0s] ----------
#> 
#> Ok
#> 
#> ------ Nonidentifiable MPs  [+0.0s | 0.0s] -------
#> 
#> None 
#> 
#> ------ Exclusions, iteration 1  [+0.0s | 0.0s] ---
#> 
#> No exclusions
#> 
#> ------ Undisputed, iteration 1  [+0.0s | 0.0s] ---
#> 
#> No change; breaking loop
#> 
#> ------ AM-driven: Simple families  [+0.1s | 0.1s] 
#> 
#> 0 simple families
#> 
#> ------ AM-driven: Complex families  [+0.0s | 0.1s] 
#> 
#> 1 complex family: F1
#> 
#> Number of assignments: 34 (exact)
#> 
#>   Family Missing Sample       LR        GLR      Conclusion                            Comment
#> 1     F1      M1     V1 121.3476   70582.39     Match (GLR)                  Joint: {M1,M2,M3}
#> 2     F1      M2  V2/V3       NA 2947149.10 Symmetric match Full siblings: {M2, M3} = {V2, V3}
#> 3     F1      M3  V2/V3       NA 2947149.10 Symmetric match Full siblings: {M2, M3} = {V2, V3}
#> 
#> ------ PM-driven: Remaining victims  [+0.2s | 0.4s] 
#> 
#> No remaining victims
#> 
#> ------ Analysis complete  [+0.0s | 0.4s] ---------
```

The output contains AM-oriented and PM-oriented summary tables,
including conclusions and comments. With `detailedOutput = TRUE`, the
result also contains the pairwise LR matrix, the exclusion matrix, and
the joint tables used in the analysis:

#### `res$AM`

| Family | Missing | Sample | LR | GLR | Conclusion | Comment |
|:---|:---|:---|---:|---:|:---|:---|
| F1 | M1 | V1 | 121.35 | 70582.39 | Match (GLR) | Joint: {M1,M2,M3} |
| F1 | M2 | V2/V3 | NA | 2947149.10 | Symmetric match | Full siblings: {M2, M3} = {V2, V3} |
| F1 | M3 | V2/V3 | NA | 2947149.10 | Symmetric match | Full siblings: {M2, M3} = {V2, V3} |

#### `res$PM`

| Sample | Missing | Family | LR | GLR | Conclusion | Comment |
|:---|:---|:---|---:|---:|:---|:---|
| V1 | M1 | F1 | 121.35 | 70582.39 | Match (GLR) | Joint: {M1,M2,M3} |
| V2 | M2/M3 | F1 | 36.56 | 2947149.10 | Symmetric match | Full siblings: {V2, V3} = {M2, M3} |
| V3 | M2/M3 | F1 | 1.31 | 2947149.10 | Symmetric match | Full siblings: {V2, V3} = {M2, M3} |

#### `res$LRmatrix`

|     |      M1 |      M2 |      M3 |
|:----|--------:|--------:|--------:|
| V1  | 121.348 | 398.582 | 398.582 |
| V2  |   0.583 |  36.564 |  36.564 |
| V3  |   0.002 |   1.312 |   1.312 |

#### `res$exclusionMatrix`

|     |  M1 |  M2 |  M3 |
|:----|----:|----:|----:|
| V1  |   0 |   0 |   0 |
| V2  |   0 |   0 |   0 |
| V3  |   0 |   0 |   0 |

#### `res$jointTables` (truncated)

| V1  | V2  | V3  |  loglik |         LR | M1  | M2  | M3  |
|:----|:----|:----|--------:|-----------:|:----|:----|:----|
| M1  | M2  | M3  | -257.73 |       1.00 | V1  | V2  | V3  |
| M1  | M3  | M2  | -257.73 |       1.00 | V1  | V3  | V2  |
| \*  | M2  | M3  | -268.90 |   70582.39 | \*  | V2  | V3  |
| \*  | M3  | M2  | -268.90 |   70582.39 | \*  | V3  | V2  |
| M2  | M1  | M3  | -272.63 | 2947149.10 | V2  | V1  | V3  |
| M3  | M1  | M2  | -272.63 | 2947149.10 | V2  | V3  | V1  |

### Conclusion

In the `fire` example, V1 is identified as M1 by joint GLR analysis
(Egeland & Vigeland, 2025). The two remaining victims are matched to the
two missing full siblings M2 and M3, but the data do not distinguish
between the two sibling assignments. This is reported as a symmetric
match.

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/dvir)](https://CRAN.R-project.org/package=dvir)
[![](https://cranlogs.r-pkg.org/badges/grand-total/dvir?color=yellow)](https://cran.r-project.org/package=dvir)
[![](https://cranlogs.r-pkg.org/badges/last-month/dvir?color=yellow)](https://cran.r-project.org/package=dvir)
<!-- badges: end -->

# Disaster Victim Identification in R

The purpose of **dvir** is to implement state-of-the-art algorithms for
DNA-based disaster victim identification (DVI). In particular, **dvir**
performs *joint* identification of multiple victims.

The methodology and algorithms of **dvir** are described in

- Egeland & Vigeland (2025): [Kinship cases with partially specified
  hypotheses](https://doi.org/10.1016/j.fsigen.2025.103270)
- Vigeland & Egeland (2021): [Joint DNA-based Disaster Victim
  Identification](https://www.nature.com/articles/s41598-021-93071-5)

The **dvir** package is part of the
[pedsuite](https://magnusdv.github.io/pedsuite/), a collection of R
packages for pedigree analysis. Much of the machinery behind **dvir** is
imported from other pedsuite packages, especially **pedtools** for
handling pedigree data, and **forrel** for the calculation of likelihood
ratios. A comprehensive presentation of these packages, and much more,
can be found in the book [Pedigree Analysis in
R](https://shop.elsevier.com/books/pedigree-analysis-in-r/vigeland/978-0-12-824430-2).

## Installation

To get the current official version of **dvir**, install from CRAN as
follows:

``` r
install.packages("dvir")
```

Alternatively, the latest development version may be obtained from
GitHub:

``` r
# install.packages("remotes")
remotes::install_github("magnusdv/dvir")
```

## Example

We illustrate joint DVI analysis using a toy example from the paper
(Vigeland & Egeland, 2021), shown in the plot below. Three victim
samples (V1, V2, V3) are to be matched against three missing persons
(M1, M2, M3) in two different families.

<img src="man/figures/README-example-plot1-1.png" alt="" width="85%" style="display: block; margin: auto;" />

The hatched symbols indicate genotyped individuals. In this simple
example we consider only a single marker, with 10 equifrequent alleles
denoted 1, 2,…, 10. The available genotypes are shown in the figure.

DNA profiles from victims are referred to as *post mortem* (PM) data,
while the *ante mortem* (AM) data contains profiles from the reference
individuals R1 and R2.

### Assignments

A possible solution to the DVI problem is called an *assignment*. In our
toy example, there are *a priori* 14 possible assignments respecting the
individuals’ sex:

| V1  | V2  | V3  |
|:---:|:---:|:---:|
| \*  | \*  | \*  |
| \*  | \*  | M3  |
| \*  | M1  | \*  |
| \*  | M1  | M3  |
| \*  | M2  | \*  |
| \*  | M2  | M3  |
| M1  | \*  | \*  |
| M1  | \*  | M3  |
| M1  | M2  | \*  |
| M1  | M2  | M3  |
| M2  | \*  | \*  |
| M2  | \*  | M3  |
| M2  | M1  | \*  |
| M2  | M1  | M3  |

Each row indicates the missing persons corresponding to V1, V2 and V3
(in that order) with `*` meaning *not identified*. For example, the
first line contains the *null model* corresponding to none of the
victims being identified, while the last line gives the assignment where
`(V1, V2, V3) = (M1, M2, M3)`, For each assignment `a` we can calculate
the likelihood, denoted `L(a)`. The null likelihood is denoted `L0`.

Our aim in this brief tutorial is to perform the following tasks using
**dvir**:

1)  Rank the assignments according to how likely they are. For each
    assignment `a` we compute the LR comparing it to the null model:
    `LR = L(a)/L0`.
2)  Find the *posterior pairing probabilities* `P(V_i = M_j | data)` for
    all pairs of victims and missing persons.

### The data

The pedigrees and genotypes for this toy example are available within
**dvir** as a built-in dataset, under the name `example2`.

``` r
library(dvir)
```

``` r
example2
#> DVI dataset: 
#>  3 victims (2M/1F): V1, V2, V3
#>  3 missing (2M/1F): M1, M2, M3
#>  2 reference individuals: R1, R2
#>  2 pedigrees: 
#> Markers: 1
```

Internally, all DVI datasets in **dvir** have the structure of a list,
with elements `pm` (the victim data), `am` (the reference data) and
`missing` (a vector naming the missing persons): We can inspect the data
by printing each object. For instance, in this case `am` is a list of
two pedigrees:

``` r
example2$am
#> [[1]]
#>  id fid mid sex  L1
#>  M1   *   *   1 -/-
#>  R1   *   *   2 2/2
#>  M2  M1  R1   1 -/-
#> 
#> [[2]]
#>   id fid mid sex  L1
#>   R2   *   *   1 3/3
#>  MO2   *   *   2 -/-
#>   M3  R2 MO2   2 -/-
```

Each pedigree is printed in so-called *ped format*, with columns `id`
(ID label), `fid` (father’s ID, or `*` if no father is included), `mid`
(mother’s ID), `sex` (1 = male; 2 = female) and `L1` (genotypes at locus
`L1`).

The code generating this dataset can be found in the github repository
of **dvir**, more specifically here:
<https://github.com/magnusdv/dvir/blob/master/data-raw/example2.R>.

A great way to inspect a DVI dataset is to plot it with the function
`plotDVI()`.

``` r
plotDVI(example2)
```

<img src="man/figures/README-example-plot2-1.png" alt="" width="80%" style="display: block; margin: auto;" />

The `plotDVI()` function offers many parameters for tweaking the plot;
see the help page `?plotDVI()` for details.

### Joint identification

The `jointDVI()` function performs joint identification of all three
victims, given the data. It returns a data frame ranking all assignments
with nonzero likelihood:

``` r
jointRes = jointDVI(example2, verbose = FALSE)

# Print the result
jointRes
#>    V1 V2 V3    loglik  LR   posterior LR0
#> 1  M1 M2 M3 -16.11810 250 0.718390805   5
#> 2  M1 M2  * -17.72753  50 0.143678161   5
#> 3   * M2 M3 -18.42068  25 0.071839080   1
#> 4  M1  * M3 -20.03012   5 0.014367816   1
#> 5   * M1 M3 -20.03012   5 0.014367816   1
#> 6   * M2  * -20.03012   5 0.014367816   1
#> 7   *  * M3 -20.03012   5 0.014367816   1
#> 8  M1  *  * -21.63956   1 0.002873563   1
#> 9   * M1  * -21.63956   1 0.002873563   1
#> 10  *  *  * -21.63956   1 0.002873563   1
```

The output shows that the most likely joint solution is (V1, V2, V3) =
(M1, M2, M3), with an LR of 250 compared to the null model.

The function `plotSolution()` shows the most likely solution:

``` r
plotSolution(example2, jointRes, marker = 1)
```

<img src="man/figures/README-solution-1.png" alt="" width="75%" style="display: block; margin: auto;" />

By default, the plot displays the assignment in the first row of
`jointRes`. To examine the second most likely, add `k = 2` (and so on to
go further down the list).

### Posterior pairing probabilities

Next, we compute the posterior pairing (and non-pairing) probabilities.
This is done by feeding the output from `jointDVI()` into the function
`Bmarginal()`.

``` r
Bmarginal(jointRes, example2$missing, prior = NULL)
#>            M1        M2        M3          *
#> V1 0.87931034 0.0000000 0.0000000 0.12068966
#> V2 0.01724138 0.9482759 0.0000000 0.03448276
#> V3 0.00000000 0.0000000 0.8333333 0.16666667
```

Here we used a default flat prior for simplicity, assigning equal prior
probabilities to all assignments.

We see that the posterior pairing probabilities for the most likely
solution are

- *P*(V1 = M1 \| data) = 0.88,
- *P*(V2 = M2 \| data) = 0.95,
- *P*(V3 = M3 \| data) = 0.83.

------------------------------------------------------------------------

#' Joint DVI analysis
#'
#' Joint matching of victims and missing persons in a disaster victim identification (DVI)
#' case, as described by Vigeland & Egeland (2021).
#'  
#' NOTE: This is a legacy function. For new projects we recommend these instead:
#'   * `dviSolve()` (for a complete DVI pipeline)
#'   * `dviJoint()` (for fully joint analysis)
#'   
#' The `jointDVI()` function was the first implementation of a complete DVI analysis in 
#' `dvir`. It both manages the analysis workflow (including exclusion of impossible
#' pairings and detection of undisputed matches) and performs the final joint analysis.
#' As the package evolved, it became beneficial to split these two roles, resulting in
#' the two functions above.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param pairings A list of possible pairings for each victim. If NULL, all
#'   sex-consistent pairings are used.
#' @param ignoreSex A logical.
#' @param assignments A data frame containing the assignments to be considered
#'   in the joint analysis. By default, this is automatically generated by
#'   taking all combinations from `pairings`.
#' @param limit A positive number, by default 0. Only pairwise LR values above
#'   this are considered.
#' @param nkeep An integer, or NULL. If given, only the `nkeep` most likely
#'   pairings are considered for each victim.
#' @param markers A vector indicating which markers should be included in the
#'   analysis. By default all markers are included.
#' @param disableMutations A logical, or NA (default). The default action is to
#'   disable mutations in all reference families without Mendelian errors.
#' @param undisputed A logical, by default TRUE.
#' @param maxAssign A positive integer. If the number of assignments going into
#'   the joint calculation exceeds this, the function will abort with an
#'   informative error message. Default: 1e5.
#' @param threshold A positive number, passed onto [findUndisputed()]. Default:
#'   1e4.
#' @param strict A logical, passed onto [findUndisputed()]. Default: FALSE.
#' @param relax Deprecated.
#' @param numCores An integer; the number of cores used in parallelisation.
#'   Default: 1.
#' @param check A logical, indicating if the input data should be checked for
#'   consistency.
#' @param verbose A logical.
#' @param jointRes A data frame produced by `jointDVI()`.
#' @param LRthresh A positive number, used as upper limit for the LR comparing the
#'   top result with all others.
#' 
#' @return A data frame. Each row describes an assignment of victims to missing
#'   persons, accompanied with its log likelihood, the LR compared to the null
#'   (i.e., no identifications), and the posterior corresponding to a flat
#'   prior.
#'
#'   The function `compactJointRes()` removes columns without assignments, and
#'   solutions whose LR compared with the top result is below `1/LRthresh`. 
#' 
#' @references 
#' Vigeland, MD., Egeland, T. Joint DNA-based disaster victim identification. 
#' Sci Rep 11, 13661 (2021).
#' 
#' @seealso [pairwiseLR()], [findUndisputed()]
#'
#' @examples
#' jointDVI(example2)
#'
#' @importFrom utils setTxtProgressBar txtProgressBar
#'
#' @export
jointDVI = function(dvi, pairings = NULL, ignoreSex = FALSE, assignments = NULL, 
                    limit = 0, nkeep = NULL, undisputed = TRUE, markers = NULL, 
                    threshold = 1e4, strict = FALSE, relax = !strict, disableMutations = NA, 
                    maxAssign = 1e5, numCores = 1, check = TRUE, verbose = TRUE) {
  
  message(
"NOTE: `jointDVI()` is a legacy function. Please consider these instead:
 * `dviSolve()` (for a complete DVI pipeline)
 * `dviJoint()` (for fully joint analysis)"
)
  
  st = Sys.time()
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  if(length(dvi$pm) == 0)
    undisputed = FALSE
  
  origVics = vics = names(dvi$pm)
  
  if(!is.null(markers)) {
    stop2("Marker selection not implemented")
    # pm = selectMarkers(pm, markers)
    # am = selectMarkers(am, markers)
  }
  
  if(verbose)
    print.dviData(dvi)
  
  if(check)
    checkDVI(dvi, pairings = pairings, ignoreSex = ignoreSex, verbose = verbose)
  
  ### Mutation disabling
  if(any(allowsMutations(dvi$am))) {
    am = dvi$am
    
    if(verbose) 
      cat("\nMutation modelling:\n")
    
    if(isTRUE(disableMutations)) {
      if(verbose) cat(" Disabling mutations in all families\n")
      disableFams = seq_along(am)
    }
    else if(identical(disableMutations, NA)) {
      am.nomut = setMutmod(am, model = NULL)
      badFams = vapply(am.nomut, loglikTotal, FUN.VALUE = 1) == -Inf
      if(verbose) {
        if(any(badFams)) 
          cat("", sum(badFams), "inconsistent families:", trunc(which(badFams)), "\n")
        cat("", sum(!badFams), "consistent families. Disabling mutations in these\n")
      }
      disableFams = which(!badFams)
    }
    else disableFams = NULL
  
    if(length(disableFams)) {
      am[disableFams] = setMutmod(am[disableFams], model = NULL)
    }
    
    # Update DVI object
    dvi$am = am
  }
  
  ### Identify and fixate "undisputed" matches
  undisp = as.data.frame(list())
  
  if(undisputed && is.null(assignments)) {
    
    r = findUndisputed(dvi, pairings = pairings, ignoreSex = ignoreSex, 
                       threshold = threshold, strict = strict, limit = limit, 
                       nkeep = nkeep, numCores = numCores, verbose = verbose)
    
    # List of undisputed, and their LR's
    undisp = r$summary
    Nun = nrow(undisp) %||% 0
    
    # If all are undisputed, return early
    # Either all *victims* are matched or all *missing* are identified
    # The code below covers both scenarios
    if(Nun == length(dvi$pm) || Nun == length(dvi$missing)) {
      
      # Build solution assignment (must be a data frame)
      sol = rep(list("*"), length(dvi$pm))
      names(sol) = vics
      sol[undisp$Sample] = undisp$Missing
      sol = as.data.frame(sol)
      
      # Run through jointDVI() with the solution as the only assignment 
      res = jointDVI(dvi, ignoreSex = ignoreSex, assignments = sol, undisputed = FALSE,
                     markers = markers, threshold = NULL, check = FALSE, verbose = FALSE)
      return(res)
    }
    
    # Reduced DVI problem to be used in the joint analysis
    dvi = r$dviReduced
    vics = names(dvi$pm)
    
    # Pairings: These exclude those with LR = 0!
    pairings = r$pairings
    
    if(verbose && Nun > 0)
      print.dviData(dvi, heading = "\nReduced DVI dataset:")
  }
  
  pm = dvi$pm 
  am = dvi$am
  
  if(is.null(pairings) && is.null(assignments))
    pairings = pairwiseLR(dvi, pairings = pairings, ignoreSex = ignoreSex, limit = limit, nkeep = nkeep)$pairings
 
  if(is.null(assignments)) {
    if(verbose) cat("\nCalculating pairing combinations\n")
    # Expand pairings to assignment data frame
    assignments = expand.grid.nodup(pairings, max = maxAssign)
  }
  else {
    if(verbose) cat("\nChecking supplied pairing combinations\n")
    if(!setequal(names(assignments), origVics))
      stop2("Names of `assignments` do not match `pm` names")
    assignments = assignments[origVics]
  }
  
  nAss = nrow(assignments)
  if(nAss == 0)
    stop2("No possible solutions!")
  if(verbose)
    cat("Assignments to consider in the joint analysis:", nAss, "\n\n")
  
  # Convert to list; more handy below
  assignmentList = lapply(1:nAss, function(i) as.character(assignments[i, ]))
  
  # Initial loglikelihoods
  logliks.PM = vapply(pm, loglikTotal, FUN.VALUE = 1)
  logliks.AM = vapply(am, loglikTotal, FUN.VALUE = 1)
  
  loglik0 = sum(logliks.PM) + sum(logliks.AM)
  if(loglik0 == -Inf)
    stop2("Impossible initial data: AM component ", which(logliks.AM == -Inf))
  
  # Prepare no-ref analysis (to catch related victims)
  am.noref = removeGenotypes(am)
  logliks.AM.noref = rep(0, length(am)) |> setNames(names(am))
  loglik0.noref = sum(logliks.PM)
  
  # Parallelise
  if(numCores > 1) {
    
    if(verbose) 
      cat("Using", numCores, "cores\n")
    
    cl = makeCluster(numCores)
    on.exit(stopCluster(cl))
    clusterEvalQ(cl, library(dvir))
    clusterExport(cl, "loglikAssign", envir = environment())

    # Main likelihood calculation: Loop through assignments
    loglik = parLapply(cl, assignmentList, function(a) 
      loglikAssign(pm, am, vics, a, loglik0, logliks.PM, logliks.AM))
	
  	# Same with empty AM
  	loglik.noref = parLapply(cl, assignmentList, function(a) 
        loglikAssign(pm, am.noref, vics, a, loglik0.noref, logliks.PM, logliks.AM.noref))
  }
  else {
    # Setup progress bar
    if(progbar <- verbose && interactive())
      pb = txtProgressBar(min = 0, max = nAss, style = 3)

    # Main likelihood calculation: Loop through assignments    
    loglik = lapply(seq_len(nAss), function(i) {
      if(progbar) setTxtProgressBar(pb, i)
      loglikAssign(pm, am, vics, assignmentList[[i]], loglik0, logliks.PM, logliks.AM)
    })
	
  	# Same with empty AM
  	loglik.noref = lapply(seq_len(nAss), function(i) {
        loglikAssign(pm, am.noref, vics, assignmentList[[i]], loglik0.noref, logliks.PM, logliks.AM.noref)
  	})
    
    # Close progress bar
    if(progbar) close(pb)
  }
  
  loglik = unlist(loglik)
  
  LR = exp(loglik - loglik0)
  posterior = LR/sum(LR) # assumes a flat prior
  
  LR0 = exp(unlist(loglik.noref) - loglik0.noref)
  
  # Add undisputed matches
  Nun = nrow(undisp) %||% 0
  if(Nun > 0) {
    # Add ID columns
    assignments[, undisp$Sample] = rep(undisp$Missing, each = nAss)
    
    # Bug fix: Add * columns for victims lost in `subsetDVI` in undisputed
    missVic = setdiff(origVics, names(assignments))
    for(v in missVic)
      assignments[[v]] = "*"
    
    # Fix ordering
    assignments = assignments[origVics]
    
    # Fix LR: Multiply with that of the undisputed
    LR = LR * prod(undisp$LR)
  }
    
  # Collect results
  tab = cbind(assignments, loglik = loglik, LR = LR, posterior = posterior, LR0 = LR0)
  
  # Sort in decreasing likelihood, break ties with grid
  g = assignments
  g[g == "*"] = NA
  tab = tab[do.call(order, g), , drop = FALSE] # first sort assignments alphabetically
  tab = tab[order(round(tab$loglik, 10), decreasing = TRUE), , drop = FALSE]
  
  rownames(tab) = NULL
  
  if(verbose)
    cat("Time used:", format(Sys.time() - st, digits = 3), "\n")
  
  tab
}

#' @rdname jointDVI
#' @export
compactJointRes = function(jointRes, LRthresh = NULL) {
  if(!is.null(LRthresh)) {
    # Require LR >= threshold
    keepRows = jointRes$LR >= LRthresh
    
    # Also require LR_1:a > threshold (i.e. with top result as numerator)
    logLRtop = jointRes$loglik[1] - jointRes$loglik
    keepRows = keepRows & logLRtop - log(LRthresh) < sqrt(.Machine$double.eps)
    
    # Last row to keep
    last = match(TRUE, keepRows, nomatch = 0L) + 1L
    jointRes = jointRes[1:last, , drop = FALSE]
  }
  
  empty = sapply(jointRes, function(v) all(v == "*"))
  jointRes[, !empty, drop = FALSE]
}


# Function for computing the total log-likelihood of a single assignment
loglikAssign = function(pm, am, vics, assignment, loglik0, logliks.PM, logliks.AM) {
  
  # Victims which actually move
  vicMove = vics[assignment != "*"]
  mpsMove = assignment[assignment != "*"]
  
  if(length(vicMove) == 0)
    return(loglik0)
  
  # The relevant AM components
  compNo = unique.default(getComponent(am, mpsMove, checkUnique = TRUE))
  
  # Move victim data
  changedComps = transferMarkers(pm[vicMove], am[compNo], idsFrom = vicMove, 
                                 idsTo = mpsMove, erase = FALSE)

  # Update likelihood of modified AM comps
  logliks.AM.new = logliks.AM
  logliks.AM.new[compNo] = vapply(changedComps, function(a) loglikTotal(a), FUN.VALUE = 1)
  
  # Likelihood of remaining PMs
  logliks.PM.new = logliks.PM[setdiff(vics, vicMove)]
  
  # Return total loglik of assignments
  sum(logliks.PM.new) + sum(logliks.AM.new)
}

#' Finds Generalised Likelihood Ratios (GLRs)
#'
#' Based on a `dviData` object, or output from `dviJoint()``or `jointDVI()`, 
#' the GLR, the ratio of the maximum likelihood under H0 to the maximum under H1,
#' is calculated for specified hypotheses.
#'
#' @param dvi A `dviData` object.
#' @param pairings List. See details.
#' @param dviRes data frame. Output from `jointDVI()`.
#'
#' @return A data frame with GLRs and SGLR (strict GLR, max replaced by min in 
#'   the numerator).
#'
#' @details The Generalised Likelihood Ratio (GLR) statistic is defined as the
#'   ratio of the maximum likelihood for the alternatives in the numerator
#'   to the maximum in the denominator. The default 
#'   `pairings = dvir::generatePairings(dvi)` tests all hypotheses. Specific 
#'   tests can be specified as shown in an example:
#'   `pairings = list(V1 = "M1")` gives a test for H0: V1 = M1 against 
#'   H1: V1 != M1. `dviRes` will be calculated using `jointDVI()` if not provided.
#'
#' @examples
#' # dviGLR(example2, pairings = list(V1 = "M1"))
#'
#' # All tests with output from jointDVI
#' # dviGLR(example2)
#'
dviGLR = function(dvi, pairings = generatePairings(dvi), dviRes = NULL){
  if(!inherits(dvi, "dviData"))
    stop("First argument needs to be a dviData object")

  # Do joint analysis, if not provided, and find max loglikelihood
  if(is.null(dviRes))
    dviRes = jointDVI(dvi)
  loglik = dviRes$loglik
  na = dim(dviRes)[1]

  # The number of hypotheses to be tested
  nhyp = sum(unlist(lapply(pairings, function(x) length(x))))

  # Initialise log likelihoods, max and min under H0, and max under H1
  l0Max = l0Min = l1 = rep(-Inf, nhyp)

  # To contain names of null hypotheses in output:
  hyp = rep(NA, nhyp)

  # Loop through hypotheses for each victim
  npairs = length(pairings)
  s = 0
  for (i in 1:npairs){
    victim = names(pairings[i])
    for(j in pairings[[i]]){
      s = s + 1
      hyp[s] = paste(victim, j, sep = "=")
      
      # Find indices for numerator loglikelihoods
      I0 = (1:na)[dviRes[,victim] == j]
      I1 = setdiff(1:na, I0)
      if(length(I0)) {
        l0Max[s] = max(loglik[I0])
        l0Min[s] = min(loglik[I0])
      }
      if(length(I1)) l1[s] = max(loglik[I1])
    }
  }
  GLR = exp(l0Max - l1)
  SGLR = exp(l0Min - l1)
  data.frame(H0 = hyp, GLR = GLR, GLRmin = SGLR)
}

#' Posterior pairing probabilities
#'
#' Compute posterior pairing and non-pairing probabilities, based on a prior and
#' the output from [jointDVI()].
#'
#' The prior assigns a probability to each assignment, each row of `jointRes`.
#' If the prior is not specified, a flat prior is used. The prior needs not sum
#' to 1 since the user may rather choose a flat prior on the *a priori* possible
#' assignments.
#'
#' @param jointRes Output from [jointDVI()].
#' @param missing Character vector with names of missing persons.
#' @param prior A numeric vector of length equal the number of rows in
#'   `jointRes`. Default is a flat prior.
#'
#' @return A matrix. Row `i` gives the posterior probability that victim `i` is
#'   one of the missing persons or someone else, denoted '*'.
#'
#' @seealso [jointDVI()]
#'
#' @examples
#' jointRes = jointDVI(example1)
#'
#' Bmarginal(jointRes, example1$missing)
#'
#' # Artificial example: all but optimal solution excluded by prior
#' Bmarginal(jointRes, example1$missing, prior = c(1, rep(0,26)))
#'
#'
#' @export
Bmarginal = function(jointRes, missing, prior = NULL){
  
  d = dim(jointRes)
  
  # Deal with prior
  if(is.null(prior))
    prior = rep(1/d[1], d[1])  
  else if(length(prior) != d[1])
    stop2("Length of prior must equal number of rows in first argument")
  
  # Find names of victims
  nv = (1:d[2])[names(jointRes) == "loglik"] - 1
  victims = names(jointRes)[1:nv]
  
  # Initialise result table 
  val = c(missing, '*')
  res = matrix(nrow = nv, ncol = length(val), 0)
  dimnames(res) = list(victims, val)
  
  # Find denominator, const, and do calculations
  # Subtract max loglik in numerator and denominator to avoid underflow
  lmax = max(jointRes$loglik) 
  x = data.frame(jointRes, term = prior * exp(jointRes$loglik-lmax))
  const = sum(x$term)  
  for(v in 1:nv){
    r1 = split(x$term, x[,v])
    p = sapply(r1, sum)/const
    res[v, names(p)] = p
  }
  
  res
}

#' Plot a DVI problem
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param pm Either a logical indicating if the PM data should be plotted (as a
#'   set of singletons), or a vector of indices selecting a subset of the PM
#'   samples. Default: TRUE.
#' @param am Either a logical indicating if the AM families data should be
#'   plotted, or a vector of indices selecting a subset of the families.
#'   Default: TRUE.
#' @param style An integer (currently 1 or 2) indicating the style of the plot.
#' @param hatched A character vector of ID labels, or the name of a function. By
#'   default, typed individuals are hatched.
#' @param fill,cex,col,lwd,carrier Arguments passed on to [pedtools::plot.ped()].
#' @param frames A logical, by default TRUE.
#' @param titles A character of length 2.
#' @param famnames A logical. If NA (default) family names are included if there
#'   are multiple families.
#' @param widths,heights Numeric with relative columns widths / row heights, to
#'   be passed on to `layout()`.
#' @param nrowPM The number of rows in the array of PM singletons.
#' @param nrowAM,ncolAM The number of rows/cols in the array of AM families.
#' @param dev.height,dev.width Plot height and widths in inches. These are
#'   optional, and only relevant if `newdev = TRUE`.
#' @param newdev A logical indicating if a new plot window should be opened.
#' @param ... Further parameters to be passed on to [pedtools::plot.ped()],
#'   e.g., `marker`, `cex`, `cex.main`, `symbolsize`.
#'
#' @return NULL
#'
#' @examples
#'
#' plotDVI(example2)
#'
#' # Override default layout
#' plotDVI(example2, nrowPM = 1, ncolAM = 1)
#'
#' # Subset
#' plotDVI(example2, pm = 1:2, am = 1, titles = c("PM (1-2)", "AM (1)"))
#'
#' # AM only
#' plotDVI(example2, pm = FALSE, titles = "AM families")
#'
#' # Further plot options
#' plotDVI(example2, frames = FALSE, marker = 1, cex = 1.2, nrowPM = 3, nrowAM = 2,
#'   textAnnot = list(inside = list(c(M1 = "?", M2 = "?", M3 = "?"), cex = 1.5)))
#'
#' @importFrom grDevices dev.new
#' @importFrom graphics grconvertX grconvertY layout mtext par rect
#' @export
plotDVI = function(dvi, pm = TRUE, am = TRUE, style = 1, famnames = NA,  
                   hatched = typedMembers, frames = TRUE, titles = c("PM", "AM"),
                   cex = NA, col = 1, lwd = 1, fill = NA, carrier = NULL, 
                   widths = NULL, heights = NULL, nrowPM = NA, nrowAM = NA, ncolAM = NA,
                   dev.height = NULL, dev.width = NULL, 
                   newdev = !is.null(c(dev.height, dev.width)), ...) {
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  # Extra args
  args = list(...)
  
  PM = if(isTRUE(pm)) dvi$pm else if(!is.logical(pm)) dvi$pm[pm] else NULL
  AM = if(isTRUE(am)) dvi$am else if(!is.logical(am)) dvi$am[am] else NULL
  
  npm = length(PM)
  nam = length(AM)
  if(npm + nam == 0) {
    message("Nothing to plot")
    return(invisible())
  }
    
  if(nam == 0) {
    if(newdev)
      stop2("Cannot plot only PM to new device")
    plotPM(PM, hatched = hatched, title = titles[1], nrow = nrowPM, ...)
    return(invisible())
  }
  
  # Ad hoc cex increase for > 2 panels
  if(is.na(cex)) {
    npanels = nam + (npm > 0)
    cex = if(npanels %in% 3:6) 1.3 else if(npanels %in% 7:9) 1.2 else 1
  }
  
  if(is.na(famnames))
    famnames = nam > 1
  
  # AM layout
  amDim = amArrayDim(nam, nrowAM, ncolAM)
  pmDim = pmArrayDim(npm, nrowPM)
  layoutMat = matrix(1:prod(amDim), nrow = amDim[1], ncol = amDim[2], byrow = TRUE)
  
  if(npm > 0)
    layoutMat = cbind(1, layoutMat + 1)
  
  # Relative widths/heights & plot dimensions
  plotdims = findPlotDims(AM, amDim, pmDim)
  widths = widths %||% plotdims$widths
  heights = heights %||% plotdims$heights |> (\(x) {x[1] = x[1] + famnames * 0.2; x})()

  # Open new device?
  if (newdev) 
    dev.new(height = dev.height %||% plotdims$H, 
            width = dev.width %||% plotdims$W, 
            noRStudioGD = TRUE)
  
  # Set graphical parameters (include mfrow to ensure layout is reverted on exit)
  opar = par(oma = c(0, 0, 3, 0), xpd = NA, mfrow = c(1,1), mar = c(0,0,0,0))
  on.exit(par(opar))
  
  #layout(rbind(1:(1 + length(AM))), widths = widths)
  layout(layoutMat, widths = widths, heights = heights)
  
  # Top margin
  topmar = 0.5 + famnames + !is.null(titles)
  
  # Plot PMs in panel 1 (left)
  if(npm > 0)
    plotPM(PM, nrow = pmDim[1], hatched = hatched, margins = c(2,2,topmar,2), 
           cex = cex, fill = fill, col = col, lwd = lwd, carrier = carrier,...)
  
  # Color style
  miss = dvi$missing
  if(style == 1) 
    fill = list("pink" = miss)
  else if(style == 2) {
    col = list("red" = miss); carrier = miss; lwd = list("1.2" = miss)} 
  
  # Loop through AMs
  nms = names(AM)
  for(i in 1:nam) {
    rw = ceiling(i / amDim[2])
    mar = c(1.5,2,1.5,2)
    if(rw == 1) mar[3] = topmar
    if(rw == amDim[2]) mar[1] = 2
    plot(AM[[i]], hatched = hatched, title = if(famnames) nms[i], margins = mar,
         cex = cex, cex.main = cex+0.2, fill = fill, col = col, lwd = lwd, carrier = carrier, ...)
  }
  
  # X coordinate of each plot region (converted to value in [0,1]).
  ratios = c(0, if(npm > 0) widths[1], sum(widths))/sum(widths)
  
  # Draw frames
  if(frames) {
    margIn = grconvertY(1, from = "lines", to = "inches")
    margx = grconvertX(margIn, from = "inches", to = "ndc")
    margy = grconvertY(margIn, from = "inches", to = "ndc")
    
    rect(xleft   = grconvertX(ratios[1:2] + margx, from = "ndc"),
         ybottom = grconvertY(1 - margy, from = "ndc"),
         xright  = grconvertX(ratios[2:3] - margx, from = "ndc"),
         ytop    = grconvertY(margy, from = "ndc"), 
         xpd = NA)
  }
  
  # Add titles
  if(npm == 0 && length(titles) == 2)
    titles = titles[2]
  
  midpoints = ratios[1:2] + diff(ratios)/2
  cex.main = args[["cex.main"]] %||% cex + 0.3
  
  font.main = args[["font.main"]] %||% 1
  
  mtext(titles, outer = TRUE, at = midpoints, cex = cex.main, font = font.main)
  
  invisible(c(plotdims, list(layout = layoutMat)))
}

# Default layout of AM families: Up to 5 columns
amArrayDim = function(N, nrow = NA, ncol = NA) {
  if(!is.na(nrow) && !is.na(ncol)) {
    if(nrow * ncol < N)
      stop2(sprintf("'nrow * ncol' must be >= %d. Provided: %d * %d = %d", 
                    N, nrow,ncol,nrow*ncol))
    return(c(nrow, ncol))
  }

  if(!is.na(ncol)) {
    if(ncol > N) stop2("`ncolAM` is too large")
    return(c(ceiling(N/ncol), ncol))
  }
  
  if(!is.na(nrow)) {
    if(nrow > N) stop2("`nrowAM` is too large")
    return(c(nrow, ceiling(N/nrow)))
  }
  
  # No input
  maxcol = if(N <= 9) 4 else if(N <= 15) 5 else if(N<=24) 6 else 7
  nr = (N-1) %/% maxcol + 1
  c(nr, ceiling(N/nr))
}

# Default layout of PM singletons
pmArrayDim = function(N, nrow = NA) {
  if(is.na(nrow))
    nr = if(N <= 3) N else ceiling(sqrt(N))
  else {
    if(nrow > N) stop2("`nrowPM` is too large")
    nr = min(N, nrow)
  }
  c(nr, ceiling(N/nr))
}


findPlotDims = function(am, amDim = NULL, pmDim = NULL, npm = NULL) {
  amDim = amDim %||% amArrayDim(length(am))
  pmDim = pmDim %||% pmArrayDim(npm)
  
  # Alignment data for each am family
  alignlist = lapply(am, .pedAlignment)
  
  xrange = vapply(alignlist, function(a) max(1, diff(a$xrange)), 1) |> 
    `length<-`(prod(amDim)) |> 
    matrix(nrow = amDim[1], byrow = TRUE)
  
  yrange = vapply(alignlist, function(a) a$maxlev, 1) |> 
    `length<-`(prod(amDim)) |> 
    matrix(nrow = amDim[1], byrow = TRUE)
    
   # AM grid size
  maxWidth = max(rowSums(yrange, na.rm = TRUE))
  maxGen = max(colSums(yrange, na.rm = TRUE))
  
  # Relative widths & heights
  maxXrange = apply(xrange, 2, max, na.rm = TRUE)
  maxYrange = apply(yrange, 1, max, na.rm = TRUE)
  
  widths = sqrt(maxXrange) + 1
  heights = 1 + maxYrange/2
  
  # Add width of PM column if present
  if(pmDim[1] > 0) {
    pmwid = max(sqrt(pmDim[2]), sum(widths)/3)
    widths = c(pmwid, widths)
  }
  
  # Suggested plot dims in inches
  W = sum(widths + 1)
  H = max(3, 1.2 * maxGen) + 0.3
  
  list(amSize = c(maxWidth, maxGen), pmSize = pmDim, 
       widths = widths, heights = heights, W = W, H = H)
}



plotPM = function(pm, nrow = NA, ...) {

  # alignment ---------------------------------------------------------------
  
  N = length(pm)
  dims = pmArrayDim(N, nrow)
  nr = dims[1]
  nc = dims[2]
  
  nid = matrix(c(1:N, rep(0, nr*nc - N)), nrow = nr, ncol = nc, byrow = TRUE)
  pos = col(nid) - 1
  
  n = rowSums(nid > 0)
  fam = spouse = matrix(0, nrow = nr, ncol = nc)
  
  plist = list(n = n, nid = nid, pos = pos, fam = fam, spouse = spouse)
  align = pedtools::.pedAlignment(pm, plist = plist)
  
  # annotation --------------------------------------------------------------
  
  annotlist = lapply(pm, function(s) .pedAnnotation(s, ...))
  annot1 = annotlist[[1]]
  
  mrg = function(key, def)
    unlist(lapply(annotlist, function(a) a[[key]] %||% def))
  
  annot = list(
    title = annot1$title,
    textUnder = mrg("textUnder", ""),
    textAbove = mrg("textAbove", ""),
    textInside = mrg("textInside", ""),
    colvec = mrg("colvec", 1), 
    densvec = mrg("densvec", 0), 
    fillvec = mrg("fillvec", NA), 
    ltyvec = mrg("ltyvec", 1), 
    lwdvec = mrg("lwdvec", 1),
    carrierTF = mrg("carrierTF", FALSE),
           deceasedTF = mrg("deceasedTF", FALSE))
  
  p = drawPed(align, annot, ...)
  invisible(p)
}



#' Plot DVI solution
#'
#' A version of [plotDVI()] tailor-made to visualise identified individuals, for
#' example as reported by `jointDVI()`.
#'
#' @param dvi A `dviData` object.
#' @param assignment A named character of the format `c(victim = missing, ...)`,
#'   or a data frame produced by [jointDVI()].
#' @param k An integer; the row number when `assignment` is a data frame.
#' @param format A string indicating how identified individuals should be
#'   labelled, using `[M]` and `[S]` as place holders for the missing person and
#'   the matching sample, respectively. (See Examples.)
#' @param ... Further arguments passed on to [plotDVI()].
#'
#' @return NULL.
#'
#' @examples
#'
#' res = jointDVI(example2, verbose = FALSE)
#'
#' plotSolution(example2, res)
#'
#' # With line break in labels
#' plotSolution(example2, res, format = "[M]=\n[S]")
#'
#' # With genotypes for marker 1
#' plotSolution(example2, res, marker = 1)
#'
#' # Non-optimal solutions
#' plotSolution(example2, res, k = 2, pm = FALSE)
#' plotSolution(example2, res, k = 2, cex = 1.3)
#'
#' @export
plotSolution = function(dvi, assignment, k = 1, format = "[S]=[M]", ...) {
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  a = assignment
  
  if(is.data.frame(a))
    a = unlist(a[k, names(dvi$pm), drop = FALSE])
  
  # Vector of matching pairs
  mtch = a[a != "*"]
  vics = names(mtch)
  
  ### Label format for matching individuals
  fmt = sub("[S]", "%{sample}s", sub("[M]", "%{miss}s", format, fixed = TRUE), fixed = TRUE)
  newlabs = sprintfNamed(fmt, sample = vics, miss = mtch)
  
  # Avoids error if constant format
  if(length(newlabs) != length(mtch))
    newlabs = rep_len(newlabs, length.out = length(mtch))
  
  refs = typedMembers(dvi$am)
  
  if(length(mtch)) {
    am = relabel(dvi$am, old = mtch, new = newlabs)
    am = transferMarkers(dvi$pm[vics], am, idsFrom = vics, idsTo = newlabs, erase = FALSE)
    dvi$am = am
  }
  
  dvi$pm = dvi$pm[a == "*"]
  
  stillmissing = setdiff(dvi$missing, mtch)
  
  plotDVI(dvi, style = 0, fill = list("green" = newlabs), hatched = refs, col = list("red" = stillmissing), 
          lwd = list("1.5" = c(stillmissing)), ...)
}

#' Automatic labelling of a DVI dataset
#'
#' Relabel the individuals and families in a DVI dataset.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param victims A named vector of the form `c(old = new)` with names for the
#'   PM samples, or a function to be applied to the existing names.
#' @param victimPrefix Prefix used to label PM individuals.
#' @param familyPrefix Prefix used to label the AM families.
#' @param refs A named vector of the form `c(old = new)` with names for the
#'   typed references, or a function to be applied to the existing names.
#' @param refPrefix Prefix used to label the reference individuals, i.e., the
#'   typed members of the AM families.
#' @param missingPrefix Prefix used to label the missing persons. At most one of
#'   `missingPrefix` and `missingFormat` can be given.
#' @param missingFormat A string indicating family-wise labelling of missing
#'   persons, using `[FAM]`, `[IDX]`, `[MIS]` as place holders with the
#'   following meanings (see Examples):
#'   * `[FAM]`: family index
#'   * `[IDX]`: index of missing person within the family
#'   * `[MIS]`: index within all missing persons
#' @param othersPrefix Prefix used to label other untyped individuals. Use ""
#'   for numeric labels ( 1, 2, ...).
#'
#' @return A [dviData()] object.
#'
#' @examples
#'
#' # Builtin dataset `example2`
#' relabelDVI(example2,
#'            victimPrefix  = "vic",
#'            familyPrefix  = "fam",
#'            refPrefix     = "ref",
#'            missingPrefix = "mp")
#'
#' # Family-wise labelling of missing persons
#' relabelDVI(example2, missingFormat = "M[FAM]-[IDX]")
#' relabelDVI(example2, missingFormat = "M[IDX] (F[FAM])")
#' relabelDVI(example2, missingFormat = "fam[FAM].m[IDX]")
#'
#' @export
relabelDVI = function(dvi, victims = NULL, victimPrefix = NULL, familyPrefix = NULL,
                      refs = NULL, refPrefix = NULL, missingPrefix = NULL, 
                      missingFormat = NULL, othersPrefix = NULL) {
  
  dvi = consolidateDVI(dvi)
  
  pm = dvi$pm
  am = dvi$am
  missing = dvi$missing
  
  npm = length(pm)
  nam = length(am)
  nmiss = length(missing)
  
  if(identical(missingPrefix, "family") && is.null(names(am)) && is.null(familyPrefix))
    familyPrefix = "F"
  
  # Relabel PM data
  if(npm > 0) {
    if(!is.null(victims))
      pm = relabel(pm, new = victims)
    else if(!is.null(victimPrefix)) {
      if(length(victimPrefix) != 1)
        stop2("`victimPrefix` must have length 1: ", victimPrefix)
      victims = paste0(victimPrefix, 1:npm)
      pm = relabel(pm, new = victims)
    }
  }
  
  # If no AM data: Return early
  if(nam == 0)
    return(dviData(pm = pm, am = am, missing = missing))  
  

  # AM data -------------------------------------------------------------------------------------

  # Relabel typed reference individuals
  if(!is.null(refPrefix)) {
    
    if(length(refPrefix) != 1)
      stop2("`refPrefix` must have length 1: ", refPrefix)
    
    refs = paste0(refPrefix, seq_along(typedMembers(am)))
  }
  
  if((!is.null(refs))) {
    am = relabel(am, new = refs, old = typedMembers(am))
  }
  
  # Rename AM families
  if(!is.null(familyPrefix)) {
    
    if(length(familyPrefix) != 1)
      stop2("`familyPrefix` must have length 1: ", familyPrefix)
    
    names(am) = paste0(familyPrefix, 1:nam)
  }
  
  
  # Relabel missing persons
  if(!is.null(missingPrefix) || !is.null(missingFormat)) {
  
    if(!is.null(missingPrefix) && !is.null(missingFormat))
      stop2("At most one of `missingPrefix` and `missingFormat` can be used")
    
    # Temporary fix if same label is used in multiple components (e.g. "Missing person")
    labs = unlist(labels(am), use.names = FALSE)
    if(any(labs[duplicated(labs)] %in% missing)) {
      tmpmiss = character()
      
      for(i in seq_along(am)) {
        comp = am[[i]]
        missi = intersect(missing, comp$ID)
        if(!length(missi))
          next
        am[[i]] = relabel(comp, missi, new = paste(missi, i, sep = "."))
        tmpmiss = c(tmpmiss, paste(missi, i, sep = "."))
      }
      
      missing = tmpmiss
      nmiss = length(missing)
    }
    
    # Regular prefix format
    if(!is.null(missingPrefix)){
      
      if(length(missingPrefix) != 1)
        stop2("`missingPrefix` must have length 1: ", missingPrefix)
      
      # Default case
      newmiss = paste0(missingPrefix, 1:nmiss)
    }    
    
    # Special format
    if(!is.null(missingFormat)) {
      
      if(length(missingFormat) != 1)
        stop2("`missingFormat` must have length 1: ", missingFormat)
      
      mis = seq_len(nmiss)
      fam = getComponent(am, missing)
      idx = integer(length(missing))
      for(i in unique.default(fam))
        idx[fam == i] = seq_len(sum(fam == i))
      
      fmt = sub("[FAM]", "%{fam}d", fixed = TRUE,
                sub("[IDX]", "%{idx}d", fixed = TRUE,
                    sub("[MIS]", "%{mis}d", missingFormat, fixed = TRUE)))
      newmiss = sprintfNamed(fmt, fam = fam, idx = idx, mis = mis)
      
      if(length(newmiss) != nmiss)
        newmiss = rep_len(newmiss, length.out = nmiss)
    } 
    
    # Relabel
    am = relabel(am, old = missing, new = newmiss)
    
    # Update missing vector
    missing = as.character(newmiss)
  }
  
  # Other untyped individuals
  if(!is.null(othersPrefix)) {
    
    if(length(othersPrefix) != 1)
      stop2("`othersPrefix` must have length 1: ", othersPrefix)
  
    k = 0
    for(i in 1:nam) {
      oth = setdiff(untypedMembers(am[[i]]), missing)
      if(length(oth)) {
        am[[i]] = relabel(am[[i]], old = oth, new = paste0(othersPrefix, k + seq_along(oth)))
        k = k + length(oth)
      }
    }
  }
  
  # Return new object
  dviData(pm = pm, am = am, missing = missing)
}

#' Sequential DVI search
#'
#' Performs a sequential matching procedure based on the pairwise LR matrix. In
#' each step the pairing corresponding to the highest LR is selected and
#' included as a match if the LR exceeds the given threshold. By default,
#' (`updateLR = TRUE`) the pairwise LRs are recomputed in each step after
#' including the data from the identified sample.
#'
#' If, at any point, the highest LR is obtained by more than one pairing, the
#' process branches off and produces multiple solutions. (See Value.)
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param updateLR A logical. If TRUE, the LR matrix is updated in each
#'   iteration.
#' @param threshold A non-negative number. If no pairwise LR values exceed this,
#'   the iteration stops.
#' @param check A logical, indicating if the input data should be checked for
#'   consistency.
#' @param verbose A logical, by default TRUE.
#' @param debug A logical, by default FALSE. If TRUE, the LR matrix is printed
#'   in each step.
#'
#' @return A list with two elements:
#'  * `matches`: A single assignment vector, or (if multiple branches)
#'   a data frame where each row is an assignment vector.
#'  * `details`: A data frame (or a list of data frames, if multiple branches)
#'   including the LR of each identification in the order they were made.
#'
#' @examples
#' # Without LR updates
#' sequentialDVI(example1, updateLR = FALSE)
#'
#' # With LR updates (default). Note two branches!
#' r = sequentialDVI(example1)
#'
#' # Plot the two solutions
#' plotSolution(example1, r$matches, k = 1)
#' plotSolution(example1, r$matches, k = 2)
#'
#' # Add `debug = T` to see the LR matrix in each step
#' sequentialDVI(example1, debug = TRUE)
#'
#' # The output of can be fed into `jointDVI()`:
#' jointDVI(example1, assignments = r$matches)
#'
#' @importFrom stats setNames
#' @export
sequentialDVI = function(dvi, updateLR = TRUE, threshold = 1, check = TRUE, 
                         verbose = TRUE, debug = FALSE) {
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  vics = names(dvi$pm)
  nVics = length(vics)
  
  # AM components (for use in output)
  comp = getFamily(dvi, dvi$missing)
  
  if(verbose) {
    print.dviData(dvi)
    cat(sprintf("\nSequential search %s LR updates\n", ifelse(updateLR, "with", "without")))
  }
  
  # Initialise 'null' solution
  matches = list() 
  
  # LR matrix
  B = pairwiseLR(dvi, check = check)$LRmatrix
  
  # Environment for keeping parameters and storing solutions
  env = list2env(list(RES = list(), updateLR = updateLR, threshold = threshold, 
                      nVics = nVics, verbose = verbose, debug = debug))
  
  # Start recursion
  addPairing(dvi, B, matches, env)
  
  # Collect result table(s) with LR
  restabs = lapply(env$RES, function(lst) {
    df = do.call(rbind.data.frame, lst)
    if(nrow(df)) {
      df = cbind(df, Sample = names(lst), Family = comp[df$Missing])
      df = df[order(df$Step), c("Sample", "Missing", "Family", "LR", "Step")]
      rownames(df) = NULL
    }
    df
  })
  
  # Extract assignments
  A = setNames(rep("*", nVics), vics)
  mList = lapply(restabs, function(tab) {a = A; a[tab$Sample] = tab$Missing; a})
  matches = as.data.frame(do.call(rbind, unique(mList)))
  
  res = list(matches = matches, details = restabs)
  
  # Simplify output if no branching
  if(length(restabs) == 1)
    res = list(matches = matches[1,], details = restabs[[1]])
  
  res
}

# Recursive function, adding one new identification to the `matches` list
# If final: Store in the RES list of the environment `env`
addPairing = function(dvi, B, matches, env) {
  step = env$nVics - nrow(B)
  Bmax = .safeMax(B)
  if(is.na(Bmax))
    stop2("addPairing: all entries NA")
  
  # If all below threshold: store current solution and stop
  if(Bmax < env$threshold - sqrt(.Machine$double.eps)) {
    env$RES = append(env$RES, list(matches))
    if(env$verbose) 
      cat(sprintf("%sStep %d: Stop (max LR = %.2f)\n", strrep(" ", step), step + 1, max(B)))
    if(env$debug) {print(B); cat("\n")}
    return()
  }
  
  vics = rownames(B)
  missing = colnames(B)
  
  # Indices of maximal elements
  allmax = which(B == Bmax, arr.ind = TRUE)
  
  # For each max, set as solution and recurse
  # If multiple max'es, this creates new branches
  for(i in 1:nrow(allmax)) {
    mx = allmax[i, ]
    vic = vics[mx[1]]
    mp = missing[mx[2]]
    matches[[vic]] = list(Missing = mp, LR = Bmax, Step = step+1)
    if(env$verbose) {
      cat(sprintf("%sStep %d: %s = %s (LR = %.2g)\n", 
                      strrep(" ", step), step + 1, vic, mp, Bmax))
      if(env$debug) {print(B); cat("\n")}
    }
    
    # If all victims identified: store solution and return
    if(min(dim(B)) == 1) {
      env$RES = append(env$RES, list(matches))
      return()
    }
    
    ### Update DVI: Transfer vic data from old PM to new AM
    pm = dvi$pm
    am = dvi$am
    newPm = pm[-mx[1]]
    newAm = transferMarkers(from = pm, to = am, idsFrom = vic, idsTo = mp, erase = FALSE)
    newMissing = setdiff(missing, mp)
    newDvi = dviData(newPm, newAm, newMissing)
    
    if(env$updateLR)
      newB = pairwiseLR(newDvi, check = FALSE)$LRmatrix
    else
      newB = B[-mx[1], -mx[2], drop = FALSE]
    
    # Recurse
    addPairing(newDvi, newB, matches, env)
  }
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import pedtools
#' @importFrom pbapply pblapply pboptions
#' @importFrom parallel makeCluster stopCluster detectCores parLapply clusterEvalQ clusterExport
## usethis namespace: end
NULL

#' AM-driven DVI
#'
#' AM-driven identification, i.e., considering one AM family at a time. Simple
#' families (exactly 1 missing) are handled directly from the LR matrix, while
#' nonsimple families are analysed with [dviJoint()].
#'
#' Note: This function assumes that undisputed identifications have been
#' removed. Strange outputs may occur otherwise.
#'
#' @param dvi A `dviData` object.
#' @param fams A character; the names of families to consider. By default, all
#'   families. Special keywords: "simple" (all families with exactly 1 missing)
#'   and "nonsimple" (all families with > 1 missing).
#' @param threshold LR threshold for 'certain' match.
#' @param threshold2 LR threshold for 'probable' match (in *simple* families).
#' @param verbose A logical.
#'
#' @return A list of `dviReduced` and `summary`.
#'
#' @examples
#' w = amDrivenDVI(example2)
#' w$summary
#' w$dviReduced
#'
#' # Bigger example: Undisputed first
#' u = findUndisputed(planecrash)
#' u$summary
#'
#' # AM-driven analysis of the remaining
#' amDrivenDVI(u$dviReduced, threshold2 = 500)
#'
#' @export
amDrivenDVI = function(dvi, fams = NULL, threshold = 1e4, threshold2 = max(1, threshold/10), 
                       verbose = TRUE) {
  dvi = consolidateDVI(dvi)
  
  #if(verbose)
  #  cat("AM-driven analysis\n")
  
  if(!length(dvi$pm) || !length(dvi$missing))
    return(list(dviReduced = dvi, summary = NULL))
  
  if(is.null(fams))
    famnames = names(dvi$am)
  else if(identical(fams, "simple"))
    famnames = getSimpleFams(dvi)
  else if(identical(fams, "nonsimple"))
    famnames = setdiff(names(dvi$am), getSimpleFams(dvi))
  else
    famnames = fams

  # Number of missing in each fam
  comp = getFamily(dvi, ids = dvi$missing)
  nMiss = sapply(famnames, function(fam) sum(comp == fam))
  names(nMiss) = famnames
  
  # Compute LR matrix once and for all
  LRmat = pairwiseLR(dvi, check = FALSE, verbose = FALSE)$LRmatrix
  
  # List of summaries (to be combined below)
  summariesAM = summariesPM = NULL
  
  for(fam in famnames) {
    
    dvi1 = subsetDVI(dvi, am = fam, verbose = FALSE)
    if(verbose)
      cat(sprintf("Family %s: %d missing (%s)", fam, length(dvi1$missing), 
                  toString(dvi1$missing)))
    
    if(nMiss[fam] == 1) {
      s0 = .simpleFamDVI(dvi1, threshold = threshold2, LRmatrix = LRmat)
      s = list(AM = s0, PM = s0)
    }
    else {
      s = .complexFamDVI(dvi1, threshold = threshold, LRmatrix = LRmat, verbose = FALSE)
    }
   
    summariesAM = c(summariesAM, list(s$AM))
    summariesPM = c(summariesPM, list(s$PM))
    if(verbose)
      cat(" -->", if(is.null(s$AM)) "Inconclusive" else toString(unique(s$AM$Conclusion)), "\n")
  }
  
  if(verbose) cat("\nAM-driven results:\n")
  
  summaryAM = formatSummary(summariesAM, "AM")
  summaryPM = formatSummary(summariesPM, "PM")
  
  if(is.null(summaryAM)) {
    if(verbose)
      cat("No reduction of the dataset\n")
    return(list(dviReduced = dvi, summary = NULL))
  }
  
  # Reduce DVI
  remainMissing = setdiff(dvi$missing, summaryAM$Missing)
  remainVics = setdiff(names(dvi$pm), summaryPM$Sample)
  if(length(remainMissing))
    dviRed = subsetDVI(dvi, pm = remainVics, missing = remainMissing, verbose = FALSE)
  else
    dviRed = NULL   #TODO: Allow dviData with no AM data
  
  if(verbose)  {
    nam = length(dviRed$am)
    cat("Reduced dataset:", if(nam==1) "1 family remaining\n" else paste(nam, "families remaining\n"))
  }
  
  # Handle victims with no match
  if(length(remainVics)) {
    summList = lapply(remainVics, function(id) {
      maxLR = .safeMax(LRmat[id, ])
      if(is.na(maxLR)) {
        conclusion = "No match"
        comment = "No remaining pairings"
      }
      else {
        bestMatch = colnames(LRmat)[which.max(LRmat[id, ])]
        conclusion = ifelse(maxLR > threshold, "Disputed", "No match")
        comment = sprintf("Best: %s (LR=%.3g)", bestMatch, maxLR)
      }
      data.frame(Sample = id,  Conclusion = conclusion, Comment = comment, row.names = NULL)
    })
    summ = do.call(rbind, summList)
    summaryPM = formatSummary(list(summaryPM, summ), "PM")
  }
  
  list(dviReduced = dviRed, summary = list(AM = summaryAM, PM = summaryPM))
}


.simpleFamDVI = function(dvi1, threshold, LRmatrix = NULL) {
  
  miss = dvi1$missing
  if(length(miss) != 1)
    stop2("`.simpleFamDVI()` expects a dataset with a single missing person: ", miss)
  
  am = dvi1$am
  if(length(am) != 1 || is.null(names(am))) {
    print(dvi1)
    stop2("`.simpleFamDVI()` expects a `dviData` object with a single, named, family")
  }
  
  if(is.null(LRmatrix))
    LRmatrix = pairwiseLR(dvi1, check = FALSE, verbose = FALSE)$LRmatrix
  
  # Vector of LRs for the missing person
  lrs = LRmatrix[, miss]
  names(lrs) = rownames(LRmatrix)
  
  # Max LR and the corresponding victim (first of, if several)
  maxLR = .safeMax(lrs)
  if(is.na(maxLR))
    stop2("Error in `.simpleFamDVI()`: No LRs for missing person ", miss)
  bestMatch = names(lrs)[which.max(lrs)]
  
  # All LRs exceeding threshold
  high = lrs[!is.na(lrs) & lrs > threshold]
  
  # Runners-up
  rups = lrs[!is.na(lrs) & names(lrs) != bestMatch]
  
  if(length(high) > 1) {
    concl = "Disputed"
    also = high[names(high) != bestMatch]
    comment = paste("Also:", paste(sprintf("%s (LR=%.2g)", names(also), also), collapse = ", "))
  }
  else if(length(high) == 1) {
    concl = "Probable"
    if(length(rups) > 0) 
      ruptxt = sprintf("%s (LR=%.2g)", names(rups)[which.max(rups)], max(rups))
    else 
      ruptxt = "-"
    comment = paste("Runner-up:", ruptxt)
  }
  else if(maxLR >= 1) {
    concl =  "Inconclusive"
    comment = sprintf("Best: %s", bestMatch)
  }
  else {
    concl = "No match"
    comment = NA
  }
  
  # Summary (AM only)
  data.frame(Family = names(dvi1$am), Missing = miss, Sample = bestMatch, 
             LR = maxLR, Conclusion = concl, Comment = comment, row.names = NULL)
  
}


.complexFamDVI = function(dvi1, threshold = 1e4, LRmatrix = NULL, verbose = FALSE, progress = verbose) {
  
  origDvi = dvi1
  summAM = summPM = list()
  
  # Joint table
  j = dviJoint(dvi1, verbose = verbose, progress = progress)
  
  # Pairwise matrix (usually precomputed)
  if(is.null(LRmatrix))
    LRmatrix = pairwiseLR(dvi1, check = FALSE, verbose = FALSE)$LRmatrix
  
  # Pairwise GLR
  pairGLR = pairwiseGLR(dvi1, jointTable = j, LRmatrix = LRmatrix, threshold = threshold)

  # If any significant GLRs, add to summaries and reduce  
  if(!is.null(s <- pairGLR$summary)) {
    if(verbose) {cat("Pairwise GLR:\n)"); print(s)}
    summAM = c(summAM, list(s))
    summPM = c(summPM, list(s))
    dvi1 = pairGLR$dviReduced
    j = j[, !names(j) %in% c(s$Sample, s$Missing), drop = FALSE]
  } 
 
  # Symmetric GLR
  symGLR = symmetricGLR(dvi1, jointTable = j, threshold = threshold, verbose = verbose)
  if(!is.null(s <- symGLR)) {
    summAM = c(summAM, list(s$AM))
    summPM = c(summPM, list(s$PM))
  }

  summaryAM = formatSummary(summAM, "AM", dvi = origDvi)
  summaryPM = formatSummary(summPM, "PM")  # don't fill

  # Inconclusive: add best GLR
  inconc = is.na(summaryAM$GLR)
  if(any(inconc)) {
    GLRmatrix = pairGLR$GLRmatrix
    summaryAM$GLR[inconc] = apply(GLRmatrix[, inconc, drop = FALSE], 2, .safeMax)
    summaryAM$Conclusion[inconc] = "Inconclusive"
  }

  # Also add best LR for each missing
  summaryAM$LR = apply(LRmatrix[, origDvi$missing, drop = FALSE], 2, .safeMax)
  
  # Return summaries for this family
  list(AM = summaryAM, PM = summaryPM)
}

#' @importFrom verbalisr verbalise
symmetricGLR = function(dvi, jointTable = NULL, threshold = 1e4, verbose = FALSE) {
  fam = names(dvi$am)
  vics = names(dvi$pm)
  missing = dvi$missing
  j = jointTable %||% dviJoint(dvi, verbose = verbose)
  if(nrow(j) < 3)
    return(NULL)
  
  # Diffs between 1st and 2nd rows
  r1 = j[1, vics]; r2 = j[2, vics]
  diffs = r1 != r2
  if(sum(diffs) > 2 || any(r1[diffs] == "*", r2[diffs] == "*"))
    return(NULL)
  
  # Summaries
  summaryAM = summaryPM = NULL
  
  # Case 1: Single vic matching 2 missing
  # TODO: Doesn't handle vic matching 3 or more missing
  if(sum(diffs) == 1) {
    vic = vics[diffs]
    miss = .myintersect(missing, j[1:2, vic]) # intersect: for sorting
    endIdx = match(FALSE, j[[vic]] %in% miss) # 1st row not matching top 2
    GLR = exp(j$loglik[1] - j$loglik[endIdx])
    if(is.na(GLR) || GLR < threshold)
      return(NULL)

    rel = verbalisr::verbalise(dvi$am, miss) |> 
      format(cap = FALSE, simplify = TRUE, collapse = " + ")
    summaryAM = data.frame(Family = fam, Missing = miss, Sample = vic, GLR = GLR, 
                        Conclusion = "Symmetric match", 
                        Comment = sprintf("%s also matches %s (%s)", vic, rev(miss), rel),  
                        row.names = NULL)
    
    summaryPM = data.frame(Sample = vic, Family = fam, Missing = paste(miss, collapse = "/"),
                        GLR = GLR, Conclusion = "Symmetric match", 
                        Comment = sprintf("%s and %s are %s", miss[1], miss[2], rel), 
                        row.names = NULL)
  }
  else if(sum(diffs) == 2 && setequal(r1[diffs], r2[diffs])) {  # Symmetric pair
    vic = vics[diffs]
    miss = .myintersect(missing, r1[diffs]) # intersect: for sorting
    endIdx = match(FALSE, j[[vic[1]]] %in% miss & j[[vic[2]]] %in% miss) # 1st row not matching
    GLR = exp(j$loglik[1] - j$loglik[endIdx])
    if(is.na(GLR) || GLR < threshold)
      return(NULL)

    rel = verbalisr::verbalise(dvi$am, miss) |> 
      format(cap = TRUE, simplify = TRUE, collapse = " + ")
    
    summaryAM = data.frame(Family = fam, Missing = miss, 
                        Sample = paste(vic, collapse = "/"), GLR = GLR, 
                        Conclusion = "Symmetric match", 
                        Comment = sprintf("%s: {%s} = {%s}", rel, toString(miss), toString(vic)),
                        row.names = NULL)
    
    summaryPM = data.frame(Sample = vic, Family = fam, Missing = paste(miss, collapse = "/"),
                        GLR = GLR, Conclusion = "Symmetric match", 
                        Comment = sprintf("%s: {%s} = {%s}", rel, toString(vic), toString(miss)),
                        row.names = NULL)
  }
  
  # Return summaries
  list(AM = summaryAM, PM = summaryPM)
}


.pmDrivenDVI = function(dvi, threshold2, LRmatrix = NULL, origdvi = dvi) {
  
  vics = names(dvi$pm)
  nv = length(vics)
  
  if(is.null(LRmatrix))
    LRmatrix = pairwiseLR(dvi, check = FALSE, verbose = FALSE)$LRmatrix
  
  fam = character(nv)
  miss = character(nv)
  LR = numeric(nv)
  concl = character(nv)
  comment = character(nv)
  
  for(i in seq_len(nv)) {
    v = vics[i]
    lrs = LRmatrix[v, ]
    names(lrs) = colnames(LRmatrix)

    # Max LR and the corresponding missing (first of, if several)
    maxLR = .safeMax(lrs)
    if(is.na(maxLR))
      stop2("Something is wrong: No LRs for victim ", v)
    bestMatch = names(lrs)[which.max(lrs)]
    
    # All LRs exceeding threshold
    high = lrs[!is.na(lrs) & lrs >= threshold2 - .Machine$double.eps]
  
    # Runners-up
    rups = lrs[!is.na(lrs) & names(lrs) != bestMatch]
    
    LR[i] = maxLR
    
    if(length(high) > 1) {
      fam[i] = getFamily(origdvi, bestMatch)
      miss[i] = bestMatch
      concl[i] = "Disputed"
      also = high[names(high) != bestMatch]
      comment[i] = paste("Also:", paste(sprintf("%s (LR=%.2g)", names(also), also), collapse = ", "))
    }
    else if(length(high) == 1) {
      fam[i] = getFamily(origdvi, bestMatch)
      miss[i] = bestMatch
      concl[i] = "Probable"
      if(length(rups) > 0) 
        ruptxt = sprintf("%s (LR=%.2g)", names(rups)[which.max(rups)], max(rups))
      else 
        ruptxt = "-"
      comment[i] = paste("Runner-up:", ruptxt)
    }
    else if(maxLR >= 1) {
      concl[i] =  "Inconclusive"
      comment[i] = sprintf("Best: %s", bestMatch)
    }
    else {
      concl[i] = "No match"
      comment[i] = NA
    }
  }
  
  # Summary
  data.frame(Sample = vics, Family = fam, Missing = miss, LR = LR, 
             Conclusion = concl, Comment = comment, row.names = NULL)
}

#' Set identifications manually
#'
#' Manually set one or several identifications in a DVI dataset. Typically,
#' these are obtained by external means, e.g., fingerprints, dental records etc.
#'
#' The command `setPairing(dvi, c("V" = "M"))` does the following:
#'
#' * Transfer the data of victim "V" to the individual "M" in the appropriate
#' reference family
#' * Remove "M" from the list of missing persons
#' * Remove "V" from the list of victim samples
#' * Update the list of pairings
#'
#' @param dvi A DVI dataset.
#' @param match A named vector of the format c(vic1 = miss2, vic2 = miss2, ...).
#' @param victim A vector of victim sample names. If NULL, defaulting to
#'   `names(match)`.
#' @param missing A vector of missing person names, of the same length as
#'   `victim`. If NULL, defaulting to `as.character(match)`.
#' @param Conclusion A character passed on to the `Conclusion` column of the
#'   output summary.
#' @param Comment A character passed on to the `Comment` column of the output
#'   summary.
#' @param verbose A logical, by default TRUE.
#'
#' @return A list with the following entries:
#'
#'   * `dviReduced`: The new `dviData` object, as described in Details
#'
#'   * `summary`: A data frame summarising the identifications
#'
#' @examples
#' x = setPairing(example2, match = c("V3" = "M2"))
#' x$dviReduced
#' x$summary
#'
#' # Alternative syntax, using `victim` and `missing`
#' y = setPairing(planecrash, victim = c("V4", "V5"), missing = c("M4", "M5"),
#'            Conclusion = "External evidence", Comment = "Dental")
#' y$dviReduced
#' y$summary
#' 
#' @export
setPairing = function(dvi, match = NULL, victim = NULL, missing = NULL, 
                      Conclusion = "Provided", Comment = "", verbose = TRUE) {
  dvi = consolidateDVI(dvi)
  
  if(!is.null(match)) {
    nms = names(match)
    if(is.null(nms))
      stop2("`match` must be a named vector: c(vic1 = miss1, ...)")
    if(!is.null(victim))
      stop2("When `match` is given, `victim` must be NULL, not: ", victim)
    if(!is.null(missing))
      stop2("When `match` is given, `missing` must be NULL, not: ", missing)
    
    victim = nms
    missing = as.character(match)
  }

  N = length(victim)
  if(N != length(missing))
    stop2(sprintf("Number of `victims` (%d) differs from the number of missing persons (%d)",
                  N, length(missing)))
  if(N == 0) {
    if(verbose) cat("No identifications indicated\n")
    return(list(dviReduced = dvi, summary = NULL))
  }
  
  if(length(badvic <- setdiff(victim, names(dvi$pm))))
    stop2("Unknown victim ID: ", badvic)
  if(length(badmis <- setdiff(missing, dvi$missing)))
    stop2("Unknown missing person ID: ", badmis)
  
  if(!length(Conclusion) %in% c(1, N))
    stop2(sprintf("`Conclusion` must have length either 1 or %d, not %d", N, length(Conclusion)))
  if(!length(Comment) %in% c(1, N))
    stop2(sprintf("`Comment` must have length either 1 or %d, not %d", N, length(Comment)))
  
  # AM component of each missing
  comp = getFamily(dvi, missing)
  
  if(verbose) {
    cat("Fixing the following identifications:\n")
    cat(sprintf(" %s = %s (Family %s)", victim, missing, comp), sep = "\n")
    cat("\n")
  }
  
  # Form reduced DVI Remove victims and missing
  dviRed = subsetDVI(dvi, pm = setdiff(names(dvi$pm), victim), 
                     missing = setdiff(dvi$missing, missing), verbose = verbose)
  
  # Transfer marker data
  dviRed$am = transferMarkers(from = dvi$pm[victim], to = dviRed$am, 
                              idsFrom = victim, idsTo = missing, erase = FALSE)
  
  # Build summary report
  summary = data.frame(Family = comp, Missing = missing, Sample = victim, 
                       Conclusion = Conclusion, Comment = Comment,
                       row.names = NULL)
  # Return
  list(dviReduced = dviRed,
       summary = summary)
}


#' Exclude pairings
#'
#' Disallow certain pairings by removing them from the list `dvi$pairings` of
#' candidate pairings for a given victim sample.
#' 
#' @param dvi A `dviData` object.
#' @param victim The name of a single victim sample.
#' @param missing The name(s) of one or several missing individuals.
#'
#' @return A `dviData` object.
#'
#' @examples
#' # Disallow V1 = M1 in the `example2` dataset:
#' ex = excludePairing(example2, victim = "V1", missing = "M1")
#' jointDVI(ex, verbose = FALSE)
#' 
#' # Compare with original
#' jointDVI(example2, verbose = FALSE)
#' 
#' # The only difference is in the `pairings` slot:
#' ex$pairings
#' example2$pairings
#' 
#' @export
excludePairing = function(dvi, victim, missing) {
  if(is.null(dvi$pairings))
    stop2("`dvi` object has no pairings attached")
  if(length(victim) != 1)
    stop2("Pairings can be excluded for only 1 victim at a time: ", victim)
  
  dvi$pairings[[victim]] = setdiff(dvi$pairings[[victim]], missing)
  dvi
}

#' Convert a Familias file to DVI data
#'
#' This is a wrapper for [pedFamilias::readFam()] that reads Familias files with
#' DVI information.
#'
#' @inheritParams relabelDVI
#' @param famfile Path to Familias file.
#' @param verbose A logical, passed on to `readFam()`.
#' @param missingIdentifier A character of length 1 used to identify missing
#'   persons in the Familias file. The default chooses everyone whose label
#'   begins with "Missing".
#'
#' @return A `dviData` object.
#'
#' @details The sex of the missing persons need to be checked as this
#'   information may not be correctly recorded in the fam file.
#'
#' @seealso [dviData()], [relabelDVI()]
#'
#' @examples
#'
#' # Family with three missing
#' file = system.file("extdata", "dvi-example.fam", package="dvir")
#'
#' # Read file without relabelling
#' y = familias2dvir(file)
#' plotDVI(y)
#'
#' # With relabelling
#' z = familias2dvir(file, missingFormat = "M[FAM]-[IDX]",
#'                    refPrefix = "ref", othersPrefix = "E")
#' plotDVI(z)
#'
#' @export
familias2dvir = function(famfile, victimPrefix = NULL, familyPrefix = NULL,
                         refPrefix = NULL, missingPrefix = NULL, 
                         missingFormat = NULL, othersPrefix = NULL,
                         verbose = TRUE, missingIdentifier = "^Missing"){
  
  # Read fam file
  x = pedFamilias::readFam(famfile, useDVI = TRUE, verbose = verbose)

  # PM data -----------------------------------------------------------------
  
  pm = x$`Unidentified persons`
  if(is.null(pm))
    stop2("No `Unidentified persons` found")

  # AM data -----------------------------------------------------------------
  
  am = x[-1]
  
  # Remove untyped components
  am = lapply(am, function(ref) {
    if(is.ped(ref))
      return(ref)
    
    # Remove Reference pedigree if present
    idx = match("Reference pedigree", names(ref), nomatch = 0)
    if(length(ref) == 2 && idx > 0)
      ref = ref[[3 - idx]] # the other
    
    # Expect single component with typed references
    cmp = getComponent(ref, typedMembers(ref))
    if(max(cmp) > min(cmp)) 
      stop2("Disconnected reference family")
    
    ref[[cmp[1]]]
  })

  # Check for duplicated names among reference individuals
  if(!is.null(am)) {
    typed = typedMembers(am)
    dups = anyDuplicated.default(typed)
    if(dups) {
      if(verbose) cat(sprintf("Warning: Duplicated reference name '%s'\n", typed[dups]), 
                      "        Adding family prefix\n")
      for(nam in names(am)) {
        y = am[[nam]]
        old = intersect(typed, y$ID)
        am[[nam]] = relabel(y, old = old, new = paste(nam, old, sep = "_"))
      }
    }
  }
  
  # Missing individuals -----------------------------------------------------

  missing = grep(missingIdentifier, unlist(labels(am)), value = TRUE)
  
  if(!length(missing))
    stop2("Reference pedigree without missing")
  

  # Create DVI object -------------------------------------------------------

  dvi0 = dviData(pm = pm, am = am, missing = missing, generatePairings = FALSE)
  
  # Relabel missing persons (NB: pairings are generated here)
  relabelDVI(dvi0, victimPrefix = victimPrefix, familyPrefix = familyPrefix,
             refPrefix = refPrefix, missingPrefix = missingPrefix, 
             missingFormat = missingFormat, othersPrefix = othersPrefix)
}

pairwiseGLR = function(dvi, jointTable = NULL, LRmatrix = NULL, threshold = 1e4, verbose = FALSE) {
  
  dvi = consolidateDVI(dvi)
  pm = dvi$pm
  am = dvi$am
  vics = names(pm)
  missing = dvi$missing
  
  if(threshold <= 1) 
    stop2("Argument `threshold` must exceed 1: ", threshold)
  
  # Pairwise LR matrix (as additional info)
  LRmat = LRmatrix %||% pairwiseLR(dvi, verbose = FALSE)$LRmatrix
  
  # Joint table (should normally be supplied)
  joint = jointTable %||% dviJoint(dvi, verbose = verbose)
  
  # Initialise GLR matrix
  G = matrix(NA_real_, nrow = length(vics), ncol = length(missing), 
             dimnames = list(vics, missing))
  
  # Fill matrix row-wise
  for (v in vics) {
    # corresponding column of joint table
    a = joint[[v]]
    
    # If all equal (to '*' or some Mj): GLR not defined
    if(length(unique.default(a)) < 2)
      next
    
    # Index of 1st row with v = m, for each m in miss
    idxNumer = match(missing, a)
    
    # Index of 1st row with v != m
    idxDenom = ifelse(idxNumer == 1, which(a != a[1])[1], 1)
    
    # Insert in matrix
    ll = joint$loglik
    G[v,  ] = exp(ll[idxNumer] - ll[idxDenom])
  }
  
  # Indices of matches exceeding GLR threshold
  highGLR = which(G >= threshold - sqrt(.Machine$double.eps), arr.ind = TRUE)
  
  foundVics = vics[highGLR[, 1]]
  foundMiss = missing[highGLR[, 2]]
  
  # Summary
  summ = NULL
  
  if(length(foundVics)) {
    summ = data.frame(
      Family = getFamily(dvi, foundMiss),
      Missing = foundMiss, 
      Sample = foundVics, 
      LR = LRmat[highGLR],
      GLR = G[highGLR],
      Conclusion = "Match (GLR)",
      Comment = sprintf("Joint analysis {%s}", paste0(missing, collapse = ",")),
      row.names = NULL)
  }
  
  # Data from identified samples (keep for updating dviRed below)
  foundData = pm[foundVics]
   
  # Reduced DVI dataset
  newvics = setdiff(vics, foundVics)
  newmissing = setdiff(missing, foundMiss)
  dviRed = subsetDVI(dvi, pm = newvics, missing = newmissing, verbose = verbose)
  
  # Move vic data to AM data
  names(foundVics) = foundMiss
  relevantMP = intersect(foundMiss, labels(am))
  if(length(relevantMP))
    dvi$am = transferMarkers(from = foundData, to = am, 
                             idsFrom = foundVics[relevantMP], 
                             idsTo = relevantMP, erase = FALSE)
    
  list(dviReduced = dviRed, GLRmatrix = G, summary = summ)
}

#' Combinations without duplications
#'
#' This is similar to [expand.grid()] except that combinations with repeated
#' elements are not included. The element "*" is treated separately, and is
#' allowed to be repeated.
#'
#' @param lst A list of vectors.
#' @param max A positive integer. If the number of combinations (according to a
#'   preliminary lower bound) exceeds this, the function aborts with an
#'   informative error message. Default: 1e5.
#'
#' @return A data frame.
#' @seealso [expand.grid()]
#'
#' @examples
#'
#' lst = list(1, 1:2, 3:4)
#'
#' # Compare
#' expand.grid.nodup(lst)
#' expand.grid(lst)
#'
#' # Typical use case for DVI
#' lst2 = generatePairings(example1)
#' expand.grid.nodup(lst2)
#'
#' @export
expand.grid.nodup = function(lst, max = 1e5) {
  if(is.data.frame(lst))
    stop2("Unexpected input: Argument `lst` is a data frame")
  if(!is.list(lst))
    stop2("Argument `lst` should be a list")
  
  len = length(lst)  
  if(len == 0)
    return(data.frame())
  
  nms = names(lst)
  if(is.null(nms)) 
    nms = paste0("Var", 1:len)
  
  # If empty: Return early
  if(any(lengths(lst) == 0))
    return(data.frame(matrix(nrow = 0, ncol = len, dimnames = list(NULL, nms))))
  
  
  # Warn if too many
  b = estimateGridSize(lst)
  if(b > max) {
    msg1 = sprintf("\nNumber of assignments is at least %g, exceeding the current limit of %d.\n", b, max)
    msg2 = "Possible strategies:\n * Increase `limit`\n * Decrease `threshold`\n * Increase `maxAssign`\n * Set `strict` to FALSE"
    stop2(paste0(msg1, msg2))
  }
  
  # Recursive function
  recurse = function(a) {
    if(length(a) == 1)
      return(as.list.default(a[[1]]))
    
    r = lapply(a[[1]], function(val) {
      b = a[-1]
      if(val != "*")
        b = lapply(b, function(vec) vec[vec != val])
      lapply(recurse(b), function(vec) c(val, vec))
    })
    
    unlist(r, recursive = FALSE)
  }
  
  # Call it!
  reslist = recurse(lst)
  if(!length(reslist))
    stop2("No possible solutions (empty grid)")
  
  # Convert to data frame and add names
  resmat = matrix(unlist(reslist, recursive = FALSE, use.names = FALSE), 
                  byrow = TRUE, ncol = len)
  res = as.data.frame.matrix(resmat)
  names(res) = nms
  
  res
}


# Compute a lower bound on the number of assignments for a given list of pairings
# Exact if entries of lst are all contained in each other
estimateGridSize = function(lst) {
  
  n = length(lst)
  lens = sort.int(lengths(lst))
  
  if(n == 0 || any(lens == 0))
    return(0)
  
  allstars = all(vapply(lst, function(v) length(v) > 0 && "*" %in% v, TRUE))
  
  if(!allstars)
    # Assume no stars. Trivial formula (Theorem 2 from paper 2022)
    return(prod(lens - 0:(n-1)))
  
  # From here: All sets contain "*"
  C = lens - 1
  a = matrix(0, ncol = n + 1, nrow = n)
  
  # first column: a_n,0 (no stars)
  a[, 1] = cumprod(C - 0:(n-1))
  
  # Diagonal: k stars
  for(k in 1:n) 
    a[k, k+1] = 1
  
  # Remaining entries: Recurse!
  for(k in seq_len(n)) for(s in seq_len(k-1))
    a[k,s+1] = a[k-1, s] + a[k-1, s+1] * (C[k] - (k-1-s))
  
  sum(a[n, ])
}

#' A complete workflow for solving a DVI case
#'
#' Wraps the main`dvir` workflow for DVI analysis and identification of likely AM--PM matches.
#'
#' The function roughly implements the following workflow:
#'
#' - **Preparation and QC**
#'   - Consolidate data and validate input: [checkDVI()]
#'   - Generate pairings: [generatePairings()]
#'   
#' - **Initial reduction**
#'   - Detect and remove nonidentifiable missing persons: [findNonidentifiable()]
#'   
#' - **Iterative reduction** (repeat until stable)
#'   - Excluded individuals and pairings: [findExcluded()]
#'   - Undisputed matches from pairwise LRs: [findUndisputed()]
#'   
#' - **AM-driven analysis**: [amDrivenDVI()]
#'   - Simple families (1 missing): Summarise conflicting results
#'   - Complex families (>1 missing): Joint analysis + GLR
#'
#' - **PM-driven analysis**
#'   - Summarise remaining victim samples.
#'   
#' - **Output**
#'   - Format AM and PM summaries: [formatSummary()]
#'
#' @param dvi A `dviData` object.
#' @param threshold LR threshold for 'significant' match.
#' @param threshold2 LR threshold for 'probable' match. By default set to
#'   `threshold/10`.
#' @param maxIncomp	An integer passed onto [findExcluded()]. A pairing is
#'   excluded if the number of incompatible markers exceeds this.
#' @param ignoreSex A logical, by default FALSE.
#' @param limit	A number passed onto [findUndisputed()]; only pairwise LR values
#'   above this are considered.
#' @param detailedOutput A logical, by default FALSE. See Details.
#' @param verbose,debug Logicals.
#'
#' @return A list of data frames `AM` and `PM`. 
#' 
#' If `detailedOutput = TRUE`, the result also includes the exclusion matrix 
#' from the first call to `findExcluded()` and the LR matrix from the first 
#' call to `findUndisputed()`.
#' 
#' @examples
#' dviSolve(example2)
#' dviSolve(example2, threshold = 5, detailedOutput = TRUE, verbose = FALSE)
#'
#' @export
dviSolve = function(dvi, threshold = 1e4, threshold2 = max(1, threshold/10), 
                    maxIncomp = 2, ignoreSex = FALSE, limit = 0, 
                    detailedOutput = FALSE, verbose = TRUE, debug = FALSE) {

  if(ignoreSex)
    dvi$pairings = generatePairings(dvi, ignoreSex = TRUE)
  
  # Check dataset -----------------------------------------------------------
  if(verbose)
    cat("Checking dataset" |> dashpad())
  
  origdvi = dvi = consolidateDVI(dvi)
  checkDVI(dvi, verbose = debug, ignoreSex = ignoreSex)
  if(verbose)
    cat("Ok\n")
  
  summariesAM = summariesPM = list()
  
  # Nonidentifiable ---------------------------------------------------------

  if(verbose)
    cat("Nonidentifiable missing persons" |> dashpad())
  
  non = findNonidentifiable(dvi)
  dvi = non$dviReduced
  summariesAM = c(summariesAM, list(non$summary))
  if(verbose)
    printSummary(non$summary, nulltext = "None")
  
  iter = 0

  # Loop until no change: excl, undisp, excl, ...
  while(TRUE) {
    
    iter = iter + 1
    
    # Exclusions --------------------------------------------------------------
    
    if(verbose)
      cat("Exclusions, iteration" |> paste(iter) |> dashpad())
    
    excl = findExcluded(dvi, maxIncomp = maxIncomp, verbose = debug)

    # Store first exclusion matrix
    if(iter == 1)
      EXCLmat1 = excl$exclusionMatrix
    
    # Break loop if no change - but not in first iter!
    if(dviEqual(excl$dviReduced, dvi)) {
      if(verbose) cat("No exclusions", if(iter > 1) "; breaking loop", "\n", sep = "")
      if(iter > 1)
        break
    }
    else {
      s = excl$summary
      summariesAM = c(summariesAM, list(s$AM))
      summariesPM = c(summariesPM, list(s$PM))
      if(verbose) {
        removedPrs = removedPairings(excl$dviReduced, dvi)
        printSummary(s, addNewline = removedPrs > 0)
        
        if(removedPrs)
          cat(sprintf("Excluded %s\n", pluralise("individual pairing", removedPrs)))
      }
      dvi = excl$dviReduced
    }
    
    # Undisputed --------------------------------------------------------------
    
    if(verbose)
      cat("Undisputed, iteration" |> paste(iter) |> dashpad())
    
    und = findUndisputed(dvi, threshold = threshold, limit = limit, 
                         keepLRmatrs = (iter == 1), verbose = debug)

    # Store first LR matrix
    if(iter == 1) {
      LRmat1 = und$LRmatrix[[1]]
      LRmat = und$LRmatrix[[length(und$LRmatrix)]]
    }
    else 
      LRmat = und$LRmatrix
    
    if(dviEqual(und$dviReduced, dvi)) {
      if(verbose) cat("No change; breaking loop\n")
      break
    }
    else {
      s = und$summary
      summariesAM = c(summariesAM, list(s))
      summariesPM = c(summariesPM, list(s))
      if(verbose) {
        removedPrs = removedPairings(und$dviReduced, dvi)
        printSummary(s, addNewline = removedPrs > 0)
        if(removedPrs) {
          a = pluralise("pairing", removedPrs)
          if(limit == 0) 
            cat(sprintf("Removed %s with LR = 0\n", a))
          else 
            cat(sprintf("Removed %s with LR < %g (`limit`)\n", a, limit))
        }
      }
      dvi = und$dviReduced
    }
  }

  # AM-driven: Simple-----------------------------------------------------------
  
  nam = length(dvi$am)
  nMiss = nMissFam(dvi)
  simpleFams = names(nMiss[nMiss == 1])
  nsimp = length(simpleFams)
  
  if(verbose) {
    cat("AM-driven analysis: Simple families" |> dashpad())
    if(nsimp == 0)
      cat("0 simple families\n")
    else
      cat(sprintf("%d simple famil%s: %s\n\n", nsimp, ifelse(nsimp == 1, "y", "ies"), toString(simpleFams)))
  }
  
  for(fam in simpleFams) {
    dvi1 = subsetDVI(dvi, am = fam, verbose = FALSE)
    if(verbose)
      cat(sprintf("Family %s: %s", fam, dvi1$missing))
    s = .simpleFamDVI(dvi1, threshold = threshold2, LRmatrix = LRmat)
    summariesAM = c(summariesAM, list(s))
    if(verbose)
      cat(" -->", s$Conclusion, "\n")
  }
  
  # AM-driven: Complex-----------------------------------------------------------
  
  complexFams = names(nMiss[nMiss > 1])
  ncomp = length(complexFams)
  
  if(verbose) {
    cat("AM-driven analysis: Complex families" |> dashpad())
    if(ncomp == 0)
      cat("0 complex families\n")
    else
      cat(sprintf("%d complex famil%s: %s\n\n", ncomp, 
                  ifelse(ncomp == 1, "y", "ies"), toString(complexFams)))
  }

  jointTabs = list()
  
  # Loop through complex families. Store joints and remove identified victims in each iteration
  for(fam in complexFams) {
    dvi1 = subsetDVI(dvi, am = fam, verbose = FALSE)
    
    s = .complexFamDVI(dvi1, threshold = threshold, LRmatrix = LRmat, verbose = FALSE)
    summariesAM = c(summariesAM, list(s$AM))
    summariesPM = c(summariesPM, list(s$PM))
    jointTabs[[fam]] = s$joint
    
    # Reduce main DVI dataset
    dvi = subsetDVI(dvi, 
                    am = .mysetdiff(names(dvi$am), fam),
                    pm = .mysetdiff(names(dvi$pm), s$PM$Sample), 
                    removeUnpairedPM = FALSE, verbose = FALSE)
    if(verbose) {
      print(utils::head(s$joint, 5))
      cat(sprintf("<%d rows total>\n\n", nrow(s$joint)))
      
      concs = s$AM$Conclusion
      if(all(concs == concs[1])) concs = concs[1]
      cat(sprintf("Conclusions: %s --> %s\n", toString(dvi1$missing), toString(concs)))
    }
  }
  
  # PM driven analysis----------------------------------------------------------
  
  if(verbose)
    cat("Remaining victim samples" |> dashpad())
  
  vics = names(dvi$pm)
  nv = length(vics)
  if(verbose) {
    if(nv == 0) cat("No remaining victims\n")
    else cat(sprintf("%d remaining victim%s: %s\n", nv, if(nv == 1) "" else "s", toString(vics)))
  }

  if(nv > 0) {
    s = .pmDrivenDVI(dvi, threshold2 = threshold2, LRmatrix = LRmat, origdvi = origdvi)
    summariesPM = c(summariesPM, list(s))
  }

  # Collect results
  resultAM = formatSummary(summariesAM, orientation = "AM", dvi = origdvi)
  resultPM = formatSummary(summariesPM, orientation = "PM", dvi = origdvi)

  res = list(AM = resultAM, PM = resultPM)
  if(detailedOutput) {
    res$LRmatrix = LRmat1
    res$exclusionMatrix = EXCLmat1
    res$jointTables = jointTabs
  }
  
  res
}



dashpad = function(x, width = 50) {
  y = paste0("\n", strrep("-", 6), " ", x, " ")
  paste0(y, strrep("-", max(0, width - nchar(y))), "\n\n")
}

printSummary = function(s, nulltext = NULL, addNewline = FALSE) {
  if(is.null(s) || (length(s) == 2 && is.null(s$AM) && is.null(s$PM))) {
    if(!is.null(nulltext))
      cat(nulltext, "\n")
    return()
  }
  if(is.data.frame(s)) {
    print(s)
    return(invisible())
  }
  
  if(!is.null(s$AM)) {
    cat("$AM\n"); print(s$AM)
  }
  if(!is.null(s$PM)) {
    cat("$PM\n"); print(s$PM)
  }
  if(addNewline)
    cat("\n")
}

#' Nonidentifiable missing persons
#'
#' A missing person in a DVI case is nonidentifiable if they are not related to
#' any (genotyped) reference individuals, neither directly nor through other
#' missing persons in the family. This function identifies such individuals, and
#' reduces the DVI dataset accordingly.
#'
#' Two individuals are unrelated if their kinship coefficient is zero. The
#' implementation uses `ribd::kinship()` to calculate all pairwise kinship
#' coefficients.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#'
#' @return A list with the following entries:
#'   * `nonidentifiable`: A character vector (possibly empty) with the names of
#'   the nonidentifiable missing persons.
#'   * `dviReduced`: A reduced `dviData` object, where the nonidentifiable
#'   individuals are removed from the list of missing persons. If there are no
#'   `nonidentifiable`, this is just a copy of `dvi`.
#'   * `summary`: A data frame summarising the findings.
#'
#' @examples
#' # Example 1: No nonidentifiables in dataset `example1`
#' findNonidentifiable(example1)
#'
#' # Example 2: Add nonidentifiable person "A"
#' amNew = example1$am[[1]] |>
#'   addSon(parents = c("NN", "A"))
#' missNew = c(example1$missing, "A")
#'
#' dvi = dviData(pm = example1$pm, am = amNew, missing = missNew)
#' plotDVI(dvi, textAbove = c(A = "nonidentif."))
#'
#' findNonidentifiable(dvi)
#'
#' @export
findNonidentifiable = function(dvi) {
  dvi = consolidateDVI(dvi)
  miss = dvi$missing
  refs = typedMembers(dvi$am)
  
  # Kinship matrix of refs and missing
  ids = c(typedMembers(dvi$am), dvi$missing)
  k = ribd::kinship(dvi$am, ids = ids, simplify = FALSE)
  
  # Columns with only zeroes = nonidentifiable
  nonident = .nonidentif(k, miss, refs)
  
  # If none, return unchanged
  if(!length(nonident))
    return(list(nonidentifiable = nonident,
                dviReduced = dvi,
                summary = NULL))
  
  # Reduce by removing nonidentifiable from "missing"
  dviRed = subsetDVI(dvi, missing = setdiff(miss, nonident), verbose = FALSE)
  
  # For summary: Family of nonidentifiable indivs
  fams = getFamily(dvi, ids = nonident)

  # Special comment if there are no refs in a family
  nRefs = sapply(dvi$am, function(x) length(typedMembers(x)))

  # Comment column
  cmt = ifelse(nRefs[fams] == 0, "No references in family", "Unrelated to all references")
  
  # Build summary report
  summary = data.frame(Family = fams,
                       Missing = nonident,
                       Conclusion = "Nonidentifiable",
                       Comment = cmt,
                       row.names = NULL)
                        
  list(nonidentifiable = nonident,
       dviReduced = dviRed,
       summary = summary)
}

.nonidentif = function(k, miss, refs) {
  # Candidates for being nonidentifiable
  cand = miss
  
  # Individuals linked to the refs
  linked = refs
  
  # Iteratively remove those who are related to the previous links
  while(length(cand) > 0 && length(linked) > 0) {
    linked = names(which(rowSums(k[cand, linked, drop = FALSE]) > 0))
    cand = setdiff(cand, linked)
  }
  
  cand
}

#' @importFrom pedprobr likelihood
loglikTotal = function(x, markers = seq_len(nMarkers(x))) {
  sum(likelihood(x, marker = markers, logbase = exp(1)))
}


# Modified version of stop()
stop2 = function(...) {
  a = lapply(list(...), toString)
  a = append(a, list(call. = FALSE))
  do.call(stop, a)
}


`%||%` = function(x, y) {
  if(is.null(x)) y else x
}

ftime = function(st, digits = 3) {
  format(Sys.time() - st, digits = digits)
}

# Fast intersection. NB: assumes no duplicates!
.myintersect = function (x, y) {
  y[match(x, y, 0L)]
}

.mysetdiff = function(x, y) {
  unique.default(x[match(x, y, 0L) == 0L])
}

pluralise = function(noun, n, numberFirst = TRUE) {
  s = if(n == 1) noun else sprintf("%ss", noun)
  if(numberFirst) 
    s = paste(n, s)
  s
}

.safeMax = function(v) {
  if(length(v) == 0 || all(is.na(v))) 
    return(NA)
  max(v, na.rm = TRUE)
}

trunc = function(x, printMax = 8, head = 3) {
  if(length(x) <= printMax)
    return(toString(x))
  maxHead = max(1, min(head, printMax - 2))
  y = c(head(x, maxHead), "...", x[length(x)])
  toString(y)
}

# example: sprintfNamed("I am %{name}s", name = "your father")
sprintfNamed = function(fmt, ...) {
  arglist = list(...)
  if(length(arglist) == 1 && is.list(arglist[[1]]))
    arglist = arglist[[1]]
  
  nms = names(arglist)
  if(is.null(nms) || any(nms == "")) 
    stop2("Arguments must have names")
  
  patterns = sprintf("%%{%s}", nms)
  hasPat = vapply(patterns, function(p) grepl(p, fmt, fixed = TRUE), FALSE)
  
  # Remove unused arguments
  arglist = arglist[hasPat]
  patterns = patterns[hasPat]

  for(i in seq_along(patterns)) {
    fmt = gsub(pattern = patterns[i],
               replacement = sprintf("%%%d$", i),
               fmt, fixed = TRUE)
  }

  miss = regmatches(fmt, m = gregexec("%\\{([^%]*)}", fmt, perl = TRUE))[[1]]
  if(length(miss))
    stop2("Missing values for variable: ", sprintf("'%s'", miss[2, ]))
  
  do.call(sprintf, append(arglist, fmt, 0))
}

# Undisputed entries in a LR/GLR matrix
undisputedEntries = function(M, threshold, strict = TRUE) {
  
  if(!length(M))
    return(matrix(nrow = 0, ncol = 2, dimnames = list(NULL, c("row", "col"))))
  
  # Replace NAs with -1 to allow numeric comparisons below
  M[is.na(M)] = -1
  
  # Indices of matches exceeding threshold
  highIdx = which(M >= threshold - sqrt(.Machine$double.eps), arr.ind = TRUE)
  
  # Return if empty
  if(!nrow(highIdx)) 
    return(highIdx)
  
  # Identify which rows of `highIdx` to keep
  if(strict) { # undisputed = no others in same row or column exceed 1
    goodRows = which(rowSums(M <= 1) == ncol(M) - 1)
    goodCols = which(colSums(M <= 1) == nrow(M) - 1)
    isUndisp = highIdx[, "row"] %in% goodRows & highIdx[, "col"] %in% goodCols
  }
  else { # undisputed = no others in same row or column exceed LR/threshold
    isUndisp = sapply(seq_len(nrow(highIdx)), function(k) {  # safer than apply(.., 1)!
      rw = highIdx[k,1]
      cl = highIdx[k,2]
      all(c(M[rw, -cl], M[-rw, cl]) <= M[rw, cl]/threshold)
    })
  }

  # Return matrix of indices of undisputed matches   
  highIdx[isUndisp, , drop = FALSE]
}

#' Extract a subset of a DVI dataset
#'
#' @param dvi A [dviData()] object
#' @param pm A vector with names or indices of victim samples. By default, all
#'   are included.
#' @param am A vector with names or indices of AM components. By default,
#'   components without remaining missing individuals are dropped.
#' @param missing A vector with names or indices of missing persons. By default,
#'   all missing persons in the remaining AM families are included.
#' @param removeUnpairedPM A logical, by default TRUE, removing PM samples with
#'   no remaining pairings.
#' @param verbose A logical.
#'
#' @return A `dviData` object.
#'
#' @examples
#'
#' subsetDVI(example2, pm = 1:2) |> plotDVI()
#' subsetDVI(example2, pm = "V1", am = 1) |> plotDVI()
#' subsetDVI(example2, missing = "M3") |> plotDVI()
#'
#' @export
subsetDVI = function(dvi, pm = NULL, am = NULL, missing = NULL, removeUnpairedPM = TRUE, verbose = TRUE) {
  
  if(verbose)
    cat("Reducing DVI dataset\n")
  
  dvi = consolidateDVI(dvi)
  
  pmNew = dvi$pm
  amNew = dvi$am
  missNew = dvi$missing
  
  if(!is.null(pm)) {
    pmNew = dvi$pm[pm]
    
    err = vapply(pmNew, is.null, FALSE)
    if(any(err))
      stop2("Unknown name/index of PM singleton: ", pm[err])
  }
  
  if(!is.null(am)) {
    amNew = dvi$am[am]
    
    err = vapply(amNew, is.null, FALSE)
    if(any(err))
      stop2("Unknown name/index of AM family: ", am[err])
    
    if(is.null(missing)) {
      
      if(!length(amNew))
        missNew = character(0)
      else {
        comp = getComponent(amNew, dvi$missing, checkUnique = FALSE, errorIfUnknown = FALSE)
        missNew = dvi$missing[!is.na(comp)]
      }
      missRem = .mysetdiff(dvi$missing, missNew)
      if(verbose && length(missRem))
        cat(sprintf("Removing %s in excluded families: %s\n", 
                    pluralise("missing person", length(missRem)), trunc(missRem, 6)))
    }
  }
  
  if(!is.null(missing)) {
    
    # Add names to allow subsetting by name or index
    miss0 = dvi$missing
    names(miss0) = miss0
    
    missNew = unname(miss0[missing])
    
    err = is.na(missNew)
    if(any(err))
      stop2("Unknown name/index of missing person: ", missing[err])
    
    # If AM subset not given by user, remove components without missing persons
    if(is.null(am)) {
      comps = getComponent(dvi$am, missNew, checkUnique = FALSE, errorIfUnknown = FALSE)
      if(anyNA(comps))
        stop2("Missing person not found in AM data: ", missNew[is.na(comps)])
      
      unused = setdiff(seq_along(dvi$am), comps)
      if(length(unused)) {
        if(verbose) {
          nun = length(unused)
          famNames = names(amNew)[unused] %||% unused
          cat(sprintf("Removing %d AM famil%s with no remaining missing persons: %s\n", 
                          nun, if(nun == 1) "y" else "ies", trunc(famNames, 6)))
        }
        amNew[unused] = NULL
      }
    }
  }
  
  dviNew = dviData(pmNew, amNew, missNew)
  
  # Fix pairings if they were included in original dataset
  if(!is.null(dvi$pairings)) {
    removedMiss = setdiff(dvi$missing, missNew)
    dviNew$pairings = lapply(dvi$pairings[names(pmNew)], function(v) v[!v %in% removedMiss])
  } 
    
  # PMs with no remaining pairings?
  if(removeUnpairedPM) {
    excl = vapply(dviNew$pairings, function(v) length(v) == 1 && v == "*", FUN.VALUE = FALSE)
    if(sum(excl) > 0) {
      if(verbose)
        cat(sprintf("Removing %s with no remaining pairings: %s\n", 
                    pluralise("PM sample", sum(excl)), trunc(names(excl)[excl], 6)))
      dviNew$pm[excl] = dviNew$pairings[excl] = NULL
    }
  }
  
  dviNew
}

#' Undisputed identifications in a DVI problem
#'
#' This function uses the pairwise LR matrix to find *undisputed* matches
#' between victims and missing individuals. An identification \eqn{V_i = M_j} is
#' called undisputed, relative to a threshold T, if the corresponding likelihood
#' ratio \eqn{LR_{i,j} \geq T} AND \eqn{LR_{i,j}} is at least T times greater
#' than all other pairwise LRs involving \eqn{V_i} or \eqn{M_j}.
#'
#' If the parameter `strict` is set to TRUE, the last criterion is replaced with
#' the stronger requirement that all other pairwise LRs involving \eqn{V_i} or
#' \eqn{M_j} must be at most 1.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param pairings A list of possible pairings for each victim. If NULL, all
#'   sex-consistent pairings are used.
#' @param ignoreSex A logical.
#' @param threshold A non-negative number. If no pairwise LR exceed this, the
#'   iteration stops.
#' @param strict A logical affecting the definition of being undisputed (see
#'   Details). Default: FALSE.
#' @param relax Deprecated; use `strict = FALSE` instead.
#' @param limit A positive number. Only pairwise LR values above this are
#'   considered.
#' @param nkeep An integer, or NULL. If given, only the `nkeep` most likely
#'   pairings are kept for each victim.
#' @param numCores An integer; the number of cores used in parallelisation.
#'   Default: 1.
#' @param keepLRmatrs A logical; for internal purposes. Default: FALSE.
#' @param verbose A logical. Default: TRUE.
#'
#' @seealso [pairwiseLR()], [findExcluded()]
#'
#' @return A list with the following entries:
#'
#'  * `dviReduced`: A reduced version of `dvi`, where undisputed
#'   victims/missing persons are removed, and data from undisputed victims
#'   inserted into the reference data.
#'
#'  * `summary`: A data frame summarising the undisputed matches.
#'
#'  * `LRmatrix`: Output from `pairwiseLR()` applied to
#'   the reduced problem.
#'
#' @examples
#'
#' \donttest{
#' u1 = findUndisputed(planecrash, verbose = FALSE)
#' u1$summary 
#' 
#' # With `strict = TRUE`, the match M3 = V2 goes away
#' u2 = findUndisputed(planecrash, strict = TRUE, verbose = FALSE)
#' u2$summary
#' 
#' # Reason: M3 has LR > 1 also against V7
#' u2$LRmatrix[, "M3"] |> round(2)
#' }
#'
#' @export
findUndisputed = function(dvi, pairings = NULL, ignoreSex = FALSE, 
                          threshold = 1e4, strict = FALSE, relax = !strict, 
                          limit = 0, nkeep = NULL, numCores = 1, 
                          keepLRmatrs = FALSE, verbose = TRUE) {
  
  if(!missing(relax)) {
    cat("Warning: `relax` is deprecated; replaced by (its negation) `strict`")
    strict = !relax
  }
  
  if(verbose) {
    cat("Finding undisputed matches\n")
    cat("Pairwise LR threshold =", threshold, "\n")
  }

  if(!isTRUE(length(threshold) == 1 && threshold > 1))
    stop2("`threshold` must be a number larger than 1: ", threshold %||% "NULL")
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  if(!length(dvi$pm) || !length(dvi$missing))
    return(list(dviReduced = dvi, summary = NULL))
  
  # AM components (for use in output)
  comp = getFamily(dvi, dvi$missing)
  
  # Initialise pairings if not given
  dvi$pairings = pairings %||% dvi$pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  
  # Initialise output
  RES = list()
  
  if(keepLRmatrs)
    LRmatrices = list()
  
  # Loop until problem solved - or no more undisputed matches
  stp = 0
  while(TRUE) {
    
    stp = stp + 1
    
    if(verbose)
      cat(sprintf("\nStep %d:\n", stp))
    
    vics = names(dvi$pm)
    missing = dvi$missing
    
    # Pairwise LR matrix
    ss = pairwiseLR(dvi, limit = limit, nkeep = nkeep, numCores = numCores, 
                    check = FALSE, verbose = verbose)
    B = ss$LRmatrix
    if(keepLRmatrs)
      LRmatrices[[stp]] = B
    
    # Indices of undisputed matches
    undisp = undisputedEntries(B, threshold, strict = strict)
    nu = nrow(undisp)
    
    if(verbose) {
      if(nu == 0) 
        cat(sprintf("No%s undisputed matches\n", if(stp > 1) " further" else ""))
      else 
        cat(sprintf("%d undisputed match%s\n", nu, if(nu == 1) "" else "es"))
    }
    
    if(nu == 0)
      break
    
    for(i in seq_len(nu)) {
      rw = undisp[i,1]
      cl = undisp[i,2]
      vic = vics[rw] 
      RES[[vic]] = list(Step = stp, Missing = missing[cl], LR = B[rw,cl])
      
      if(verbose)
        cat(sprintf(" %s = %s (LR = %.3g)\n", vic, missing[cl], B[rw,cl]))
    }
    
    undispVics = vics[undisp[, 1]]
    undispMP = missing[undisp[, 2]]
    
    # Data from identified samples (keep for updating dviRed below)
    undispData = dvi$pm[undispVics]
    
    # Reduced DVI dataset
    newvics = setdiff(vics, undispVics)
    newmissing = setdiff(missing, undispMP)
    dvi = subsetDVI(dvi, pm = newvics, missing = newmissing, verbose = verbose)
    
    # Victims removed with no remaining pairings
    for(v in setdiff(newvics, names(dvi$pm)))
      RES[[v]] = list(Step = stp, Missing = NA, LR = NA)
    
      
    # Move vic data to AM data
    names(undispVics) = undispMP
    relevantMP = intersect(undispMP, unlist(labels(dvi$am)))
    if(length(relevantMP))
      dvi$am = transferMarkers(from = undispData, to = dvi$am, 
                               idsFrom = undispVics[relevantMP], 
                               idsTo = relevantMP, erase = FALSE)
    
    # Break?
    if(!length(newmissing) || !length(newvics))
      break
  }
  
  summary = do.call(rbind.data.frame, RES)
  isExcl = is.na(summary$Missing)
  step = paste("Step", summary$Step)
  
  if(nrow(summary)) {
    summary$Sample = names(RES)
    summary$Family = comp[summary$Missing]
    summary$Conclusion = ifelse(isExcl, "Excluded", "Undisputed")
    summary$Comment = ifelse(isExcl, "No compatible pairings", step)
    summary = summary[c("Family", "Missing", "Sample", "LR", "Conclusion", "Comment")]
    rownames(summary) = NULL
  }
  else {
    summary = NULL
  }
  
  # Update pairings using output from the last pairwiseLR
  dvi$pairings = ss$pairings
  
  # Keep all or last LR matrix
  LRmatrix = if(keepLRmatrs) LRmatrices else B
  
  list(dviReduced = dvi, summary = summary, LRmatrix = LRmatrix)
}

#' Excluded individuals and pairings in a DVI dataset
#'
#' Analysing exclusions is often an efficient way to reduce large DVI datasets.
#' A pairing V = M is *excluded* if it implies (too many) genetic
#' inconsistencies. The function `findExcluded()` identifies and removes (i)
#' victim samples with too many inconsistencies against all missing persons,
#' (ii) missing persons with too many inconsistencies against all victim
#' samples, and (iii) inconsistent pairings among the remaining.
#'
#' The main calculation in `findExcluded()` is done by `exclusionMatrix()`,
#' which records number of incompatible markers of each pairwise comparison.
#'
#' @param dvi A [dviData()] object.
#' @param maxIncomp An integer. A pairing is excluded if the number of
#'   incompatible markers exceeds this.
#' @param pairings A list of possible pairings for each victim. By default,
#'   `dvi$pairings` is used, or, if this is NULL, `generatePairings(dvi,
#'   ignoreSex)`.
#' @param ignoreSex A logical, by default: FALSE.
#' @param verbose A logical, by default TRUE.
#'
#' @return A list with the following entries:
#'
#'   * `exclusionMatrix`: A matrix showing the number of inconsistencies for
#'   each pair (or NA if the pairing was not considered)
#'
#'   * `excluded`: A list of three character vectors:
#'       * `sample`: victim samples excluded against all missing persons
#'       * `missing`: missing persons excluded against all victims
#'       * `fam`: families in which all missing members are excluded against
#'   all victim samples
#'
#'   * `dviReduced`: A reduced version of `dvi`, where the excluded elements
#'   are removed, and the pairings are updated.
#'
#'   * `summary`: A list of data frames `PM` and `AM`, summarising the excluded
#'   individuals.
#'
#' @seealso [findUndisputed()]. See also [forrel::findExclusions()] for analysis
#'   of a specific pairwise comparison.
#'
#' @examples
#' \donttest{
#' e = findExcluded(icmp)
#' e$summary
#' e$exclusionMatrix
#'
#' # The exclusion matrix can also be computed directly:
#' exclusionMatrix(icmp)
#'
#' # Inspect a particular pair: M4 vs V4
#' forrel::findExclusions(icmp$am, id = "M4", candidate = icmp$pm$V4)
#'
#' # Plot one of the incompatible markers
#' plotDVI(icmp, pm = 4, marker ="D7S820")
#' }
#' @export
findExcluded = function(dvi, maxIncomp = 2, pairings = NULL, ignoreSex = FALSE, verbose = TRUE) {
  
  if(verbose) {
    cat("Finding exclusions\n")
    cat("Max incompatible markers =", maxIncomp, "\n")
  }
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  # Pairings
  if(!missing(pairings) || !missing(ignoreSex))
    pairings = pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  else
    pairings = dvi$pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  
  am = dvi$am
  missing = dvi$missing
  vics = names(dvi$pm)
  comp = getFamily(dvi, missing)
  
  # Main calculation: Exclusion matrix
  mat = exclusionMatrix(dvi, pairings = pairings)

  # Utilities for analysing rows/cols of `mat`
  minEx = function(v)
    if(all(is.na(v))) Inf else min(v, na.rm = TRUE)
  
  commnt = function(v) 
    ifelse(v == Inf, "No compatible pairings", sprintf("%s+ inconsistencies", v))
  
  # Victims excluded against all --------------------------------------------
  
  pmMinEx = apply(mat, 1, minEx)
  pmNomatch = pmMinEx > maxIncomp
  excludedVics = vics[pmNomatch]
  
  if(!length(excludedVics)) {
    summaryPM = NULL
    if(verbose)
      cat("\nPM samples excluded against all missing: None\n\n")
  }
  else {
    summaryPM = data.frame(Sample = excludedVics, 
                           Conclusion = "Excluded", 
                           Comment = commnt(pmMinEx[pmNomatch]),
                           row.names = NULL)
    if(verbose) {
      cat("\nPM samples excluded against all missing:\n")
      cat(sprintf(" %s (%s)", summaryPM$Sample, summaryPM$Comment), sep = "\n")
      cat("\n")
    }
  }
  
  # Missing excluded against all --------------------------------------------
  
  missMinEx = apply(mat, 2, minEx)
  missNomatch = missMinEx > maxIncomp
  excludedMissing = missing[missNomatch]
  
  if(!length(excludedMissing)) {
    summaryAM = NULL
    if(verbose)
      cat("Missing persons excluded against all PM samples: None\n\n")
  }
  else {
    summaryAM = data.frame(Family = comp[excludedMissing],
                           Missing = excludedMissing,
                           Conclusion = "Excluded",
                           Comment = commnt(missMinEx[missNomatch]),
                           row.names = NULL)
    if(verbose) {
      cat("Missing persons excluded against all PM samples:\n")
      cat(sprintf(" %s (%s)", summaryAM$Missing, summaryAM$Comment), sep = "\n")
      cat("\n")
    }
  }
  
    # Reduced DVI problem -----------------------------------------------------
  
  keepVics = vics[!pmNomatch]
  keepMissing = missing[!missNomatch]
  
  # Removed families
  famnames = names(am) %||% 1:length(am)
  excludedFams = setdiff(famnames, comp[keepMissing])
  
  if(length(excludedVics) + length(excludedMissing) > 0) {
    dviRed = subsetDVI(dvi, pm = keepVics, missing = keepMissing, verbose = verbose)
  } else {
    dviRed = dvi
  }
  
  # Updated pairings ----------------------------------------------------
  
  keepPairs = apply(mat[keepVics, keepMissing, drop = FALSE], 1, function(rw)
      c("*", keepMissing[!is.na(rw) & rw <= maxIncomp]), simplify = FALSE)
  
  dviRed$pairings = keepPairs

  nRemov = sum(lengths(dvi$pairings)) - sum(lengths(keepPairs))
  if(verbose)
    cat(sprintf("Pairings excluded in total: %d\n", nRemov))
  

  # Return list -------------------------------------------------------------

  excluded = list(sample = excludedVics, missing = excludedMissing, fam = excludedFams)
  summary = list(PM = summaryPM, AM = summaryAM)
  
  list(exclusionMatrix = mat, excluded = excluded, dviReduced = dviRed, 
       summary = summary)
}



#' @rdname findExcluded
#' @importFrom forrel findExclusions
#' @export
exclusionMatrix = function(dvi, pairings = NULL, ignoreSex = FALSE) {
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  am = dvi$am
  missing = dvi$missing
  vics = names(dvi$pm)
  
  # Pairings
  if(!missing(pairings) || !missing(ignoreSex))
    pairings = pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  else
    pairings = dvi$pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  
  # AM components (for use in output)
  comp = getFamily(dvi, missing)
  
  # Initialise matrix
  mat = matrix(NA_integer_, nrow = length(vics), ncol = length(missing), 
               dimnames = list(vics, missing))
  
  # Loop through each pair of victim vs missing
  for(vic in vics) {
    vicdata = dvi$pm[[vic]]
    compatMiss = setdiff(pairings[[vic]], "*")
    for(mis in compatMiss) {
      ref = am[[comp[mis]]]
      mat[vic, mis] = length(findExclusions(ref, id = mis, candidate = vicdata))
    }
  }
  
  mat
}

#' Plot undisputed identifications
#' 
#' @inheritParams plotSolution
#' @param undisputed A data frame containing the undisputed matches, typically
#'   the entry `undisputed` in output from [findUndisputed()] (only three first
#'   columns used).
#' @param ... Further arguments passed on to [plotSolution()].
#'
#' @return NULL
#' 
#' @seealso [findUndisputed()], [plotSolution()]
#' @examples
#'
#' # Example
#' res = findUndisputed(example2, threshold = 2, verbose = FALSE)
#' u = res$summary
#' plotUndisputed(example2, u, marker = 1)
#'
#' @export
plotUndisputed = function(dvi, undisputed, ...){
  
  d = dim(undisputed)[1]
  if(d == 0)
    stop2("No undisputed to plot.")
  
  # Include only families with identification(s)
  dvi = subsetDVI(dvi, am = unique(undisputed$Family), verbose = FALSE)
  
  # Assignment as named vector (vic = miss)
  a = undisputed$Missing
  names(a) = undisputed$Sample

  plotSolution(dvi, assignment = a, ...)
}

#' Compare DVI approaches
#'
#' Compare the efficiency of different computational approaches to DVI.
#'
#' The following methods are available for comparison, through the `methods`
#' parameter:
#'
#' 1. Sequential, without LR updates
#'
#' 2. Sequential, with LR updates
#'
#' 3. Sequential (undisputed) + joint (remaining). Always return the most likely
#' solution(s).
#'
#' 4. Joint - brute force. Always return the most likely solution(s).
#'
#' 5. Like 3, but return winner(s) only if LR > `threshold`; otherwise the empty
#' assignment.
#'
#' 6. Like 4, but return winner(s) only if LR > `threshold`; otherwise the empty
#' assignment.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param true A character of the same length as `dvi$pm`, with the true
#'   solution, e.g., `true = c("M2", "M3", "*")` if the truth is V1 = M2, V2 =
#'   M3 and V3 unmatched.
#' @param refs Character vector with names of the reference individuals. By
#'   default the typed members of `dvi$am`.
#' @param methods A subset of the numbers 1,2,3,4,5,6.
#' @param markers If `simulate = FALSE`: A vector indicating which markers
#'   should be used.
#' @param threshold An LR threshold passed on to the sequential methods.
#' @param simulate A logical, indicating if simulations should be performed.
#' @param db A frequency database used for simulation, e.g.,
#'   `forrel::NorwegianFrequencies`. By default the frequencies attached to
#'   `dvi$am` are used.
#' @param Nsim A positive integer; the number of simulations.
#' @param returnSims A logical: If TRUE, the simulated data are returned without
#'   any DVI comparison.
#' @param seed A seed for the random number generator, or NULL.
#' @param numCores The number of cores used in parallelisation. Default: 1.
#' @param verbose A logical.
#'
#' @return A list of solution frequencies for each method, and a vector of true
#'   positive rates for each method.
#'
#' @examples
#'
#' refs = "R1"
#' db = forrel::NorwegianFrequencies[1:3]
#'
#' # True solution
#' true = c("M1", "M2", "M3")
#'
#' # Run comparison
#' \donttest{
#' # dviCompare(example1, refs, true = true, db = db, Nsim = 2, seed = 123)
#' }
#'
#' # Alternatively, simulations can be done first...
#' sims = dviCompare(example1, refs, true = true, simulate = TRUE,
#'                   db = db, Nsim = 2, seed = 123, returnSims = TRUE)
#'
#' # ... and computations after:
#' \donttest{
#' # dviCompare(sims, refs, true = true, simulate = FALSE)
#' }
#'
#' @importFrom forrel profileSim
#' @importFrom parallel clusterSetRNGStream
#' @export
dviCompare = function(dvi, true, refs = typedMembers(am), methods = 1:6, 
                      markers = NULL, threshold = 1, simulate = TRUE, 
                      db = getFreqDatabase(am), Nsim = 1, returnSims = FALSE, 
                      seed = NULL, numCores = 1, verbose = TRUE) {
  st = Sys.time()
  
  refs = as.character(refs)
  true = as.character(true)
  
  if(verbose) {
    if(simulate) 
      print(dvi)
    else
      print(dvi[[1]])
    cat("\nParameters for DVI comparison:\n")
    cat(" True solution:", toString(true), "\n")
    cat(" Simulate data:", simulate, "\n")
    cat(" Number of sims:", if(simulate) Nsim else length(dvi$pm), "\n")
    cat(" Reference IDs:", toString(refs))
    cat(" LR threshold:", threshold, "\n\n")
  }
  
  if(simulate) {
    dvi = consolidateDVI(dvi)
    
    pm = dvi$pm
    am = dvi$am
    missing = as.character(dvi$missing)
    
    vics = names(pm)
    isMatch = true != "*"
    
    stopifnot(length(true) == length(vics), 
              all(true[isMatch] %in% missing), 
              all(getSex(am, true[isMatch]) == getSex(pm)[isMatch]))
    
    if(!is.null(seed))
      set.seed(seed)
    
    am = setMarkers(am, locusAttributes = db)
    pm = setMarkers(pm, locusAttributes = db)
    
    # Simulate AM
    AMsims = profileSim(am, N = Nsim, ids = c(refs, true[isMatch]))
    
    # Simulate the unrelated victims
    PMsims = forrel::profileSim(pm, N = Nsim, ids = vics[!isMatch])
    
    # For the true matches, transfer from MP to vics
    PMsims = lapply(1:Nsim, function(i) 
      transferMarkers(from = AMsims[[i]], to = PMsims[[i]], 
                      idsFrom = true[isMatch], idsTo = vics[isMatch], erase = FALSE))
    
    # Remove data from missing
    AMsims = lapply(AMsims, function(s) removeGenotypes(s, missing))
    
    # Collect into list of dviData objects
    dviSims = lapply(1:Nsim, function(i) dviData(pm = PMsims[[i]], am = AMsims[[i]], missing = missing))
    
    # Return sims
    if(returnSims)
      return(dviSims)
  }
  else {
    dviSims = dvi
    
    dvi1 = dvi[[1]]
    
    vics = names(dvi1$pm)
    stopifnot(length(true) == length(vics), 
              all(true %in% c(dvi1$missing, "*")),
              setequal(refs, typedMembers(dvi1$am)))
    
    #PMsims = pm
    #AMsims = am
    if(!is.null(markers)) {
      stop2("Marker selection not implemented yet.")
      #PMsims = lapply(PMsims, selectMarkers, markers)
      #AMsims = lapply(AMsims, selectMarkers, markers)
    }
  }
  
  N = length(dviSims)
  
  # Setup parallelisation
  if(paral <- (numCores > 1)) {
    cl = makeCluster(numCores)
    if(verbose) 
      cat("Using", length(cl), "cores\n")
    on.exit(stopCluster(cl))
    clusterEvalQ(cl, library(dvir))
    # clusterExport(cl, c("missing"), envir = environment())
    clusterSetRNGStream(cl, iseed = sample.int(1e6,1))  
  }
  
  # DVI functions (just to reduce typing)
  fun1 = function(i) sequentialDVI(dviSims[[i]], threshold = threshold, updateLR = FALSE, check = FALSE, verbose = FALSE)$matches
  fun2 = function(i) sequentialDVI(dviSims[[i]], threshold = threshold, updateLR = TRUE, check = FALSE, verbose = FALSE)$matches
  fun3 = function(i) top(jointDVI(dviSims[[i]], undisputed = TRUE, threshold = threshold, check = FALSE, verbose = FALSE))
  fun4 = function(i) top(jointDVI(dviSims[[i]], undisputed = FALSE, check = FALSE, verbose = FALSE))
  fun5 = function(i) top(jointDVI(dviSims[[i]], undisputed = TRUE, threshold = threshold, check = FALSE, verbose = FALSE), threshold)
  fun6 = function(i) top(jointDVI(dviSims[[i]], undisputed = FALSE, check = FALSE, verbose = FALSE), threshold)
  
  # Initialise list of results
  res = list()
  
  # Approach 1: Sequential - naive
  if(1 %in% methods) {
    method1 = if(paral) parLapply(cl, 1:N, fun1) else lapply(1:N, fun1)
    res$method1 = summar(method1)
    if(verbose) print(res['method1'])
  }
  
  # Approach 2: Sequential - with LR update
  if(2 %in% methods) {
    method2 = if(paral) parLapply(cl, 1:N, fun2) else lapply(1:N, fun2)
    res$method2 = summar(method2)
    if(verbose) print(res['method2'])
  }
  
  # Approach 3: Sequential undisputed + joint
  if(3 %in% methods) {
    method3 = if(paral) parLapply(cl, 1:N, fun3) else lapply(1:N, fun3); 
    res$method3 = summar(method3)
    if(verbose) print(res['method3'])
  }
  
  # Approach 4: Joint
  if(4 %in% methods) {
    method4 = if(paral) parLapply(cl, 1:N, fun4) else lapply(1:N, fun4)
    res$method4 = summar(method4)
    if(verbose) print(res['method4'])
  }
  
  # Approach 5: Sequential undisputed + joint + threshold (LR < thresh => *-*-*)
  if(5 %in% methods) {
    method5 = if(paral) parLapply(cl, 1:N, fun5) else lapply(1:N, fun5); 
    res$method5 = summar(method5)
    if(verbose) print(res['method5'])
  }
  
  # Approach 6: Joint + threshold (LR < thresh => *-*-*)
  if(6 %in% methods) {
    method6 = if(paral) parLapply(cl, 1:N, fun6) else lapply(1:N, fun6)
    res$method6 = summar(method6)
    if(verbose) print(res['method6'])
  }
  
  # True positive rates
  TPR = sapply(res, function(r) as.numeric(r[paste(true, collapse = "-")]))
  TPR[is.na(TPR)] = 0
  res$TPR = TPR
  
  if(verbose)
    cat("Total time used:", format(Sys.time() - st, digits = 3), "\n")
  
  res
}

# Utility for summarising list of solutions
#' @importFrom stats aggregate
summar = function(x) {
  sols = lapply(x, apply, 1, paste, collapse = "-")
  L = lengths(sols)
  df = data.frame(sol = unlist(sols), wei = rep(1/L, L))
  agg = aggregate(wei ~ sol, data = df, FUN = sum)
  freqs = agg$wei/length(x)
  names(freqs) = agg$sol
  sort(freqs, decreasing = T)
}

# Utility for including ties in output of jointDVI()
top = function(res, thresh = 1) {
  nvic = ncol(res) - 3
  if(res$LR[1] < thresh) return(rbind(rep("*", nvic)))
  mx = res$LR == res$LR[1]
  res[mx, seq_len(nvic), drop = FALSE]
}

#' Sex-consistent pairings
#'
#' Generate a list of sex-consistent pairings for each victim in a DVI problem.
#' By default, the empty pairing (denoted `*`) is included for each victim.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param includeEmpty A logical. If TRUE (default), the do-nothing symbol (`*`)
#'   is included for each victim.
#' @param ignoreSex A logical.
#'
#' @return A list of character vectors. Each vector is a subset of `missing`,
#'   plus the character `*` denoting no pairing.
#'
#' @seealso [jointDVI()]
#'
#' @examples
#'
#' pm = singletons(c("V1", "V2"), sex = 1:2)
#'           
#' missing = paste0("M", 1:4)
#' am = list(nuclearPed(children = missing[1:3]),
#'           nuclearPed(children = missing[4], sex = 2))
#' 
#' dvi = dviData(pm, am, missing)
#' generatePairings(dvi)
#'
#' @export
generatePairings = function(dvi, includeEmpty = TRUE, ignoreSex = FALSE){
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  pm = dvi$pm
  am = dvi$am
  missing = dvi$missing
  
  # ID of victims
  vics = names(pm)
  
  if(ignoreSex) {
    mp = if(includeEmpty) c("*", missing) else missing
    lst = replicate(length(vics), mp, simplify = FALSE)
    names(lst) = vics
    return(lst)
  }
    
  # Vectors with sex of victims and missing
  sexVics = getSex(pm, vics, named = TRUE)
  sexMP = getSex(am, missing)
  
  # For each victim, find missing of same sex
  lst = lapply(vics, function(v) {
    mp = missing[sexMP == sexVics[v]]
    if(includeEmpty) 
      mp = c("*", mp)
    mp
  })
  names(lst) = vics

  lst
}

#' DVI dataset: Generational trio
#'
#' A proof-of-concept dataset involving three missing members (child, father,
#' grandfather) of a single family. With the given data, stepwise victim
#' identification fails to find the correct solution, while joint identification
#' succeeds.
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 3 singletons (victims).
#'
#'   * `am`: A pedigree with three missing persons and one typed reference
#'   individual.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples    
#' example1
#' plotDVI(example1, marker = 1)
#' 
#' jointDVI(example1)  
#' 
"example1"


#' DVI dataset: Two reference families
#'
#' A small DVI example with three victims, and three missing persons from two reference
#' families
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 3 singletons (victims).
#'
#'   * `am`: A list of 2 pedigrees with three missing persons and one typed reference
#'   individual.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'   
#' @examples    
#' example2
#' plotDVI(example2, marker = 1)
#' 
#' jointDVI(example2)
#' 
"example2"

#' DVI dataset: Family of fire victims
#'
#' A family with three missing persons after a fire, and one reference
#' individual. This example is featured in the GLR paper (Egeland & Vigeland,
#' 2024).
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 3 singletons (victims).
#'
#'   * `am`: A pedigree with three missing persons and one typed reference
#'   individual.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples
#' fire
#' plotDVI(fire, marker = 1)
#' dviSolve(fire, verbose = FALSE)
#' 
"fire"

#' DVI dataset: Simulated plane crash
#'
#' A simulated dataset based on Exercise 3.3 in Egeland et al. "Relationship
#' Inference with Familias and R" (2015).
#'
#' The 15 markers are `CSF1PO`, `D13S317`, `D16S539`, `D18S51`, `D21S11`,
#' `D3S1358`, `D5S818`, `D7S820`, `D8S1179`, `FGA`, `PENTA_D`, `PENTA_E`,
#' `TH01`, `TPOX`, and `VWA`.
#'
#' Source code for the simulation, and a file containing the allele frequencies,
#' can be found in the `data-raw` folder of the GitHub repository:
#' https://github.com/magnusdv/dvir.
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 8 female singletons (victims).
#'
#'   * `am`: A list of 5 pedigrees, each with one missing member and one
#'   genotyped member.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples
#' planecrash
#' plotDVI(planecrash)
#'
#' # Markers and allele frequencies
#' db = getDatabase(planecrash)
#' db[1:3]
#' 
"planecrash"


#' DVI dataset: A large reference pedigree
#'
#' DVI dataset based loosely on the ICMP 2017 workshop material
#' https://www.few.vu.nl/~ksn560/Block-III-PartI-KS-ISFG2017.pdf (page 18).
#' There are 3 female victims, 2 male victims and 6 missing persons of both
#' sexes. We have renamed the individuals and simulated data for 13 CODIS
#' markers (see Details).
#'
#' The 13 markers are, in order: `CSF1PO`, `D3S1358`, `D5S818`,`D7S820`,
#' `D8S1179`, `D13S317`, `D16S539`, `D18S51`, `D21S11`, `FGA`, `TH01`, `TPOX`,
#' and `vWA`.
#'
#' Source code for the simulation, and a file containing the allele frequencies,
#' can be found in the `data-raw` folder of the GitHub repository:
#' https://github.com/magnusdv/dvir.
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 5 singletons (victims).
#'
#'   * `am`: A reference pedigree with 6 genotyped members and 12 missing
#'   persons.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#'
#' @examples
#' icmp
#' plotDVI(icmp)
#' 
#' # Markers and allele frequencies
#' db = getDatabase(icmp)
#' db[1:3]
#' 
"icmp"


#' DVI dataset: Family grave
#'
#' Family grave data in Kling et al. (2021) "Mass Identifications: Statistical
#' Methods in Forensic Genetics". There are 5 female victims and 3 male victims.
#' There is one reference family with 5 missing females and 3 missing males.
#' There are 23 markers, no mutation model.
#'
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 8 singletons (victims).
#'
#'   * `am`: A pedigree with 8 missing persons.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples
#' grave
#' plotDVI(grave, marker = 1)
#'
#' # dviSolve(grave)
#' 
"grave"


#' Data used in the book Kling et al. (2021)
#'
#' Data used in last example of Chapter 4 in Kling et al. (2021) "Mass
#' Identifications: Statistical Methods in Forensic Genetics". There are 2
#' female victims, 2 male victims. There are four reference families with 2
#' missing females and 2 missing males. There are 21 markers. An `equal mutation
#' mode with rate 0.005 is specified.
#'
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 4 singletons (victims).
#'
#'   * `am`: A list of 3 pedigrees.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples
#' KETPchap484
#' 
#' plotDVI(KETPchap484, nrowPM = 4)
#' 
#' 
"KETPchap484"


#' Data used in the book Kling et al. (2021)
#'
#' Data used in Exercise 4.9.7 in Kling et al. (2021) "Mass Identifications:
#' Statistical Methods in Forensic Genetics". There are 3 female victims and 3
#' reference families with 3 missing females. There are 23 markers, equal
#' mutation model, rate 0.001.
#'
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 3 singletons (victims).
#'
#'   * `am`: A list of 3 pedigrees.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples
#' plotDVI(KETPexer497, nrowPM = 3)
#'
#'
"KETPexer497"



#' Data used in the book Kling et al. (2021)
#'
#' Data used in Exercise 4.9.8 in Kling et al. (2021) "Mass Identifications:
#' Statistical Methods in Forensic Genetics". There are 2 female victims and 1
#' male. There is one reference family with 2 missing females and one missing
#' male. There are 16 markers, equal mutation model, rate 0.001.
#'
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 3 singletons (victims).
#'
#'   * `am`: A list of 1 pedigree.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'   
#' @examples 
#' 
#' plotDVI(KETPexer498)
#' 
"KETPexer498"


#' Data used in the book Kling et al. (2021)
#'
#' Data used in the example in Chapter 4.8.1 in Kling et al. (2021) "Mass
#' Identifications: Statistical Methods in Forensic Genetics". The victims are
#' V1 and V2, both females. There is one reference family with 2 missing
#' persons, both females. There are 21 markers, no mutation model.
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 2 singletons (victims).
#'
#'   * `am`: A list of 1 pedigree.
#'
#'   * `missing`: A vector containing the names of the missing persons.
#'
#' @examples
#' plotDVI(KETPchap481, marker = 1)
#' 
"KETPchap481"


#' Dataset: Exclusion example
#'
#' This data is based on a real case, but pedigrees have been changed and
#' marker data simulated to preserve anonymity.
#' 
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 16 singletons (male victims).
#'
#'   * `am`: A list of 15 pedigrees, each with one missing person 
#'
#'   * `missing`: A vector containing the names of the 15 missing persons.
#'   
#' @examples
#' 
#' exclusionExample
#' 
"exclusionExample"

#' Dataset: Symmetry examples
#'
#' A toy DVI dataset illustrating various forms of undecidability due to
#' symmetries in the solutions.
#'
#' @format A `dviData` object with the following content:
#'
#'   * `pm`: A list of 5 singletons (male victims).
#'
#'   * `am`: A list of 3 pedigrees
#'
#'   * `missing`: A character vector of length 5, naming the missing persons.
#'
#' @examples
#'
#' symmetricSibs
#' 
"symmetricSibs"

#' Extract the marker database from a DVI dataset
#'
#' @param dvi A [dviData()] object.
#' @param check A logical. If TRUE (default), the functions checks if the AM and
#'   PM databases are identical, and raises an error if not. If FALSE, the AM
#'   database is returned without checking PM.
#' @param includeAttrs A logical, indicating if the full list of attributes of
#'   each marker (including mutation model and position data) should be
#'   returned, or only the allele frequencies. Default: FALSE.
#'
#' @returns A list of marker information.
#'
#' @examples
#' getDatabase(example1)
#' getDatabase(example1, includeAttrs = TRUE)
#'
#' @export
getDatabase = function(dvi, check = TRUE, includeAttrs = FALSE) {
  dvi = consolidateDVI(dvi)
  getFUN = if(includeAttrs) getLocusAttributes else getFreqDatabase
  
  dbAM = getFUN(dvi$am) 
  if(!check) 
    return(dbAM)
  
  dbPM = getFUN(dvi$pm)
  
  if(!identical(dbAM, dbPM))
    stop2("Databases differ between AM and PM")
  
  dbAM
}

#' Joint DVI search
#'
#' This is a redesign of [jointDVI()], with narrower scope (no preprocessing
#' steps) and more informative output. The output includes the pairwise GLR
#' matrix based on the joint table.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param assignments A data frame containing the assignments to be considered
#'   in the joint analysis. By default, this is automatically generated by
#'   taking all combinations from `dvi$pairings`.
#' @param disableMutations Either TRUE, FALSE (default) or NA. If NA, mutation
#'   modelling is applied only in families where the reference data are
#'   incompatible with the pedigree unless at least one mutation has occurred.
#' @param ignoreSex A logical, only relevant if `dvi$pairings` is NULL, so that
#'   candidate pairings have to be generated.
#' @param maxAssign A positive integer. If the number of assignments going into
#'   the joint calculation exceeds this, the function will abort with an
#'   informative error message. Default: 1e5.
#' @param numCores An integer; the number of cores used in parallelisation.
#'   Default: 1.
#' @param cutoff A number; if non-negative, the output table is restricted to
#'   LRs equal to or exceeding this value.
#' @param verbose A logical.
#' @param progress A logical, indicating if a progress bar should be shown.
#'
#' @return A data frame.
#' @examples
#' dviJoint(example2)
#'
#' @export
dviJoint = function(dvi, assignments = NULL, ignoreSex = FALSE, disableMutations = FALSE, 
                    maxAssign = 1e5, numCores = 1, cutoff = 0, verbose = TRUE, progress = verbose) {
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  if(verbose)
    print.dviData(dvi)
  
  ### Mutation disabling
  if(any(allowsMutations(dvi$am))) {
    am = dvi$am
    
    if(verbose) 
      cat("\nMutation modelling:\n")
    
    if(isTRUE(disableMutations)) {
      if(verbose) cat(" Disabling mutations in all families\n")
      disableFams = seq_along(am)
    }
    else if(identical(disableMutations, NA)) {
      am.nomut = setMutmod(am, model = NULL)
      badFams = vapply(am.nomut, loglikTotal, FUN.VALUE = 1) == -Inf
      if(verbose) {
        if(any(badFams)) 
          cat("", sum(badFams), "inconsistent families:", trunc(which(badFams)), "\n")
        cat("", sum(!badFams), "consistent families. Disabling mutations in these\n")
      }
      disableFams = which(!badFams)
    }
    else disableFams = NULL
  
    if(length(disableFams)) {
      am[disableFams] = setMutmod(am[disableFams], model = NULL)
    }
    
    # Update DVI object
    dvi$am = am
  }
  
  pm = dvi$pm 
  am = dvi$am
  vics = names(pm)
  pairings = dvi$pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  
  if(is.null(assignments)) {
    if(verbose) cat("\nCalculating pairing combinations...")
    # Expand pairings to assignment data frame
    assignments = expand.grid.nodup(pairings, max = maxAssign)
    if(verbose) cat(nrow(assignments), "\n")
  }
  else {
    if(verbose) cat("\nSupplied pairing combinations: ")
    if(!setequal(names(assignments), vics))
      stop2("Names of `assignments` do not match `pm` names")
    assignments = assignments[vics]
    if(verbose) cat(nrow(assignments), "\n")
  }
  
  nAss = nrow(assignments)
  if(nAss == 0)
    stop2("No possible solutions")
  
  # Initial loglikelihoods
  logliks.PM = vapply(pm, loglikTotal, FUN.VALUE = 1)
  logliks.AM = vapply(am, loglikTotal, FUN.VALUE = 1)
  
  loglik0 = sum(logliks.PM) + sum(logliks.AM)
  if(loglik0 == -Inf)
    stop2("Impossible initial data: AM component ", which(logliks.AM == -Inf))
  
  # Parallelise?
  if(is.na(numCores))
    numCores = max(detectCores() - 1, 1)

  cl = NULL
  if(numCores > 1) {
    if(verbose) 
      cat("Using", numCores, "cores\n")
    
    cl = makeCluster(numCores)
    on.exit(stopCluster(cl))
    clusterEvalQ(cl, library(dvir))
    clusterExport(cl, "loglikAssign", envir = environment())
  }
  
  # Convert to list; more handy below
  assignmentList = lapply(1:nAss, function(i) as.character(assignments[i, ]))
  
  # Max 20 chunks to each worker (to reduce overhead but maintain informative PB)
  assignmentList = split(assignmentList, cut(1:nAss, numCores * 20, labels = FALSE))
  
  # Progress bar?
  op = pboptions(type = if(progress) "timer" else "none")
  
  # Main calculation
  loglik = pblapply(cl = cl, assignmentList, function(chunk)
    lapply(chunk, function(a) loglikAssign(pm, am, vics, a, loglik0, logliks.PM, logliks.AM)))
  
  op = pboptions(type = if(progress) "timer" else "none")

  loglik = unlist(loglik, use.names = FALSE)
  
  # Sort in decreasing likelihood, break ties with assignments (alphabetically)
  g = assignments
  g[g == "*"] = NA
  g = cbind(ll = -loglik, g)
  ORD = do.call(order, g)

  assignments = assignments[ORD, , drop = FALSE]
  rownames(assignments) = NULL
  
  # Sorted (joint) logliks and LR
  loglik = loglik[ORD]
  
  # Skip these!?
  #LR = exp(loglik - loglik0) # joint LR
  #posterior = LR/sum(LR) # assuming flat prior
  
  # Collect joint results, including both PM and AM centric assignment tables
  joint = cbind(assignments, loglik = loglik, # LR = LR, posterior = posterior, 
                swapOrientation(assignments))
  joint
}


#' Swap orientation of an assignment table
#'
#' This function switches the roles of victims and missing persons in a table of
#' assignments, from PM-oriented (victims as column names) to AM-oriented (missing
#' persons as column names), and _vice versa_. In both version, each row
#' describes the same assignment vector.
#'
#' @param df A data frame. Each row is an assignment, with `*` representing
#'   non-pairing.
#' @param from A character vector; either victims or missing persons. By
#'   default, the column names of `df`. The only time this argument is needed,
#'   if when `df` has other columns in addition, as in output tables of
#'   `dviJoint()`.
#' @param to The column names of the transformed data frame. If missing, the
#'   unique elements of `df` are used. An error is raised if `to` does not
#'   contain all elements of `df` (except `*`).
#'
#' @return A data frame with `nrow(df)` rows and `length(to)` columns.
#'
#' @examples
#' df = example1 |> generatePairings() |> expand.grid.nodup()
#' df
#' swapOrientation(df)
#' 
#' # Swap is idempotent
#' stopifnot(identical(swapOrientation(swapOrientation(df)), df))
#' 
#' @export
swapOrientation = function(df, from = NULL, to = NULL) {

  if(is.null(from))
    from = names(df)
  
  # Matrix for speed
  amat = as.matrix(df[from])
  
  if(is.null(to))
    to = setdiff(sort.default(unique.default(amat)), "*")
  
  dims = dim(amat)
  
  # Match
  toExt = c(to, "*")
  idx = match(amat, toExt, nomatch = 0L)
  if(any(idx == 0))
     stop2("Table entry not included in `to`: ", setdiff(amat, toExt))
  idx[idx == length(toExt)] = 0L
  dim(idx) = dims
  
  # Melt to long format
  long = cbind(which(idx > 0, arr.ind = TRUE), 
               val = idx[idx > 0])
  
  # Create result matrix
  new = matrix("*", nrow = dims[1], ncol = length(to), dimnames = list(NULL, to))
  
  # Pivot back to wide
  new[long[, c(1,3)]] = from[long[,2]]
  
  res = as.data.frame(new)
  if(ncol(df) > length(from))
    res = cbind(res, df[!names(df) %in% from])
  res
}

#' Identity and merge matching PM samples
#'
#' Computes the direct matching LR of each pair of samples, and merges the
#' matching samples.
#'
#' The available methods for merging matched samples are:
#'
#' * "mostcomplete": Use the sample with the highest number of non-missing genotypes
#'
#' * "first": Use the first in each group, according to the input order
#'
#' * "combine": Not implemented yet.
#'
#' @param pm A list of typed singletons.
#' @param threshold LR threshold for positive identification.
#' @param method A keyword indicating how to merging matching samples. See Details.
#' @param verbose A logical.

#' @seealso [directMatch()].
#'
#' @returns A list with the following entries:
#' * `groups`: A list containing the groups of matching samples.
#'
#' * `LRmat`: A symmetric matrix (with 0s on the diagonal) containing the direct
#'   matching LR values.
#'
#' * `nonmissing`: A named vector reporting the number of non-missing genotypes 
#' for each sample.
#' 
#' * `pmReduced`: A list of singletons. If `use` is "best" or "first", this is 
#' a subset of the input `pm`.
#' 
#' @examples
#'
#' pm = singletons(c("V1", "V2", "V3")) |> 
#'   addMarker(V1 = "1/1", V2 = "2/2", V3 = "1/1", 
#'             afreq = c("1" = 0.01, "2" = 0.99), name = "L1")
#' 
#' mergePM(pm)
#'
#' @export
mergePM = function(pm, threshold = 1e4, method = c("mostcomplete", "first", "combine"), 
                   verbose = TRUE) {
  
  n = length(pm)
  method = match.arg(method)
  
  if(verbose) {
    msg = c(sprintf("Number of singletons: %d", n),
            sprintf("LR threshold: %g", threshold),
            sprintf("Merging method: '%s'", method))
    cat(msg, sep = "\n")
  }
  if(n < 2) {
    if(verbose) cat("Nothing to do\n")
    return(list())
  }
  
  g = getGenotypes(pm)
  ids = rownames(g)
  names(pm) = ids
  
  # Number of non-missing for each sample
  nonmissing = rowSums(g != "-/-")
  
  # LR matrix (upper triangular)
  LRs = matrix(0, nrow = n, ncol = n, dimnames = list(ids, ids))
  for(i in 1:(n-1)) for(j in (i+1):n)
    LRs[i,j] = directMatch(pm[[i]], pm[[j]], g1 = g[i, ], g2 = g[j, ], 
                           .skipChecks = TRUE)
  
  # Find clusters of matching samples. NB: Indices!
  clust = list()
  for(i in 1:n) {
    # Matches in row i (including diagonal entry)
    rmatch = c(i, which(LRs[i, ] >= threshold))
    new = TRUE
    
    # Loop over current clusters
    K = length(clust)
    for(k in seq_len(K)) {
      this = clust[[k]]
      if(any(rmatch %in% this)) {
        clust[[k]] = c(this, rmatch)
        new = FALSE
        break
      }
    }
    # If not in previous comps, create new
    if(new)
      clust[[K+1]] = rmatch
  }
  
  # Convert indices to names (sorted by input order)
  groups = lapply(clust, function(idx) ids[sort.default(unique.default(idx))])
  
  # For "mostcomplete", re-sort and add names
  if(method == "mostcomplete") {
    groups = lapply(groups, function(g) g[order(nonmissing[g], decreasing = TRUE)])
    names(groups) = sapply(groups, '[', 1)
  }
    
  # Merge matching samples
  pmReduced = switch(method,
    mostcomplete = pm[names(groups)],
    first = pm[unlist(lapply(groups, function(g) g[1]))],
    combine = stop2("Method 'combine' is not implemented yet")
  )
    
  # Make LR matrix symmetric (with 0 on diag)
  LRmat = LRs + t.default(LRs)
  
  if(verbose) {
    cat("-----\n")
    clust = groups[lengths(groups) > 1]
    if(length(clust)) {
      s = unlist(lapply(clust, function(g) sprintf(" * [%s]\n", toString(g))), use.names = FALSE)
      cat("Groups of matching samples:\n", s, sep = "")
    }
    else
      cat("Groups of matching samples: None\n")
  }
  
  list(groups = groups, 
       LRmat = LRmat,
       nonmissing = nonmissing,
       pmReduced = pmReduced)
}



#' Direct match LR
#'
#' Computes the likelihood ratio comparing if two samples are from the same
#' individual or from unrelated individuals.
#'
#' @param x,y Typed singletons.
#' @param g1,g2 (Optional) Named character vectors with genotypes for `x` and
#'   `y` respectively.
#' @param .skipChecks A logical indicating that various input checks can be skipped,
#'   e.g. when called by `mergePM()`.
#'
#' @return A nonnegative likelihood ratio.
#' @seealso [mergePM()].
#'
#' @examples
#'
#' pm = singletons(c("V1", "V2", "V3")) |>
#'   addMarker(V1 = "1/1", V2 = "2/2", V3 = "1/1",
#'             afreq = c("1" = 0.01, "2" = 0.99), name = "L1")
#'
#' directMatch(pm[[1]], pm[[2]])
#' directMatch(pm[[1]], pm[[3]])
#'
#' @export
directMatch = function(x, y, g1 = NULL, g2 = NULL, .skipChecks = FALSE) {
  if(!.skipChecks) {
    if(!is.singleton(x))
      stop2("First argument is not a singleton: ", class(x)[1])
    if(!is.singleton(y))
      stop2("Second argument is not a singleton: ", class(y)[1])
    
    if(is.null(g1))
      g1 = getGenotypes(x)[1,]
    if(is.null(g2))
      g2 = getGenotypes(y)[1,]
    
    # Add missing name (occurs in cases with only 1 marker)
    if(is.null(names(g1)))
      names(g1) = name(x)
    if(is.null(names(g2)))
      names(g2) = name(y)
    
    commonM = .myintersect(names(g1), names(g2))
    if(!length(commonM)) {
      message("No shared markers")
      return(1)
    }
    
    g1 = g1[commonM]
    g2 = g2[commonM]
  }
  
  if(x$SEX != y$SEX && x$SEX * y$SEX > 0)
    return(0)
  
  miss1 = g1 == "-/-"
  miss2 = g2 == "-/-"
  if(!all(miss1 | miss2 | g1 == g2))
    return(0)
  
  nonmiss = which(!miss1 & !miss2)
  
  lik = likelihood(x, markers = nonmiss)
  prod(1/lik)
}

#' DVI data
#'
#' @param pm A list of singletons: The victim samples.
#' @param am A list of pedigrees: The reference families.
#' @param missing A character vector with names of missing persons.
#' @param generatePairings A logical. If TRUE (default) a list of sex-compatible
#'   pairings is included as part of the output.
#' @param ignoreSex A logical.
#' @param dvi A `dviData` object.
#' @param pairings A list of pairings.
#' @param errorIfEmpty A logical.
#' @param verbose A logical.
#'
#' @return An object of class `dviData`, which is basically a list of `pm`,
#'   `am`, `missing` and `pairings`.
#'
#' @examples
#' dvi = dviData(pm = singleton("V1"), am = nuclearPed(1), missing = "3")
#' dvi
#' 
#' checkDVI(dvi)
#' 
#' @export
dviData = function(pm, am, missing, generatePairings = TRUE) {
  if(inherits(pm, "dviData"))
    return(pm)
  
  # Enforce lists
  if(is.null(pm))
    pm = list()
  else if(is.singleton(pm)) 
    pm = list(pm)
   
  # Ensure `pm` is named
  if(length(pm))
    names(pm) = unlist(labels(pm))
  
  if(is.null(am))
    am = list()
  
  if(!(length(am) == 0 || is.ped(am) || is.pedList(am)))
    stop2("Argument `am` must be a `ped` object or a list of such")
  
  if(is.null(missing))
    missing = character()
  
  if(!all(missing %in% unlist(labels(am))))
    stop2("Missing person not found in AM data: ", setdiff(missing, unlist(labels(am))))
  
  missing = as.character(missing)
  
  dvi = structure(list(pm = pm, am = am, missing = missing, pairings = NULL), 
                  class = "dviData")
  
  if(generatePairings) 
    dvi$pairings = generatePairings(dvi, ignoreSex = FALSE)
  
  consolidateDVI(dvi)
}


#' @export
print.dviData = function(x, ..., heading = "DVI dataset:", printMax = 8) {
  
  dvi = x
  pm = dvi$pm
  am = dvi$am
  missing = dvi$missing
  
  npm = length(pm)
  nam = length(am)
  
  vics = names(pm)
  refs = if(nam) typedMembers(am) else NULL
  amNames = names(am)
  
  # Number of victim males and females:
  nVfemales = length(females(pm))
  nVmales = length(males(pm))
  
  # Number of missing males and females:
  nMPsex = tabulate(getSex(x$am, missing), nbins = 2)
  
  # Need to know later if there are no, equal (>0) or different number of markers
  rangePM = if(npm) range(nMarkers(pm, compwise = TRUE)) else c(0,0)
  rangeAM = if(nam) range(nMarkers(am, compwise = TRUE)) else c(0,0)
  
    
  cat(heading, "\n")
  
  if(printMax == 0) {
    cat(sprintf(" %d victim%s (%dM/%dF)\n", length(pm), if(length(pm) == 1) "" else "s", nVmales, nVfemales))
    cat(sprintf(" %d missing (%dM/%dF)\n", length(missing), nMPsex[1], nMPsex[2]))
    cat(sprintf(" %d reference individual%s\n", length(refs), if(length(refs) == 1) "" else "s"))
    cat(sprintf(" %d pedigree%s\n", nam, if(nam == 1) "" else "s"))  
  }
  else {
    cat(sprintf(" %d victim%s (%dM/%dF): %s\n", 
                length(pm), if(length(pm) == 1) "" else "s", nVmales, nVfemales, trunc(vics, printMax)))
    cat(sprintf(" %d missing (%dM/%dF): %s\n", 
                length(missing), nMPsex[1], nMPsex[2], trunc(missing, printMax)))
    cat(sprintf(" %d reference individual%s: %s\n", length(refs), if(length(refs) == 1) "" else "s", trunc(refs, printMax)))
    cat(sprintf(" %d pedigree%s: %s\n", nam, if(nam == 1) "" else "s", trunc(amNames, printMax)))
  }
  
  ### Number of markers
  
  # Simple case: PM and AM equal, same number for all
  if(min(rangePM, rangeAM) == max(rangePM, rangeAM))
    cat("Markers:", rangePM[1], "\n")
  else {
    if(npm > 0) {
      if(rangePM[1] == rangePM[2])
        cat("Markers (PM):", rangePM[1], "\n")
      else 
        cat(sprintf("Markers (PM): Ranges from %d to %d\n", 
                    rangePM[1], rangePM[2]))
    }
    if(nam > 0) {
      if(rangeAM[1] == rangeAM[2])
        cat("Markers (AM):", rangeAM[1], "\n")
      else 
        cat(sprintf("Markers (AM): Ranges from %d to %d\n", 
                    rangeAM[1], rangeAM[2]))
    }
  }
}


# Utility to be run in the beginning of all major functions, 
# ensuring that the input is correctly formatted.
# Particularly important if the user has modified the `dvi` object.
consolidateDVI = function(dvi, dedup = FALSE) {

  if(!inherits(dvi, "dviData"))
    stop2("Cannot consolidate; input is not a `dviData` object")
  
  # Make sure pm is a list
  if(is.ped(dvi$pm))   # include non-singletons, so that they can be caught below
    dvi$pm = list(dvi$pm)
  
  # Ensure `pm` is correctly named
  labs = unlist(lapply(dvi$pm, function(x) x$ID), use.names = FALSE) # faster than labels(pm)
  if(length(labs) != length(dvi$pm))
    stop2("PM part is not a list of singletons")
  
  names(dvi$pm) = labs
  
  # Make sure am is a list
  if(is.ped(dvi$am))
    dvi$am = list(dvi$am)
  
  # Enforce family names if not present (F1, F2, ...)
  if(is.null(names(dvi$am)) && length(dvi$am))
    names(dvi$am) = paste0("F", seq_along(dvi$am))

  # Ensure all comps have the same markers in the same order
  dvi$pm = harmoniseMarkers(dvi$pm)
  dvi$am = harmoniseMarkers(dvi$am)
  
  # Ensure `missing` is an unnamed character
  dvi$missing = as.character(dvi$missing)
  
  if(dedup)
    dvi = relabelDVI(dvi, othersPrefix = "")
  
  dvi
}



#' @rdname dviData
#' @export
checkDVI = function(dvi, pairings = NULL, errorIfEmpty = FALSE, 
                    ignoreSex = FALSE, verbose = TRUE){
  
  #TODO: Check for duplicated labels! (e.g. among refs)
  
  if(verbose)
    cat("Checking DVI dataset consistency\n")
  
  # Assumes `dvi` has been consolidated
  
  pm = dvi$pm
  am = dvi$am
  missing = dvi$missing
  
  # Added to avoid crash in certain cases.
  if(length(pm) == 0 || length(missing) == 0) {
    if(errorIfEmpty) stop2("Empty DVI problem") 
    else return()
  }
  
  # Check that PM is a list of singletons
  if(!is.list(pm) || !all(vapply(pm, is.singleton, TRUE)))
    stop2("PM data is not a list of singletons")
  
  # Check that all missing are members of a ref pedigree
  comps = getComponent(am, missing, checkUnique = TRUE, errorIfUnknown = FALSE)
  if(anyNA(comps))
    stop2("Missing person not found in the AM pedigree(s): ", missing[is.na(comps)])
  
  unused = setdiff(seq_along(am), comps)
  if(length(unused))
    warning("AM families with no missing individuals: ", toString(unused), 
            call. = FALSE, immediate. = TRUE)
  
  # Check if marker sets are identical
  markersAM = name(am)
  markersPM = name(pm)
  if(!setequal(markersAM, markersPM)) {
    msg = "Warning: Unequal marker sets in PM and AM\n"
    if(length(notinPM <- setdiff(markersAM, markersPM))) {
      n1 = length(notinPM)
      line1 = sprintf(" %d marker%s in AM but not in PM: %s\n", n1, if(n1 == 1) "" else "s", toString(notinPM))
      msg = paste0(msg, line1)
    }
    if(length(notinAM <- setdiff(markersPM, markersAM))) {
      n2 = length(notinAM)
      line2 = sprintf(" %d marker%s in PM but not in AM: %s\n", n2, if(n2 == 1) "" else "s", toString(notinAM))
      msg = paste0(msg, line2)
    }
    cat(msg)
  }
  
  pairings = pairings %||% dvi$pairings
  
  if(length(pairings) == 0)
    return(invisible()) #stop2("Argument `pairings` has length 0")
  
  vics = names(pm)
  vicSex = getSex(pm, vics, named = TRUE)
  
  candidMP = setdiff(unlist(pairings), "*")
  candidSex = getSex(am, candidMP, named = TRUE)
  
  if(!all(candidMP %in% missing))
    stop2("Pairing error: Candidate is not a missing person: ", setdiff(candidMP, missing))
  
  for(v in vics) {
    candid = pairings[[v]]
    if(length(candid) == 0)
      stop2("No available candidate for victim ", v)
    
    if(any(duplicated(candid)))
      stop2("Duplicated candidate for victim ", v)
    
    cand = setdiff(candid, "*")
    if(length(cand) == 0)
      next
    
    if(!ignoreSex) {
      correctSex = candidSex[cand] == vicSex[v]
      if(!all(correctSex)) 
        stop2("Candidate for victim ", v, " has wrong sex: ", cand[correctSex])
    }
  }
  
  if(verbose)
    cat("No problems found\n")
}



#' Get AM component of selected individuals
#'
#' @param dvi A [dviData()] object.
#' @param ids A vector of ID labels of members of `dvi$am`.
#'
#' @return A vector of the same length as `ids`, containing the family names of
#'   the `ids` individuals.
#'
#' @examples
#' getFamily(example2, ids = example2$missing)
#'
#' @export
getFamily = function(dvi, ids) {
  if(!length(ids))
    return(character(0))
  
  dvi = consolidateDVI(dvi)
  
  famnames = names(dvi$am)
  if(is.null(famnames))
    stop2("AM data has no family names")
  
  comp = getComponent(dvi$am, ids, checkUnique = TRUE, errorIfUnknown = TRUE)
  comp = famnames[comp]
  names(comp) = ids
  comp
}


#' Find the simple families of a DVI dataset
#'
#' Extract the names (if present) or indices of the *simple* reference families,
#' i.e., the families containing exactly 1 missing person.
#'
#' @param dvi A `dviData` object.
#'
#' @return A character (if `dvi$am` has names) or integer vector.
#' @seealso [getFamily()]
#' 
#' @examples
#' # No simple families
#' simple1 = getSimpleFams(example1)
#' stopifnot(length(simple1) == 0)
#' 
#' # Second family is simple
#' simple2 = getSimpleFams(example2)
#' stopifnot(simple2 == "F2")
#' 
#' @export
getSimpleFams = function(dvi) {
  dvi = consolidateDVI(dvi)
  famnames = names(dvi$am)
  
  # AM component of each missing
  fams = getFamily(dvi, ids = dvi$missing)
  
  # Number of missing in each (better than `table`)
  nMiss = tabulate(match(fams, famnames))

  res = famnames[nMiss == 1]
  
  # Convert to integer if indices
  if(is.integer(fams)) 
    res = as.integer(res)
  
  res
}

nMissFam = function(dvi) {
  dvi = consolidateDVI(dvi)
  famnames = names(dvi$am)
  
  # AM component of each missing
  fams = getFamily(dvi, ids = dvi$missing)
  
  # Number of missing in each (better than `table`)
  nMiss = tabulate(match(fams, famnames), nbins = length(famnames))
  names(nMiss) = famnames
  
  nMiss
}

dviEqual = function(dvi1, dvi2) {
  test1 = identical(dvi1$pm, dvi2$pm) && 
    identical(dvi1$am, dvi2$am) && 
    identical(dvi1$missing, dvi2$missing) &&
    identical(lengths(dvi1$pairings), lengths(dvi2$pairings))
  if(!test1)
    return(FALSE)
  
  # Pairings may come in different order
  
  for(v in names(dvi1$pairings))
    if(!setequal(dvi1$pairings[[v]], dvi2$pairings[[v]]))
      return(FALSE)
  
  return(TRUE)
}

# How many pairings are removed in the reduced dataset?
# Only count those between vics and missing persons still included
removedPairings = function(dviRed, dvi) {
  newvics = names(dviRed$pm)
  
  # Previous pairings for the current vics
  pOld = dvi$pairings[newvics]
  
  # Remove pairings to missing individuals no longer included
  remMiss = .mysetdiff(dvi$missing, dviRed$missing)
  if(length(remMiss))
    pOld = lapply(pOld, function(a) .mysetdiff(a, remMiss))
  
  sum(lengths(pOld)) - sum(lengths(dviRed$pairings))
}

#' Format final summary table
#'
#' Combines and harmonises summary tables from different DVI analyses
#'
#' The default column order is controlled by `orientation`, which the following
#' effect:
#'
#' * `AM`:
#'     - Column order: `Family`, `Missing`, `Sample`, `LR`, `GLR`, `Conclusion`, `Comment`
#'     - Sort by: `Family` and `Missing`
#' * `PM`:
#'     - Column order: `Sample`, `Missing`, `Family`, `LR`, `GLR`, `Conclusion`, `Comment`
#'     - Sort by: `Sample`
#'
#' Columns (in any of the data frames) other than these are simply ignored.
#'
#' @param dfs A list of data frames.
#' @param orientation Either "AM" or "PM", controlling column order and sorting.
#' @param columns A (optional) character vector with column names in the wanted
#'   order.
#' @param dvi A `dviData` object used for sorting. Note that if given, this must
#'   contain all victims and families.
#'
#' @return A data frame.
#'
#' @examples
#' u = findUndisputed(planecrash)
#' a = amDrivenDVI(u$dviReduced, threshold2 = 500)
#'
#' u$summary
#' a$summary
#'
#' formatSummary(list(u$summary, a$summary$AM))
#' formatSummary(list(u$summary, a$summary$PM), orientation = "PM", dvi = planecrash)
#'
#' @export
formatSummary = function(dfs, orientation = c("AM", "PM"), columns = NULL, dvi = NULL) {
  
  centr = match.arg(orientation)
  
  # Column order: depends on orientation
  allCols = columns %||% switch(centr,
    AM = c("Family", "Missing", "Sample", "LR", "GLR", "Conclusion", "Comment"),
    PM = c("Sample", "Missing", "Family", "LR", "GLR", "Conclusion", "Comment")
  )
  
  # Harmonize data frames to have all columns
  dfsExt = lapply(dfs, function(df) {
    if(is.null(df) || (nrow(df) == 0)) 
      return(NULL)
    
    # AM must have `Missing`; PM must have `Sample`
    df = switch(centr,
      AM = df[!is.na(df$Missing), , drop = FALSE],
      PM = df[!is.na(df$Sample), , drop = FALSE]
    )
    
    if(nrow(df) == 0) 
      return(NULL)
    
    # Missing columns
    miscol = !allCols %in% names(df)
    if(all(miscol))
      return(NULL)
    
    # Add missing columns; fill with NAs
    for(cc in allCols[miscol])
      df[[cc]] = NA
    
    df[allCols]
  })
  
  final = do.call(rbind, dfsExt)
  
  # If no `dvi` object provided, order in standard way
  if(is.null(dvi)) {
    if(!is.null(final)) {
      orderBy = switch(centr, AM = c("Family", "Missing"), PM = "Sample")
      final = final[do.call(order, final[orderBy]), , drop = FALSE]
    }
    return(final)
  }
  
  fullDf = switch(orientation, 
      AM = data.frame(Family = getFamily(dvi, dvi$missing), Missing = dvi$missing, row.names = NULL),
      PM = data.frame(Sample = names(dvi$pm), row.names = NULL))
  
  # If no data, return fullDf
  if(is.null(final)) {
    for (cc in setdiff(allCols, names(fullDf)))
      fullDf[[cc]] = NA
    return(fullDf)
  }

  orderBy = switch(centr, AM = "Missing", PM = "Sample")
  final = merge(fullDf, final, all.x = TRUE, sort = FALSE)
  final = final[match(fullDf[[orderBy]], final[[orderBy]]), , drop = FALSE]
  rownames(final) = NULL
  final
}

#' Add victim samples to a DVI dataset
#'
#' This is a convenience function for adding new victim samples to an existing
#' [dviData()] object. Users are recommended to use this instead of manipulating
#' the `pm` data directly, as it ensures consistency of the output and also
#' recomputes the internal `pairings` object.
#'
#' @param dvi A [dviData()] object.
#' @param newvics A `singleton` object or a list of such.
#'
#' @returns A new [dviData()] object with the new victims added to the `pm`
#'   data.
#'
#' @examples
#' addVictims(example1, singleton("V4"))
#'
#' @export
addVictims = function(dvi, newvics) {
  
  if(is.singleton(newvics))
    newvics = list(newvics)
  else if(!is.pedList(newvics) && pedsize(newvics) == length(newvics))
    stop2("`newvics` must be a `singleton` object or a list of such")
  
  labs = unlist(lapply(newvics, \(x) x$ID), use.names = FALSE)
  if(any(labs %in% names(dvi$pm)))
    stop2("Victim already exists: ", .myintersect(labs, names(dvi$pm)))
  
  # Extend PM data
  newPM = c(dvi$pm, newvics)
  dviData(pm = newPM, am = dvi$am, missing = dvi$missing)
}

#' Pairwise LR matrix
#'
#' For a given DVI problem, compute the matrix consisting of pairwise likelihood
#' ratios \eqn{LR_{i,j}} comparing \eqn{V_i = M_j} to the null. The output may
#' be reduced by specifying arguments `limit` or `nkeep`.
#'
#' @param dvi A `dviData` object, typically created with [dviData()].
#' @param pairings A list of possible pairings for each victim. If NULL, all
#'   sex-consistent pairings are used.
#' @param ignoreSex A logical.
#' @param limit A nonnegative number controlling the `pairings` slot of the
#'   output: Only pairings with LR greater or equal to `limit` are kept. If zero
#'   (default), pairings with LR > 0 are kept.
#' @param nkeep An integer, or NULL. If given, only the `nkeep` most likely
#'   pairings are kept for each victim.
#' @param check A logical, indicating if the input data should be checked for
#'   consistency.
#' @param numCores An integer; the number of cores used in parallelisation.
#'   Default: 1.
#' @param verbose A logical.
#'
#' @return A list with 3 elements:
#'
#'   * `LRmatrix`: A matrix containing the pairwise LR values.
#'
#'   * `LRlist`: A list of numerical vectors, containing the pairwise LRs in
#'   list format.
#'
#'   * `pairings`: A reduced version of the input `pairings`, keeping only
#'   entries with corresponding LR >= `limit`. For the default case `limit = 0`
#'   a strict inequality is used, i.e., LR > 0.
#'
#' @examples
#' pairwiseLR(example1, verbose = TRUE)
#'
#' @export
pairwiseLR = function(dvi, pairings = NULL, ignoreSex = FALSE, limit = 0, nkeep = NULL, 
                    check = FALSE, numCores = 1, verbose = FALSE){
  
  if(verbose)
    cat("Computing matrix of pairwise LR\n")
  
  # Ensure proper dviData object
  dvi = consolidateDVI(dvi)
  
  # Check consistency
  if(check)
    checkDVI(dvi, pairings = pairings, ignoreSex = ignoreSex, verbose = verbose)
  
  pm = dvi$pm
  am = dvi$am
  missing = dvi$missing
  
  if(length(pm) == 0)
    return(list(LRmatrix = NULL, LRlist = list(), pairings = list()))
  
  # Generate pairings
  pairings = pairings %||% dvi$pairings %||% generatePairings(dvi, ignoreSex = ignoreSex)
  
  # Loglik of each victim and each ref
  marks = 1:nMarkers(pm)
  logliks.PM = vapply(pm, loglikTotal, markers = marks, FUN.VALUE = 1)
  logliks.AM = vapply(am, loglikTotal, markers = marks, FUN.VALUE = 1)
  
  # log-likelihood of H0
  loglik0 = sum(logliks.PM) + sum(logliks.AM)
  
  if(loglik0 == -Inf)
    stop2("Impossible initial data: AM component ", which(logliks.AM == -Inf))
  
  # For each victim, compute the LR of each pairing
  vics = names(pm)
  
  # Dont use more cores than the number of vics
  numCores = min(numCores, length(vics))
  
  # Parallelise
  if(numCores > 1) {
    
    if(verbose) 
      cat("Using", numCores, "cores\n")
    
    cl = makeCluster(numCores)
    on.exit(stopCluster(cl))
    clusterEvalQ(cl, library(dvir))
    clusterExport(cl, "pairwise_singlevic", envir = environment())
    
    # Loop through victims
    LRlist = parLapply(cl, vics, function(v) 
      pairwise_singlevic(am, vics, v, pm[[v]], pairings[[v]], marks, loglik0, logliks.PM, logliks.AM))
  }
  else {
    # Default: no parallelisation
    LRlist = lapply(vics, function(v) 
      pairwise_singlevic(am, vics, v, pm[[v]], pairings[[v]], marks, loglik0, logliks.PM, logliks.AM))
  }
  
  names(LRlist) = vics
  
  # Matrix of individual LRs (filled with NA's)
  LRmatrix = matrix(NA_real_, nrow = length(vics), ncol = length(missing), 
                    dimnames = list(vics, missing))
  
  # Fill matrix row-wise
  for (v in vics) {
    lrs = LRlist[[v]]
    lrs = lrs[names(lrs) != "*"]  # remove do-nothing move
    LRmatrix[v, names(lrs)] = unname(lrs)
  }
  
  # Reduce pairings according to `limit` and/or nkeep
  pairingsReduced = lapply(LRlist, function(lrs) {
    if(limit == 0) 
      keepIdx = lrs > 0
    else
      keepIdx = lrs >= limit
    if(limit > 1) 
      keepIdx = keepIdx | names(lrs) == "*"  # Never remove "*"
    
    newpairings = names(lrs)[keepIdx]
    
    # Apply `nkeep` if given
    if(!is.null(nkeep) && length(newpairings) > nkeep) {
      starIdx = match("*", newpairings, nomatch = 0)
      length(newpairings) = nkeep
      if(starIdx > nkeep)
        newpairings = c(newpairings, "*")
    }
    
    newpairings
  })
  
  list(LRmatrix = LRmatrix, LRlist = LRlist, pairings = pairingsReduced)
}


# Function for computing the pairwise LRs for a single victim
pairwise_singlevic = function(am, vics, v, pmV, pairingsV, marks, loglik0, logliks.PM, logliks.AM) {
  
  lrs = vapply(pairingsV, function(mp) {
    
    if(mp == "*") 
      return(1)
    
    # Make copy of AM likelihoods (vector)
    logliks.AM.new = logliks.AM
    
    # The relevant AM component 
    compNo = getComponent(am, mp, checkUnique = TRUE)
    
    # Move victim data to `mp`
    comp = transferMarkers(pmV, am[[compNo]], idsFrom = v, idsTo = mp, erase = FALSE)
    
    # Update likelihood of this comp
    logliks.AM.new[compNo] = loglikTotal(comp, marks)
    
    # Likelihood of remaining PMs
    logliks.PM.new = logliks.PM[setdiff(vics, v)]
    
    # Total loglik after move
    loglik.move = sum(logliks.PM.new) + sum(logliks.AM.new)
    
    # Return LR
    exp(loglik.move - loglik0)
  }, FUN.VALUE = numeric(1), USE.NAMES = FALSE) 
  
  # Add names explicitly (avoiding that names of pairingsV are reused)
  names(lrs) = as.character(pairingsV)
  
  # Return sorted vector
  sort(lrs, decreasing = TRUE)
}

#' The number of assignments for DVI problem
#'
#' The number of victims and missing persons of each sex is given. The number of
#' possible assignments, i.e., the number of ways the victims can be identified
#' with the missing persons, is calculated.
#'
#' @param nVfemales Integer. The number of female victims.
#' @param nMPfemales Integer. The number of female missing persons.
#' @param nVmales Integer. The number of male victims.
#' @param nMPmales Integer. The number of male missing persons.
#'
#' @return The total number of possible assignments.
#'
#' @examples
#' 
#' # Example
#' m1 = ncomb(5,5,5,5) #
#' 
#' # Example: 3 male victims; 2 male missing persons.
#' # The number of a priori possible assignments is
#' m1 = ncomb(0,0,3,2) # 13
#'
#'
#' # Compare with the complete list of assignments
#' m2 = expand.grid.nodup(list(V1 = c("*", "M1", "M2"),
#'                             V2 = c("*", "M1", "M2"),
#'                             V3 = c("*", "M1", "M2")))
#' stopifnot(m1 == nrow(m2))
#'
#' @export
ncomb = function(nVfemales, nMPfemales, nVmales, nMPmales){

# The number of assignments for one sex
  fmoves = function(V,M){
    nmoves = 0
    for (k in 0:min(V,M)){
      n1 = choose(V, k)
      n2 = choose(M, k)
      nmoves = nmoves + n1*n2*factorial(k)
    }
    nmoves
  }
  
  if(nVfemales < 0| nMPfemales < 0 | nVmales < 0 | nMPmales < 0)
    stop2("All parameters must be non-negative integers")
  
# Calculate for each sex and multiply  
  nfe = fmoves(nVfemales, nMPfemales) 
  nma = fmoves(nVmales, nMPmales)
  nfe * nma
}

#' Simulate genotypes in a DVI dataset
#'
#' Simulates genotypes for the references and missing persons in each AM family,
#' transfers to the PM singletons according to the indicated matching. Remaining
#' victims are simulated as unrelated.
#'
#' @param dvi A `dviData` object.
#' @param N	The number of complete simulations to be performed.
#' @param refs A character indicating reference individuals. By default, the
#'   typed members of the input. If `conditional = TRUE`, the `refs` should be a
#'   subset of the typed members, and the simulations are conditional on these.
#' @param truth A named vector of the format `c(vic1 = mis1, vic2 = mis2, ...)`.
#' @param seed An integer seed for the random number generator.
#' @param conditional A logical, by default FALSE. If TRUE, references are kept
#'   unchanged, while the missing persons are simulated conditional on these.
#' @param simplify1	A logical, by default TRUE, removing the outer list layer
#'   when N = 1. See Value.
#' @param verbose A logical.
#'
#' @return If `N = 1`, a `dviData` object similar to the input, but with new
#'   genotypes for the pm samples. If `N > 1`, a list of `dviData` objects.
#'
#' @seealso [forrel::profileSim()].
#'
#' @examples
#'
#' # Simulate refs and missing once and plot:
#' ex = dviSim(example2, N = 1, truth = c(V1 = "M1", V2 = "M2", V3 = "M3"))
#' plotDVI(ex, marker = 1)
#'
#' # Two simulations and plot for the first
#' ex = dviSim(example2, N = 2, truth = c(V1 = "M1", V2 = "M2", V3 = "M3"),
#'             seed = 1729)
#' plotDVI(ex[[1]], marker = 1)
#'
#'
#' @export
dviSim = function(dvi, N = 1, refs = typedMembers(dvi$am), truth = NULL, 
                  seed = NULL, conditional = FALSE, simplify1 = TRUE,
                  verbose = FALSE) {
  dvi = consolidateDVI(dvi)
  
  labsAM = labels(dvi$am) |> unlist(use.names = FALSE)
  if (!all(refs %in% labsAM)) {
    stop2("Unknown reference ID of reference: ", setdiff(refs, labsAM))
  }
  
  if (any(refs %in% dvi$missing)) {
    stop2("Missing person cannot be a reference: ", intersect(refs, dvi$missing))
  }
  
  if (!all(truth %in% dvi$missing)) {
    stop2("Unknown missing person ID: ", setdiff(truth, dvi$missing))
  }
  
  if (!all(names(truth) %in% names(dvi$pm))) {
    stop2("Unknown victim ID: ", setdiff(names(truth), names(dvi$pm)))
  }
  
  if (!is.null(seed)) {
    set.seed(seed)
  }
  
  # Remove existing genotypes in pm
  pm = removeGenotypes(dvi$pm)
  missing = dvi$missing
  
  # Targets for sims: missing persons and also of refs if conditional = FALSE
  if (conditional) {
    idsClear = setdiff(typedMembers(dvi$am), refs)
    idsSim = missing
  } else {
    idsClear = typedMembers(dvi$am)
    idsSim = c(refs, missing)
  }
  
  # Clearing current genotypes
  am = removeGenotypes(dvi$am, ids = idsClear)
  
  # Simulate (conditional on remaining genotypes)
  amSim = profileSim(am, N = N, ids = idsSim, simplify1 = FALSE, verbose = verbose)
  
  # Transfer to PM according to `truth`
  pmSim = lapply(amSim, function(z) {
    transferMarkers(
      from = z, to = pm, idsFrom = truth, idsTo = names(truth), erase = FALSE
    )})
  
  # Erase genotypes of missing
  amSim = lapply(amSim, function(z) removeGenotypes(z, ids = missing))
  
  # Simulate the remaining victims
  remVics = setdiff(names(pm), names(truth))
  if (length(remVics)) 
    pmSim = lapply(pmSim, function(z) forrel::profileSim(z, ids = remVics))
  
  # Collect the list of dvi data objects  
  dviDataSimulated = lapply(seq_len(N), function(i) {
    dviData(pmSim[[i]], amSim[[i]], missing)
  })
 
  if (simplify1 && N == 1) 
    dviDataSimulated = dviDataSimulated[[1]]
  
  dviDataSimulated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/findUndisputed.R
\name{findUndisputed}
\alias{findUndisputed}
\title{Undisputed identifications in a DVI problem}
\usage{
findUndisputed(
  dvi,
  pairings = NULL,
  ignoreSex = FALSE,
  threshold = 10000,
  strict = FALSE,
  relax = !strict,
  limit = 0,
  nkeep = NULL,
  numCores = 1,
  keepLRmatrs = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{pairings}{A list of possible pairings for each victim. If NULL, all
sex-consistent pairings are used.}

\item{ignoreSex}{A logical.}

\item{threshold}{A non-negative number. If no pairwise LR exceed this, the
iteration stops.}

\item{strict}{A logical affecting the definition of being undisputed (see
Details). Default: FALSE.}

\item{relax}{Deprecated; use \code{strict = FALSE} instead.}

\item{limit}{A positive number. Only pairwise LR values above this are
considered.}

\item{nkeep}{An integer, or NULL. If given, only the \code{nkeep} most likely
pairings are kept for each victim.}

\item{numCores}{An integer; the number of cores used in parallelisation.
Default: 1.}

\item{keepLRmatrs}{A logical; for internal purposes. Default: FALSE.}

\item{verbose}{A logical. Default: TRUE.}
}
\value{
A list with the following entries:
\itemize{
\item \code{dviReduced}: A reduced version of \code{dvi}, where undisputed
victims/missing persons are removed, and data from undisputed victims
inserted into the reference data.
\item \code{summary}: A data frame summarising the undisputed matches.
\item \code{LRmatrix}: Output from \code{pairwiseLR()} applied to
the reduced problem.
}
}
\description{
This function uses the pairwise LR matrix to find \emph{undisputed} matches
between victims and missing individuals. An identification \eqn{V_i = M_j} is
called undisputed, relative to a threshold T, if the corresponding likelihood
ratio \eqn{LR_{i,j} \geq T} AND \eqn{LR_{i,j}} is at least T times greater
than all other pairwise LRs involving \eqn{V_i} or \eqn{M_j}.
}
\details{
If the parameter \code{strict} is set to TRUE, the last criterion is replaced with
the stronger requirement that all other pairwise LRs involving \eqn{V_i} or
\eqn{M_j} must be at most 1.
}
\examples{

\donttest{
u1 = findUndisputed(planecrash, verbose = FALSE)
u1$summary 

# With `strict = TRUE`, the match M3 = V2 goes away
u2 = findUndisputed(planecrash, strict = TRUE, verbose = FALSE)
u2$summary

# Reason: M3 has LR > 1 also against V7
u2$LRmatrix[, "M3"] |> round(2)
}

}
\seealso{
\code{\link[=pairwiseLR]{pairwiseLR()}}, \code{\link[=findExcluded]{findExcluded()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Bmarginal.R
\name{Bmarginal}
\alias{Bmarginal}
\title{Posterior pairing probabilities}
\usage{
Bmarginal(jointRes, missing, prior = NULL)
}
\arguments{
\item{jointRes}{Output from \code{\link[=jointDVI]{jointDVI()}}.}

\item{missing}{Character vector with names of missing persons.}

\item{prior}{A numeric vector of length equal the number of rows in
\code{jointRes}. Default is a flat prior.}
}
\value{
A matrix. Row \code{i} gives the posterior probability that victim \code{i} is
one of the missing persons or someone else, denoted '*'.
}
\description{
Compute posterior pairing and non-pairing probabilities, based on a prior and
the output from \code{\link[=jointDVI]{jointDVI()}}.
}
\details{
The prior assigns a probability to each assignment, each row of \code{jointRes}.
If the prior is not specified, a flat prior is used. The prior needs not sum
to 1 since the user may rather choose a flat prior on the \emph{a priori} possible
assignments.
}
\examples{
jointRes = jointDVI(example1)

Bmarginal(jointRes, example1$missing)

# Artificial example: all but optimal solution excluded by prior
Bmarginal(jointRes, example1$missing, prior = c(1, rep(0,26)))


}
\seealso{
\code{\link[=jointDVI]{jointDVI()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{grave}
\alias{grave}
\title{DVI dataset: Family grave}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 8 singletons (victims).
\item \code{am}: A pedigree with 8 missing persons.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
grave
}
\description{
Family grave data in Kling et al. (2021) "Mass Identifications: Statistical
Methods in Forensic Genetics". There are 5 female victims and 3 male victims.
There is one reference family with 5 missing females and 3 missing males.
There are 23 markers, no mutation model.
}
\examples{
grave
plotDVI(grave, marker = 1)

# dviSolve(grave)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{KETPexer497}
\alias{KETPexer497}
\title{Data used in the book Kling et al. (2021)}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 3 singletons (victims).
\item \code{am}: A list of 3 pedigrees.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
KETPexer497
}
\description{
Data used in Exercise 4.9.7 in Kling et al. (2021) "Mass Identifications:
Statistical Methods in Forensic Genetics". There are 3 female victims and 3
reference families with 3 missing females. There are 23 markers, equal
mutation model, rate 0.001.
}
\examples{
plotDVI(KETPexer497, nrowPM = 3)


}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/findExcluded.R
\name{findExcluded}
\alias{findExcluded}
\alias{exclusionMatrix}
\title{Excluded individuals and pairings in a DVI dataset}
\usage{
findExcluded(
  dvi,
  maxIncomp = 2,
  pairings = NULL,
  ignoreSex = FALSE,
  verbose = TRUE
)

exclusionMatrix(dvi, pairings = NULL, ignoreSex = FALSE)
}
\arguments{
\item{dvi}{A \code{\link[=dviData]{dviData()}} object.}

\item{maxIncomp}{An integer. A pairing is excluded if the number of
incompatible markers exceeds this.}

\item{pairings}{A list of possible pairings for each victim. By default,
\code{dvi$pairings} is used, or, if this is NULL, \code{generatePairings(dvi, ignoreSex)}.}

\item{ignoreSex}{A logical, by default: FALSE.}

\item{verbose}{A logical, by default TRUE.}
}
\value{
A list with the following entries:
\itemize{
\item \code{exclusionMatrix}: A matrix showing the number of inconsistencies for
each pair (or NA if the pairing was not considered)
\item \code{excluded}: A list of three character vectors:
\itemize{
\item \code{sample}: victim samples excluded against all missing persons
\item \code{missing}: missing persons excluded against all victims
\item \code{fam}: families in which all missing members are excluded against
all victim samples
}
\item \code{dviReduced}: A reduced version of \code{dvi}, where the excluded elements
are removed, and the pairings are updated.
\item \code{summary}: A list of data frames \code{PM} and \code{AM}, summarising the excluded
individuals.
}
}
\description{
Analysing exclusions is often an efficient way to reduce large DVI datasets.
A pairing V = M is \emph{excluded} if it implies (too many) genetic
inconsistencies. The function \code{findExcluded()} identifies and removes (i)
victim samples with too many inconsistencies against all missing persons,
(ii) missing persons with too many inconsistencies against all victim
samples, and (iii) inconsistent pairings among the remaining.
}
\details{
The main calculation in \code{findExcluded()} is done by \code{exclusionMatrix()},
which records number of incompatible markers of each pairwise comparison.
}
\examples{
\donttest{
e = findExcluded(icmp)
e$summary
e$exclusionMatrix

# The exclusion matrix can also be computed directly:
exclusionMatrix(icmp)

# Inspect a particular pair: M4 vs V4
forrel::findExclusions(icmp$am, id = "M4", candidate = icmp$pm$V4)

# Plot one of the incompatible markers
plotDVI(icmp, pm = 4, marker ="D7S820")
}
}
\seealso{
\code{\link[=findUndisputed]{findUndisputed()}}. See also \code{\link[forrel:findExclusions]{forrel::findExclusions()}} for analysis
of a specific pairwise comparison.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUndisputed.R
\name{plotUndisputed}
\alias{plotUndisputed}
\title{Plot undisputed identifications}
\usage{
plotUndisputed(dvi, undisputed, ...)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{undisputed}{A data frame containing the undisputed matches, typically
the entry \code{undisputed} in output from \code{\link[=findUndisputed]{findUndisputed()}} (only three first
columns used).}

\item{...}{Further arguments passed on to \code{\link[=plotSolution]{plotSolution()}}.}
}
\description{
Plot undisputed identifications
}
\examples{

# Example
res = findUndisputed(example2, threshold = 2, verbose = FALSE)
u = res$summary
plotUndisputed(example2, u, marker = 1)

}
\seealso{
\code{\link[=findUndisputed]{findUndisputed()}}, \code{\link[=plotSolution]{plotSolution()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setPairing.R
\name{excludePairing}
\alias{excludePairing}
\title{Exclude pairings}
\usage{
excludePairing(dvi, victim, missing)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{victim}{The name of a single victim sample.}

\item{missing}{The name(s) of one or several missing individuals.}
}
\value{
A \code{dviData} object.
}
\description{
Disallow certain pairings by removing them from the list \code{dvi$pairings} of
candidate pairings for a given victim sample.
}
\examples{
# Disallow V1 = M1 in the `example2` dataset:
ex = excludePairing(example2, victim = "V1", missing = "M1")
jointDVI(ex, verbose = FALSE)

# Compare with original
jointDVI(example2, verbose = FALSE)

# The only difference is in the `pairings` slot:
ex$pairings
example2$pairings

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviGLR.R
\name{dviGLR}
\alias{dviGLR}
\title{Finds Generalised Likelihood Ratios (GLRs)}
\usage{
dviGLR(dvi, pairings = generatePairings(dvi), dviRes = NULL)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{pairings}{List. See details.}

\item{dviRes}{data frame. Output from \code{jointDVI()}.}
}
\value{
A data frame with GLRs and SGLR (strict GLR, max replaced by min in
the numerator).
}
\description{
Based on a \code{dviData} object, or output from \verb{dviJoint()``or }jointDVI()`,
the GLR, the ratio of the maximum likelihood under H0 to the maximum under H1,
is calculated for specified hypotheses.
}
\details{
The Generalised Likelihood Ratio (GLR) statistic is defined as the
ratio of the maximum likelihood for the alternatives in the numerator
to the maximum in the denominator. The default
\code{pairings = dvir::generatePairings(dvi)} tests all hypotheses. Specific
tests can be specified as shown in an example:
\code{pairings = list(V1 = "M1")} gives a test for H0: V1 = M1 against
H1: V1 != M1. \code{dviRes} will be calculated using \code{jointDVI()} if not provided.
}
\examples{
# dviGLR(example2, pairings = list(V1 = "M1"))

# All tests with output from jointDVI
# dviGLR(example2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviJoint.R
\name{swapOrientation}
\alias{swapOrientation}
\title{Swap orientation of an assignment table}
\usage{
swapOrientation(df, from = NULL, to = NULL)
}
\arguments{
\item{df}{A data frame. Each row is an assignment, with \code{*} representing
non-pairing.}

\item{from}{A character vector; either victims or missing persons. By
default, the column names of \code{df}. The only time this argument is needed,
if when \code{df} has other columns in addition, as in output tables of
\code{dviJoint()}.}

\item{to}{The column names of the transformed data frame. If missing, the
unique elements of \code{df} are used. An error is raised if \code{to} does not
contain all elements of \code{df} (except \code{*}).}
}
\value{
A data frame with \code{nrow(df)} rows and \code{length(to)} columns.
}
\description{
This function switches the roles of victims and missing persons in a table of
assignments, from PM-oriented (victims as column names) to AM-oriented (missing
persons as column names), and \emph{vice versa}. In both version, each row
describes the same assignment vector.
}
\examples{
df = example1 |> generatePairings() |> expand.grid.nodup()
df
swapOrientation(df)

# Swap is idempotent
stopifnot(identical(swapOrientation(swapOrientation(df)), df))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mergePM.R
\name{mergePM}
\alias{mergePM}
\title{Identity and merge matching PM samples}
\usage{
mergePM(
  pm,
  threshold = 10000,
  method = c("mostcomplete", "first", "combine"),
  verbose = TRUE
)
}
\arguments{
\item{pm}{A list of typed singletons.}

\item{threshold}{LR threshold for positive identification.}

\item{method}{A keyword indicating how to merging matching samples. See Details.}

\item{verbose}{A logical.}
}
\value{
A list with the following entries:
\itemize{
\item \code{groups}: A list containing the groups of matching samples.
\item \code{LRmat}: A symmetric matrix (with 0s on the diagonal) containing the direct
matching LR values.
\item \code{nonmissing}: A named vector reporting the number of non-missing genotypes
for each sample.
\item \code{pmReduced}: A list of singletons. If \code{use} is "best" or "first", this is
a subset of the input \code{pm}.
}
}
\description{
Computes the direct matching LR of each pair of samples, and merges the
matching samples.
}
\details{
The available methods for merging matched samples are:
\itemize{
\item "mostcomplete": Use the sample with the highest number of non-missing genotypes
\item "first": Use the first in each group, according to the input order
\item "combine": Not implemented yet.
}
}
\examples{

pm = singletons(c("V1", "V2", "V3")) |> 
  addMarker(V1 = "1/1", V2 = "2/2", V3 = "1/1", 
            afreq = c("1" = 0.01, "2" = 0.99), name = "L1")

mergePM(pm)

}
\seealso{
\code{\link[=directMatch]{directMatch()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{example2}
\alias{example2}
\title{DVI dataset: Two reference families}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 3 singletons (victims).
\item \code{am}: A list of 2 pedigrees with three missing persons and one typed reference
individual.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
example2
}
\description{
A small DVI example with three victims, and three missing persons from two reference
families
}
\examples{
   
example2
plotDVI(example2, marker = 1)

jointDVI(example2)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/findNonidentifiable.R
\name{findNonidentifiable}
\alias{findNonidentifiable}
\title{Nonidentifiable missing persons}
\usage{
findNonidentifiable(dvi)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}
}
\value{
A list with the following entries:
\itemize{
\item \code{nonidentifiable}: A character vector (possibly empty) with the names of
the nonidentifiable missing persons.
\item \code{dviReduced}: A reduced \code{dviData} object, where the nonidentifiable
individuals are removed from the list of missing persons. If there are no
\code{nonidentifiable}, this is just a copy of \code{dvi}.
\item \code{summary}: A data frame summarising the findings.
}
}
\description{
A missing person in a DVI case is nonidentifiable if they are not related to
any (genotyped) reference individuals, neither directly nor through other
missing persons in the family. This function identifies such individuals, and
reduces the DVI dataset accordingly.
}
\details{
Two individuals are unrelated if their kinship coefficient is zero. The
implementation uses \code{ribd::kinship()} to calculate all pairwise kinship
coefficients.
}
\examples{
# Example 1: No nonidentifiables in dataset `example1`
findNonidentifiable(example1)

# Example 2: Add nonidentifiable person "A"
amNew = example1$am[[1]] |>
  addSon(parents = c("NN", "A"))
missNew = c(example1$missing, "A")

dvi = dviData(pm = example1$pm, am = amNew, missing = missNew)
plotDVI(dvi, textAbove = c(A = "nonidentif."))

findNonidentifiable(dvi)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.grid.nodup.R
\name{expand.grid.nodup}
\alias{expand.grid.nodup}
\title{Combinations without duplications}
\usage{
expand.grid.nodup(lst, max = 1e+05)
}
\arguments{
\item{lst}{A list of vectors.}

\item{max}{A positive integer. If the number of combinations (according to a
preliminary lower bound) exceeds this, the function aborts with an
informative error message. Default: 1e5.}
}
\value{
A data frame.
}
\description{
This is similar to \code{\link[=expand.grid]{expand.grid()}} except that combinations with repeated
elements are not included. The element "*" is treated separately, and is
allowed to be repeated.
}
\examples{

lst = list(1, 1:2, 3:4)

# Compare
expand.grid.nodup(lst)
expand.grid(lst)

# Typical use case for DVI
lst2 = generatePairings(example1)
expand.grid.nodup(lst2)

}
\seealso{
\code{\link[=expand.grid]{expand.grid()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotDVI.R
\name{plotDVI}
\alias{plotDVI}
\title{Plot a DVI problem}
\usage{
plotDVI(
  dvi,
  pm = TRUE,
  am = TRUE,
  style = 1,
  famnames = NA,
  hatched = typedMembers,
  frames = TRUE,
  titles = c("PM", "AM"),
  cex = NA,
  col = 1,
  lwd = 1,
  fill = NA,
  carrier = NULL,
  widths = NULL,
  heights = NULL,
  nrowPM = NA,
  nrowAM = NA,
  ncolAM = NA,
  dev.height = NULL,
  dev.width = NULL,
  newdev = !is.null(c(dev.height, dev.width)),
  ...
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{pm}{Either a logical indicating if the PM data should be plotted (as a
set of singletons), or a vector of indices selecting a subset of the PM
samples. Default: TRUE.}

\item{am}{Either a logical indicating if the AM families data should be
plotted, or a vector of indices selecting a subset of the families.
Default: TRUE.}

\item{style}{An integer (currently 1 or 2) indicating the style of the plot.}

\item{famnames}{A logical. If NA (default) family names are included if there
are multiple families.}

\item{hatched}{A character vector of ID labels, or the name of a function. By
default, typed individuals are hatched.}

\item{frames}{A logical, by default TRUE.}

\item{titles}{A character of length 2.}

\item{fill, cex, col, lwd, carrier}{Arguments passed on to \code{\link[pedtools:plot.ped]{pedtools::plot.ped()}}.}

\item{widths, heights}{Numeric with relative columns widths / row heights, to
be passed on to \code{layout()}.}

\item{nrowPM}{The number of rows in the array of PM singletons.}

\item{nrowAM, ncolAM}{The number of rows/cols in the array of AM families.}

\item{dev.height, dev.width}{Plot height and widths in inches. These are
optional, and only relevant if \code{newdev = TRUE}.}

\item{newdev}{A logical indicating if a new plot window should be opened.}

\item{...}{Further parameters to be passed on to \code{\link[pedtools:plot.ped]{pedtools::plot.ped()}},
e.g., \code{marker}, \code{cex}, \code{cex.main}, \code{symbolsize}.}
}
\description{
Plot a DVI problem
}
\examples{

plotDVI(example2)

# Override default layout
plotDVI(example2, nrowPM = 1, ncolAM = 1)

# Subset
plotDVI(example2, pm = 1:2, am = 1, titles = c("PM (1-2)", "AM (1)"))

# AM only
plotDVI(example2, pm = FALSE, titles = "AM families")

# Further plot options
plotDVI(example2, frames = FALSE, marker = 1, cex = 1.2, nrowPM = 3, nrowAM = 2,
  textAnnot = list(inside = list(c(M1 = "?", M2 = "?", M3 = "?"), cex = 1.5)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviJoint.R
\name{dviJoint}
\alias{dviJoint}
\title{Joint DVI search}
\usage{
dviJoint(
  dvi,
  assignments = NULL,
  ignoreSex = FALSE,
  disableMutations = FALSE,
  maxAssign = 1e+05,
  numCores = 1,
  cutoff = 0,
  verbose = TRUE,
  progress = verbose
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{assignments}{A data frame containing the assignments to be considered
in the joint analysis. By default, this is automatically generated by
taking all combinations from \code{dvi$pairings}.}

\item{ignoreSex}{A logical, only relevant if \code{dvi$pairings} is NULL, so that
candidate pairings have to be generated.}

\item{disableMutations}{Either TRUE, FALSE (default) or NA. If NA, mutation
modelling is applied only in families where the reference data are
incompatible with the pedigree unless at least one mutation has occurred.}

\item{maxAssign}{A positive integer. If the number of assignments going into
the joint calculation exceeds this, the function will abort with an
informative error message. Default: 1e5.}

\item{numCores}{An integer; the number of cores used in parallelisation.
Default: 1.}

\item{cutoff}{A number; if non-negative, the output table is restricted to
LRs equal to or exceeding this value.}

\item{verbose}{A logical.}

\item{progress}{A logical, indicating if a progress bar should be shown.}
}
\value{
A data frame.
}
\description{
This is a redesign of \code{\link[=jointDVI]{jointDVI()}}, with narrower scope (no preprocessing
steps) and more informative output. The output includes the pairwise GLR
matrix based on the joint table.
}
\examples{
dviJoint(example2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviData.R
\name{dviData}
\alias{dviData}
\alias{checkDVI}
\title{DVI data}
\usage{
dviData(pm, am, missing, generatePairings = TRUE)

checkDVI(
  dvi,
  pairings = NULL,
  errorIfEmpty = FALSE,
  ignoreSex = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{pm}{A list of singletons: The victim samples.}

\item{am}{A list of pedigrees: The reference families.}

\item{missing}{A character vector with names of missing persons.}

\item{generatePairings}{A logical. If TRUE (default) a list of sex-compatible
pairings is included as part of the output.}

\item{dvi}{A \code{dviData} object.}

\item{pairings}{A list of pairings.}

\item{errorIfEmpty}{A logical.}

\item{ignoreSex}{A logical.}

\item{verbose}{A logical.}
}
\value{
An object of class \code{dviData}, which is basically a list of \code{pm},
\code{am}, \code{missing} and \code{pairings}.
}
\description{
DVI data
}
\examples{
dvi = dviData(pm = singleton("V1"), am = nuclearPed(1), missing = "3")
dvi

checkDVI(dvi)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviCompare.R
\name{dviCompare}
\alias{dviCompare}
\title{Compare DVI approaches}
\usage{
dviCompare(
  dvi,
  true,
  refs = typedMembers(am),
  methods = 1:6,
  markers = NULL,
  threshold = 1,
  simulate = TRUE,
  db = getFreqDatabase(am),
  Nsim = 1,
  returnSims = FALSE,
  seed = NULL,
  numCores = 1,
  verbose = TRUE
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{true}{A character of the same length as \code{dvi$pm}, with the true
solution, e.g., \code{true = c("M2", "M3", "*")} if the truth is V1 = M2, V2 =
M3 and V3 unmatched.}

\item{refs}{Character vector with names of the reference individuals. By
default the typed members of \code{dvi$am}.}

\item{methods}{A subset of the numbers 1,2,3,4,5,6.}

\item{markers}{If \code{simulate = FALSE}: A vector indicating which markers
should be used.}

\item{threshold}{An LR threshold passed on to the sequential methods.}

\item{simulate}{A logical, indicating if simulations should be performed.}

\item{db}{A frequency database used for simulation, e.g.,
\code{forrel::NorwegianFrequencies}. By default the frequencies attached to
\code{dvi$am} are used.}

\item{Nsim}{A positive integer; the number of simulations.}

\item{returnSims}{A logical: If TRUE, the simulated data are returned without
any DVI comparison.}

\item{seed}{A seed for the random number generator, or NULL.}

\item{numCores}{The number of cores used in parallelisation. Default: 1.}

\item{verbose}{A logical.}
}
\value{
A list of solution frequencies for each method, and a vector of true
positive rates for each method.
}
\description{
Compare the efficiency of different computational approaches to DVI.
}
\details{
The following methods are available for comparison, through the \code{methods}
parameter:
\enumerate{
\item Sequential, without LR updates
\item Sequential, with LR updates
\item Sequential (undisputed) + joint (remaining). Always return the most likely
solution(s).
\item Joint - brute force. Always return the most likely solution(s).
\item Like 3, but return winner(s) only if LR > \code{threshold}; otherwise the empty
assignment.
\item Like 4, but return winner(s) only if LR > \code{threshold}; otherwise the empty
assignment.
}
}
\examples{

refs = "R1"
db = forrel::NorwegianFrequencies[1:3]

# True solution
true = c("M1", "M2", "M3")

# Run comparison
\donttest{
# dviCompare(example1, refs, true = true, db = db, Nsim = 2, seed = 123)
}

# Alternatively, simulations can be done first...
sims = dviCompare(example1, refs, true = true, simulate = TRUE,
                  db = db, Nsim = 2, seed = 123, returnSims = TRUE)

# ... and computations after:
\donttest{
# dviCompare(sims, refs, true = true, simulate = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amDrivenDVI.R
\name{amDrivenDVI}
\alias{amDrivenDVI}
\title{AM-driven DVI}
\usage{
amDrivenDVI(
  dvi,
  fams = NULL,
  threshold = 10000,
  threshold2 = max(1, threshold/10),
  verbose = TRUE
)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{fams}{A character; the names of families to consider. By default, all
families. Special keywords: "simple" (all families with exactly 1 missing)
and "nonsimple" (all families with > 1 missing).}

\item{threshold}{LR threshold for 'certain' match.}

\item{threshold2}{LR threshold for 'probable' match (in \emph{simple} families).}

\item{verbose}{A logical.}
}
\value{
A list of \code{dviReduced} and \code{summary}.
}
\description{
AM-driven identification, i.e., considering one AM family at a time. Simple
families (exactly 1 missing) are handled directly from the LR matrix, while
nonsimple families are analysed with \code{\link[=dviJoint]{dviJoint()}}.
}
\details{
Note: This function assumes that undisputed identifications have been
removed. Strange outputs may occur otherwise.
}
\examples{
w = amDrivenDVI(example2)
w$summary
w$dviReduced

# Bigger example: Undisputed first
u = findUndisputed(planecrash)
u$summary

# AM-driven analysis of the remaining
amDrivenDVI(u$dviReduced, threshold2 = 500)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generatePairings.R
\name{generatePairings}
\alias{generatePairings}
\title{Sex-consistent pairings}
\usage{
generatePairings(dvi, includeEmpty = TRUE, ignoreSex = FALSE)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{includeEmpty}{A logical. If TRUE (default), the do-nothing symbol (\code{*})
is included for each victim.}

\item{ignoreSex}{A logical.}
}
\value{
A list of character vectors. Each vector is a subset of \code{missing},
plus the character \code{*} denoting no pairing.
}
\description{
Generate a list of sex-consistent pairings for each victim in a DVI problem.
By default, the empty pairing (denoted \code{*}) is included for each victim.
}
\examples{

pm = singletons(c("V1", "V2"), sex = 1:2)
          
missing = paste0("M", 1:4)
am = list(nuclearPed(children = missing[1:3]),
          nuclearPed(children = missing[4], sex = 2))

dvi = dviData(pm, am, missing)
generatePairings(dvi)

}
\seealso{
\code{\link[=jointDVI]{jointDVI()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncomb.R
\name{ncomb}
\alias{ncomb}
\title{The number of assignments for DVI problem}
\usage{
ncomb(nVfemales, nMPfemales, nVmales, nMPmales)
}
\arguments{
\item{nVfemales}{Integer. The number of female victims.}

\item{nMPfemales}{Integer. The number of female missing persons.}

\item{nVmales}{Integer. The number of male victims.}

\item{nMPmales}{Integer. The number of male missing persons.}
}
\value{
The total number of possible assignments.
}
\description{
The number of victims and missing persons of each sex is given. The number of
possible assignments, i.e., the number of ways the victims can be identified
with the missing persons, is calculated.
}
\examples{

# Example
m1 = ncomb(5,5,5,5) #

# Example: 3 male victims; 2 male missing persons.
# The number of a priori possible assignments is
m1 = ncomb(0,0,3,2) # 13


# Compare with the complete list of assignments
m2 = expand.grid.nodup(list(V1 = c("*", "M1", "M2"),
                            V2 = c("*", "M1", "M2"),
                            V3 = c("*", "M1", "M2")))
stopifnot(m1 == nrow(m2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dvir-package.R
\docType{package}
\name{dvir-package}
\alias{dvir}
\alias{dvir-package}
\title{dvir: Disaster Victim Identification}
\description{
Joint DNA-based disaster victim identification (DVI), as described in Vigeland and Egeland (2021) \doi{10.21203/rs.3.rs-296414/v1}. Identification is performed by optimising the joint likelihood of all victim samples and reference individuals. Individual identification probabilities, conditional on all available information, are derived from the joint solution in the form of posterior pairing probabilities. 'dvir' is part of the 'pedsuite' collection of packages for pedigree analysis.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/magnusdv/dvir}
  \item Report bugs at \url{https://github.com/magnusdv/dvir/issues}
}

}
\author{
\strong{Maintainer}: Magnus Dehli Vigeland \email{m.d.vigeland@medisin.uio.no} (\href{https://orcid.org/0000-0002-9134-4962}{ORCID})

Authors:
\itemize{
  \item Magnus Dehli Vigeland \email{m.d.vigeland@medisin.uio.no} (\href{https://orcid.org/0000-0002-9134-4962}{ORCID})
  \item Thore Egeland \email{thore.egeland@nmbu.no} (\href{https://orcid.org/0000-0002-3465-8885}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{exclusionExample}
\alias{exclusionExample}
\title{Dataset: Exclusion example}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 16 singletons (male victims).
\item \code{am}: A list of 15 pedigrees, each with one missing person
\item \code{missing}: A vector containing the names of the 15 missing persons.
}
}
\usage{
exclusionExample
}
\description{
This data is based on a real case, but pedigrees have been changed and
marker data simulated to preserve anonymity.
}
\examples{

exclusionExample

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{KETPexer498}
\alias{KETPexer498}
\title{Data used in the book Kling et al. (2021)}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 3 singletons (victims).
\item \code{am}: A list of 1 pedigree.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
KETPexer498
}
\description{
Data used in Exercise 4.9.8 in Kling et al. (2021) "Mass Identifications:
Statistical Methods in Forensic Genetics". There are 2 female victims and 1
male. There is one reference family with 2 missing females and one missing
male. There are 16 markers, equal mutation model, rate 0.001.
}
\examples{

plotDVI(KETPexer498)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequentialDVI.R
\name{sequentialDVI}
\alias{sequentialDVI}
\title{Sequential DVI search}
\usage{
sequentialDVI(
  dvi,
  updateLR = TRUE,
  threshold = 1,
  check = TRUE,
  verbose = TRUE,
  debug = FALSE
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{updateLR}{A logical. If TRUE, the LR matrix is updated in each
iteration.}

\item{threshold}{A non-negative number. If no pairwise LR values exceed this,
the iteration stops.}

\item{check}{A logical, indicating if the input data should be checked for
consistency.}

\item{verbose}{A logical, by default TRUE.}

\item{debug}{A logical, by default FALSE. If TRUE, the LR matrix is printed
in each step.}
}
\value{
A list with two elements:
\itemize{
\item \code{matches}: A single assignment vector, or (if multiple branches)
a data frame where each row is an assignment vector.
\item \code{details}: A data frame (or a list of data frames, if multiple branches)
including the LR of each identification in the order they were made.
}
}
\description{
Performs a sequential matching procedure based on the pairwise LR matrix. In
each step the pairing corresponding to the highest LR is selected and
included as a match if the LR exceeds the given threshold. By default,
(\code{updateLR = TRUE}) the pairwise LRs are recomputed in each step after
including the data from the identified sample.
}
\details{
If, at any point, the highest LR is obtained by more than one pairing, the
process branches off and produces multiple solutions. (See Value.)
}
\examples{
# Without LR updates
sequentialDVI(example1, updateLR = FALSE)

# With LR updates (default). Note two branches!
r = sequentialDVI(example1)

# Plot the two solutions
plotSolution(example1, r$matches, k = 1)
plotSolution(example1, r$matches, k = 2)

# Add `debug = T` to see the LR matrix in each step
sequentialDVI(example1, debug = TRUE)

# The output of can be fed into `jointDVI()`:
jointDVI(example1, assignments = r$matches)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mergePM.R
\name{directMatch}
\alias{directMatch}
\title{Direct match LR}
\usage{
directMatch(x, y, g1 = NULL, g2 = NULL, .skipChecks = FALSE)
}
\arguments{
\item{x, y}{Typed singletons.}

\item{g1, g2}{(Optional) Named character vectors with genotypes for \code{x} and
\code{y} respectively.}

\item{.skipChecks}{A logical indicating that various input checks can be skipped,
e.g. when called by \code{mergePM()}.}
}
\value{
A nonnegative likelihood ratio.
}
\description{
Computes the likelihood ratio comparing if two samples are from the same
individual or from unrelated individuals.
}
\examples{

pm = singletons(c("V1", "V2", "V3")) |>
  addMarker(V1 = "1/1", V2 = "2/2", V3 = "1/1",
            afreq = c("1" = 0.01, "2" = 0.99), name = "L1")

directMatch(pm[[1]], pm[[2]])
directMatch(pm[[1]], pm[[3]])

}
\seealso{
\code{\link[=mergePM]{mergePM()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relabelDVI.R
\name{relabelDVI}
\alias{relabelDVI}
\title{Automatic labelling of a DVI dataset}
\usage{
relabelDVI(
  dvi,
  victims = NULL,
  victimPrefix = NULL,
  familyPrefix = NULL,
  refs = NULL,
  refPrefix = NULL,
  missingPrefix = NULL,
  missingFormat = NULL,
  othersPrefix = NULL
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{victims}{A named vector of the form \code{c(old = new)} with names for the
PM samples, or a function to be applied to the existing names.}

\item{victimPrefix}{Prefix used to label PM individuals.}

\item{familyPrefix}{Prefix used to label the AM families.}

\item{refs}{A named vector of the form \code{c(old = new)} with names for the
typed references, or a function to be applied to the existing names.}

\item{refPrefix}{Prefix used to label the reference individuals, i.e., the
typed members of the AM families.}

\item{missingPrefix}{Prefix used to label the missing persons. At most one of
\code{missingPrefix} and \code{missingFormat} can be given.}

\item{missingFormat}{A string indicating family-wise labelling of missing
persons, using \verb{[FAM]}, \verb{[IDX]}, \verb{[MIS]} as place holders with the
following meanings (see Examples):
\itemize{
\item \verb{[FAM]}: family index
\item \verb{[IDX]}: index of missing person within the family
\item \verb{[MIS]}: index within all missing persons
}}

\item{othersPrefix}{Prefix used to label other untyped individuals. Use ""
for numeric labels ( 1, 2, ...).}
}
\value{
A \code{\link[=dviData]{dviData()}} object.
}
\description{
Relabel the individuals and families in a DVI dataset.
}
\examples{

# Builtin dataset `example2`
relabelDVI(example2,
           victimPrefix  = "vic",
           familyPrefix  = "fam",
           refPrefix     = "ref",
           missingPrefix = "mp")

# Family-wise labelling of missing persons
relabelDVI(example2, missingFormat = "M[FAM]-[IDX]")
relabelDVI(example2, missingFormat = "M[IDX] (F[FAM])")
relabelDVI(example2, missingFormat = "fam[FAM].m[IDX]")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getDatabase.R
\name{getDatabase}
\alias{getDatabase}
\title{Extract the marker database from a DVI dataset}
\usage{
getDatabase(dvi, check = TRUE, includeAttrs = FALSE)
}
\arguments{
\item{dvi}{A \code{\link[=dviData]{dviData()}} object.}

\item{check}{A logical. If TRUE (default), the functions checks if the AM and
PM databases are identical, and raises an error if not. If FALSE, the AM
database is returned without checking PM.}

\item{includeAttrs}{A logical, indicating if the full list of attributes of
each marker (including mutation model and position data) should be
returned, or only the allele frequencies. Default: FALSE.}
}
\value{
A list of marker information.
}
\description{
Extract the marker database from a DVI dataset
}
\examples{
getDatabase(example1)
getDatabase(example1, includeAttrs = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwiseLR.R
\name{pairwiseLR}
\alias{pairwiseLR}
\title{Pairwise LR matrix}
\usage{
pairwiseLR(
  dvi,
  pairings = NULL,
  ignoreSex = FALSE,
  limit = 0,
  nkeep = NULL,
  check = FALSE,
  numCores = 1,
  verbose = FALSE
)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{pairings}{A list of possible pairings for each victim. If NULL, all
sex-consistent pairings are used.}

\item{ignoreSex}{A logical.}

\item{limit}{A nonnegative number controlling the \code{pairings} slot of the
output: Only pairings with LR greater or equal to \code{limit} are kept. If zero
(default), pairings with LR > 0 are kept.}

\item{nkeep}{An integer, or NULL. If given, only the \code{nkeep} most likely
pairings are kept for each victim.}

\item{check}{A logical, indicating if the input data should be checked for
consistency.}

\item{numCores}{An integer; the number of cores used in parallelisation.
Default: 1.}

\item{verbose}{A logical.}
}
\value{
A list with 3 elements:
\itemize{
\item \code{LRmatrix}: A matrix containing the pairwise LR values.
\item \code{LRlist}: A list of numerical vectors, containing the pairwise LRs in
list format.
\item \code{pairings}: A reduced version of the input \code{pairings}, keeping only
entries with corresponding LR >= \code{limit}. For the default case \code{limit = 0}
a strict inequality is used, i.e., LR > 0.
}
}
\description{
For a given DVI problem, compute the matrix consisting of pairwise likelihood
ratios \eqn{LR_{i,j}} comparing \eqn{V_i = M_j} to the null. The output may
be reduced by specifying arguments \code{limit} or \code{nkeep}.
}
\examples{
pairwiseLR(example1, verbose = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviSim.R
\name{dviSim}
\alias{dviSim}
\title{Simulate genotypes in a DVI dataset}
\usage{
dviSim(
  dvi,
  N = 1,
  refs = typedMembers(dvi$am),
  truth = NULL,
  seed = NULL,
  conditional = FALSE,
  simplify1 = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{N}{The number of complete simulations to be performed.}

\item{refs}{A character indicating reference individuals. By default, the
typed members of the input. If \code{conditional = TRUE}, the \code{refs} should be a
subset of the typed members, and the simulations are conditional on these.}

\item{truth}{A named vector of the format \code{c(vic1 = mis1, vic2 = mis2, ...)}.}

\item{seed}{An integer seed for the random number generator.}

\item{conditional}{A logical, by default FALSE. If TRUE, references are kept
unchanged, while the missing persons are simulated conditional on these.}

\item{simplify1}{A logical, by default TRUE, removing the outer list layer
when N = 1. See Value.}

\item{verbose}{A logical.}
}
\value{
If \code{N = 1}, a \code{dviData} object similar to the input, but with new
genotypes for the pm samples. If \code{N > 1}, a list of \code{dviData} objects.
}
\description{
Simulates genotypes for the references and missing persons in each AM family,
transfers to the PM singletons according to the indicated matching. Remaining
victims are simulated as unrelated.
}
\examples{

# Simulate refs and missing once and plot:
ex = dviSim(example2, N = 1, truth = c(V1 = "M1", V2 = "M2", V3 = "M3"))
plotDVI(ex, marker = 1)

# Two simulations and plot for the first
ex = dviSim(example2, N = 2, truth = c(V1 = "M1", V2 = "M2", V3 = "M3"),
            seed = 1729)
plotDVI(ex[[1]], marker = 1)


}
\seealso{
\code{\link[forrel:profileSim]{forrel::profileSim()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsetDVI.R
\name{subsetDVI}
\alias{subsetDVI}
\title{Extract a subset of a DVI dataset}
\usage{
subsetDVI(
  dvi,
  pm = NULL,
  am = NULL,
  missing = NULL,
  removeUnpairedPM = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{dvi}{A \code{\link[=dviData]{dviData()}} object}

\item{pm}{A vector with names or indices of victim samples. By default, all
are included.}

\item{am}{A vector with names or indices of AM components. By default,
components without remaining missing individuals are dropped.}

\item{missing}{A vector with names or indices of missing persons. By default,
all missing persons in the remaining AM families are included.}

\item{removeUnpairedPM}{A logical, by default TRUE, removing PM samples with
no remaining pairings.}

\item{verbose}{A logical.}
}
\value{
A \code{dviData} object.
}
\description{
Extract a subset of a DVI dataset
}
\examples{

subsetDVI(example2, pm = 1:2) |> plotDVI()
subsetDVI(example2, pm = "V1", am = 1) |> plotDVI()
subsetDVI(example2, missing = "M3") |> plotDVI()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviData.R
\name{getSimpleFams}
\alias{getSimpleFams}
\title{Find the simple families of a DVI dataset}
\usage{
getSimpleFams(dvi)
}
\arguments{
\item{dvi}{A \code{dviData} object.}
}
\value{
A character (if \code{dvi$am} has names) or integer vector.
}
\description{
Extract the names (if present) or indices of the \emph{simple} reference families,
i.e., the families containing exactly 1 missing person.
}
\examples{
# No simple families
simple1 = getSimpleFams(example1)
stopifnot(length(simple1) == 0)

# Second family is simple
simple2 = getSimpleFams(example2)
stopifnot(simple2 == "F2")

}
\seealso{
\code{\link[=getFamily]{getFamily()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addVictims.R
\name{addVictims}
\alias{addVictims}
\title{Add victim samples to a DVI dataset}
\usage{
addVictims(dvi, newvics)
}
\arguments{
\item{dvi}{A \code{\link[=dviData]{dviData()}} object.}

\item{newvics}{A \code{singleton} object or a list of such.}
}
\value{
A new \code{\link[=dviData]{dviData()}} object with the new victims added to the \code{pm}
data.
}
\description{
This is a convenience function for adding new victim samples to an existing
\code{\link[=dviData]{dviData()}} object. Users are recommended to use this instead of manipulating
the \code{pm} data directly, as it ensures consistency of the output and also
recomputes the internal \code{pairings} object.
}
\examples{
addVictims(example1, singleton("V4"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{symmetricSibs}
\alias{symmetricSibs}
\title{Dataset: Symmetry examples}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 5 singletons (male victims).
\item \code{am}: A list of 3 pedigrees
\item \code{missing}: A character vector of length 5, naming the missing persons.
}
}
\usage{
symmetricSibs
}
\description{
A toy DVI dataset illustrating various forms of undecidability due to
symmetries in the solutions.
}
\examples{

symmetricSibs

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setPairing.R
\name{setPairing}
\alias{setPairing}
\title{Set identifications manually}
\usage{
setPairing(
  dvi,
  match = NULL,
  victim = NULL,
  missing = NULL,
  Conclusion = "Provided",
  Comment = "",
  verbose = TRUE
)
}
\arguments{
\item{dvi}{A DVI dataset.}

\item{match}{A named vector of the format c(vic1 = miss2, vic2 = miss2, ...).}

\item{victim}{A vector of victim sample names. If NULL, defaulting to
\code{names(match)}.}

\item{missing}{A vector of missing person names, of the same length as
\code{victim}. If NULL, defaulting to \code{as.character(match)}.}

\item{Conclusion}{A character passed on to the \code{Conclusion} column of the
output summary.}

\item{Comment}{A character passed on to the \code{Comment} column of the output
summary.}

\item{verbose}{A logical, by default TRUE.}
}
\value{
A list with the following entries:
\itemize{
\item \code{dviReduced}: The new \code{dviData} object, as described in Details
\item \code{summary}: A data frame summarising the identifications
}
}
\description{
Manually set one or several identifications in a DVI dataset. Typically,
these are obtained by external means, e.g., fingerprints, dental records etc.
}
\details{
The command \code{setPairing(dvi, c("V" = "M"))} does the following:
\itemize{
\item Transfer the data of victim "V" to the individual "M" in the appropriate
reference family
\item Remove "M" from the list of missing persons
\item Remove "V" from the list of victim samples
\item Update the list of pairings
}
}
\examples{
x = setPairing(example2, match = c("V3" = "M2"))
x$dviReduced
x$summary

# Alternative syntax, using `victim` and `missing`
y = setPairing(planecrash, victim = c("V4", "V5"), missing = c("M4", "M5"),
           Conclusion = "External evidence", Comment = "Dental")
y$dviReduced
y$summary

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviData.R
\name{getFamily}
\alias{getFamily}
\title{Get AM component of selected individuals}
\usage{
getFamily(dvi, ids)
}
\arguments{
\item{dvi}{A \code{\link[=dviData]{dviData()}} object.}

\item{ids}{A vector of ID labels of members of \code{dvi$am}.}
}
\value{
A vector of the same length as \code{ids}, containing the family names of
the \code{ids} individuals.
}
\description{
Get AM component of selected individuals
}
\examples{
getFamily(example2, ids = example2$missing)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{example1}
\alias{example1}
\title{DVI dataset: Generational trio}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 3 singletons (victims).
\item \code{am}: A pedigree with three missing persons and one typed reference
individual.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
example1
}
\description{
A proof-of-concept dataset involving three missing members (child, father,
grandfather) of a single family. With the given data, stepwise victim
identification fails to find the correct solution, while joint identification
succeeds.
}
\examples{
   
example1
plotDVI(example1, marker = 1)

jointDVI(example1)  

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{planecrash}
\alias{planecrash}
\title{DVI dataset: Simulated plane crash}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 8 female singletons (victims).
\item \code{am}: A list of 5 pedigrees, each with one missing member and one
genotyped member.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
planecrash
}
\description{
A simulated dataset based on Exercise 3.3 in Egeland et al. "Relationship
Inference with Familias and R" (2015).
}
\details{
The 15 markers are \code{CSF1PO}, \code{D13S317}, \code{D16S539}, \code{D18S51}, \code{D21S11},
\code{D3S1358}, \code{D5S818}, \code{D7S820}, \code{D8S1179}, \code{FGA}, \code{PENTA_D}, \code{PENTA_E},
\code{TH01}, \code{TPOX}, and \code{VWA}.

Source code for the simulation, and a file containing the allele frequencies,
can be found in the \code{data-raw} folder of the GitHub repository:
https://github.com/magnusdv/dvir.
}
\examples{
planecrash
plotDVI(planecrash)

# Markers and allele frequencies
db = getDatabase(planecrash)
db[1:3]

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{fire}
\alias{fire}
\title{DVI dataset: Family of fire victims}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 3 singletons (victims).
\item \code{am}: A pedigree with three missing persons and one typed reference
individual.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
fire
}
\description{
A family with three missing persons after a fire, and one reference
individual. This example is featured in the GLR paper (Egeland & Vigeland,
2024).
}
\examples{
fire
plotDVI(fire, marker = 1)
dviSolve(fire, verbose = FALSE)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{KETPchap484}
\alias{KETPchap484}
\title{Data used in the book Kling et al. (2021)}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 4 singletons (victims).
\item \code{am}: A list of 3 pedigrees.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
KETPchap484
}
\description{
Data used in last example of Chapter 4 in Kling et al. (2021) "Mass
Identifications: Statistical Methods in Forensic Genetics". There are 2
female victims, 2 male victims. There are four reference families with 2
missing females and 2 missing males. There are 21 markers. An `equal mutation
mode with rate 0.005 is specified.
}
\examples{
KETPchap484

plotDVI(KETPchap484, nrowPM = 4)


}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jointDVI.R
\name{jointDVI}
\alias{jointDVI}
\alias{compactJointRes}
\title{Joint DVI analysis}
\usage{
jointDVI(
  dvi,
  pairings = NULL,
  ignoreSex = FALSE,
  assignments = NULL,
  limit = 0,
  nkeep = NULL,
  undisputed = TRUE,
  markers = NULL,
  threshold = 10000,
  strict = FALSE,
  relax = !strict,
  disableMutations = NA,
  maxAssign = 1e+05,
  numCores = 1,
  check = TRUE,
  verbose = TRUE
)

compactJointRes(jointRes, LRthresh = NULL)
}
\arguments{
\item{dvi}{A \code{dviData} object, typically created with \code{\link[=dviData]{dviData()}}.}

\item{pairings}{A list of possible pairings for each victim. If NULL, all
sex-consistent pairings are used.}

\item{ignoreSex}{A logical.}

\item{assignments}{A data frame containing the assignments to be considered
in the joint analysis. By default, this is automatically generated by
taking all combinations from \code{pairings}.}

\item{limit}{A positive number, by default 0. Only pairwise LR values above
this are considered.}

\item{nkeep}{An integer, or NULL. If given, only the \code{nkeep} most likely
pairings are considered for each victim.}

\item{undisputed}{A logical, by default TRUE.}

\item{markers}{A vector indicating which markers should be included in the
analysis. By default all markers are included.}

\item{threshold}{A positive number, passed onto \code{\link[=findUndisputed]{findUndisputed()}}. Default:
1e4.}

\item{strict}{A logical, passed onto \code{\link[=findUndisputed]{findUndisputed()}}. Default: FALSE.}

\item{relax}{Deprecated.}

\item{disableMutations}{A logical, or NA (default). The default action is to
disable mutations in all reference families without Mendelian errors.}

\item{maxAssign}{A positive integer. If the number of assignments going into
the joint calculation exceeds this, the function will abort with an
informative error message. Default: 1e5.}

\item{numCores}{An integer; the number of cores used in parallelisation.
Default: 1.}

\item{check}{A logical, indicating if the input data should be checked for
consistency.}

\item{verbose}{A logical.}

\item{jointRes}{A data frame produced by \code{jointDVI()}.}

\item{LRthresh}{A positive number, used as upper limit for the LR comparing the
top result with all others.}
}
\value{
A data frame. Each row describes an assignment of victims to missing
persons, accompanied with its log likelihood, the LR compared to the null
(i.e., no identifications), and the posterior corresponding to a flat
prior.

The function \code{compactJointRes()} removes columns without assignments, and
solutions whose LR compared with the top result is below \code{1/LRthresh}.
}
\description{
Joint matching of victims and missing persons in a disaster victim identification (DVI)
case, as described by Vigeland & Egeland (2021).

NOTE: This is a legacy function. For new projects we recommend these instead:
\itemize{
\item \code{dviSolve()} (for a complete DVI pipeline)
\item \code{dviJoint()} (for fully joint analysis)
}

The \code{jointDVI()} function was the first implementation of a complete DVI analysis in
\code{dvir}. It both manages the analysis workflow (including exclusion of impossible
pairings and detection of undisputed matches) and performs the final joint analysis.
As the package evolved, it became beneficial to split these two roles, resulting in
the two functions above.
}
\examples{
jointDVI(example2)

}
\references{
Vigeland, MD., Egeland, T. Joint DNA-based disaster victim identification.
Sci Rep 11, 13661 (2021).
}
\seealso{
\code{\link[=pairwiseLR]{pairwiseLR()}}, \code{\link[=findUndisputed]{findUndisputed()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dviSolve.R
\name{dviSolve}
\alias{dviSolve}
\title{A complete workflow for solving a DVI case}
\usage{
dviSolve(
  dvi,
  threshold = 10000,
  threshold2 = max(1, threshold/10),
  maxIncomp = 2,
  ignoreSex = FALSE,
  limit = 0,
  detailedOutput = FALSE,
  verbose = TRUE,
  debug = FALSE
)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{threshold}{LR threshold for 'significant' match.}

\item{threshold2}{LR threshold for 'probable' match. By default set to
\code{threshold/10}.}

\item{maxIncomp}{An integer passed onto \code{\link[=findExcluded]{findExcluded()}}. A pairing is
excluded if the number of incompatible markers exceeds this.}

\item{ignoreSex}{A logical, by default FALSE.}

\item{limit}{A number passed onto \code{\link[=findUndisputed]{findUndisputed()}}; only pairwise LR values
above this are considered.}

\item{detailedOutput}{A logical, by default FALSE. See Details.}

\item{verbose, debug}{Logicals.}
}
\value{
A list of data frames \code{AM} and \code{PM}.

If \code{detailedOutput = TRUE}, the result also includes the exclusion matrix
from the first call to \code{findExcluded()} and the LR matrix from the first
call to \code{findUndisputed()}.
}
\description{
Wraps the main\code{dvir} workflow for DVI analysis and identification of likely AM--PM matches.
}
\details{
The function roughly implements the following workflow:
\itemize{
\item \strong{Preparation and QC}
\itemize{
\item Consolidate data and validate input: \code{\link[=checkDVI]{checkDVI()}}
\item Generate pairings: \code{\link[=generatePairings]{generatePairings()}}
}
\item \strong{Initial reduction}
\itemize{
\item Detect and remove nonidentifiable missing persons: \code{\link[=findNonidentifiable]{findNonidentifiable()}}
}
\item \strong{Iterative reduction} (repeat until stable)
\itemize{
\item Excluded individuals and pairings: \code{\link[=findExcluded]{findExcluded()}}
\item Undisputed matches from pairwise LRs: \code{\link[=findUndisputed]{findUndisputed()}}
}
\item \strong{AM-driven analysis}: \code{\link[=amDrivenDVI]{amDrivenDVI()}}
\itemize{
\item Simple families (1 missing): Summarise conflicting results
\item Complex families (>1 missing): Joint analysis + GLR
}
\item \strong{PM-driven analysis}
\itemize{
\item Summarise remaining victim samples.
}
\item \strong{Output}
\itemize{
\item Format AM and PM summaries: \code{\link[=formatSummary]{formatSummary()}}
}
}
}
\examples{
dviSolve(example2)
dviSolve(example2, threshold = 5, detailedOutput = TRUE, verbose = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/familias2dvir.R
\name{familias2dvir}
\alias{familias2dvir}
\title{Convert a Familias file to DVI data}
\usage{
familias2dvir(
  famfile,
  victimPrefix = NULL,
  familyPrefix = NULL,
  refPrefix = NULL,
  missingPrefix = NULL,
  missingFormat = NULL,
  othersPrefix = NULL,
  verbose = TRUE,
  missingIdentifier = "^Missing"
)
}
\arguments{
\item{famfile}{Path to Familias file.}

\item{victimPrefix}{Prefix used to label PM individuals.}

\item{familyPrefix}{Prefix used to label the AM families.}

\item{refPrefix}{Prefix used to label the reference individuals, i.e., the
typed members of the AM families.}

\item{missingPrefix}{Prefix used to label the missing persons. At most one of
\code{missingPrefix} and \code{missingFormat} can be given.}

\item{missingFormat}{A string indicating family-wise labelling of missing
persons, using \verb{[FAM]}, \verb{[IDX]}, \verb{[MIS]} as place holders with the
following meanings (see Examples):
\itemize{
\item \verb{[FAM]}: family index
\item \verb{[IDX]}: index of missing person within the family
\item \verb{[MIS]}: index within all missing persons
}}

\item{othersPrefix}{Prefix used to label other untyped individuals. Use ""
for numeric labels ( 1, 2, ...).}

\item{verbose}{A logical, passed on to \code{readFam()}.}

\item{missingIdentifier}{A character of length 1 used to identify missing
persons in the Familias file. The default chooses everyone whose label
begins with "Missing".}
}
\value{
A \code{dviData} object.
}
\description{
This is a wrapper for \code{\link[pedFamilias:readFam]{pedFamilias::readFam()}} that reads Familias files with
DVI information.
}
\details{
The sex of the missing persons need to be checked as this
information may not be correctly recorded in the fam file.
}
\examples{

# Family with three missing
file = system.file("extdata", "dvi-example.fam", package="dvir")

# Read file without relabelling
y = familias2dvir(file)
plotDVI(y)

# With relabelling
z = familias2dvir(file, missingFormat = "M[FAM]-[IDX]",
                   refPrefix = "ref", othersPrefix = "E")
plotDVI(z)

}
\seealso{
\code{\link[=dviData]{dviData()}}, \code{\link[=relabelDVI]{relabelDVI()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{KETPchap481}
\alias{KETPchap481}
\title{Data used in the book Kling et al. (2021)}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 2 singletons (victims).
\item \code{am}: A list of 1 pedigree.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
KETPchap481
}
\description{
Data used in the example in Chapter 4.8.1 in Kling et al. (2021) "Mass
Identifications: Statistical Methods in Forensic Genetics". The victims are
V1 and V2, both females. There is one reference family with 2 missing
persons, both females. There are 21 markers, no mutation model.
}
\examples{
plotDVI(KETPchap481, marker = 1)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotDVI.R
\name{plotSolution}
\alias{plotSolution}
\title{Plot DVI solution}
\usage{
plotSolution(dvi, assignment, k = 1, format = "[S]=[M]", ...)
}
\arguments{
\item{dvi}{A \code{dviData} object.}

\item{assignment}{A named character of the format \code{c(victim = missing, ...)},
or a data frame produced by \code{\link[=jointDVI]{jointDVI()}}.}

\item{k}{An integer; the row number when \code{assignment} is a data frame.}

\item{format}{A string indicating how identified individuals should be
labelled, using \verb{[M]} and \verb{[S]} as place holders for the missing person and
the matching sample, respectively. (See Examples.)}

\item{...}{Further arguments passed on to \code{\link[=plotDVI]{plotDVI()}}.}
}
\value{
NULL.
}
\description{
A version of \code{\link[=plotDVI]{plotDVI()}} tailor-made to visualise identified individuals, for
example as reported by \code{jointDVI()}.
}
\examples{

res = jointDVI(example2, verbose = FALSE)

plotSolution(example2, res)

# With line break in labels
plotSolution(example2, res, format = "[M]=\n[S]")

# With genotypes for marker 1
plotSolution(example2, res, marker = 1)

# Non-optimal solutions
plotSolution(example2, res, k = 2, pm = FALSE)
plotSolution(example2, res, k = 2, cex = 1.3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{icmp}
\alias{icmp}
\title{DVI dataset: A large reference pedigree}
\format{
A \code{dviData} object with the following content:
\itemize{
\item \code{pm}: A list of 5 singletons (victims).
\item \code{am}: A reference pedigree with 6 genotyped members and 12 missing
persons.
\item \code{missing}: A vector containing the names of the missing persons.
}
}
\usage{
icmp
}
\description{
DVI dataset based loosely on the ICMP 2017 workshop material
https://www.few.vu.nl/~ksn560/Block-III-PartI-KS-ISFG2017.pdf (page 18).
There are 3 female victims, 2 male victims and 6 missing persons of both
sexes. We have renamed the individuals and simulated data for 13 CODIS
markers (see Details).
}
\details{
The 13 markers are, in order: \code{CSF1PO}, \code{D3S1358}, \code{D5S818},\code{D7S820},
\code{D8S1179}, \code{D13S317}, \code{D16S539}, \code{D18S51}, \code{D21S11}, \code{FGA}, \code{TH01}, \code{TPOX},
and \code{vWA}.

Source code for the simulation, and a file containing the allele frequencies,
can be found in the \code{data-raw} folder of the GitHub repository:
https://github.com/magnusdv/dvir.
}
\examples{
icmp
plotDVI(icmp)

# Markers and allele frequencies
db = getDatabase(icmp)
db[1:3]

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatSummary.R
\name{formatSummary}
\alias{formatSummary}
\title{Format final summary table}
\usage{
formatSummary(dfs, orientation = c("AM", "PM"), columns = NULL, dvi = NULL)
}
\arguments{
\item{dfs}{A list of data frames.}

\item{orientation}{Either "AM" or "PM", controlling column order and sorting.}

\item{columns}{A (optional) character vector with column names in the wanted
order.}

\item{dvi}{A \code{dviData} object used for sorting. Note that if given, this must
contain all victims and families.}
}
\value{
A data frame.
}
\description{
Combines and harmonises summary tables from different DVI analyses
}
\details{
The default column order is controlled by \code{orientation}, which the following
effect:
\itemize{
\item \code{AM}:
\itemize{
\item Column order: \code{Family}, \code{Missing}, \code{Sample}, \code{LR}, \code{GLR}, \code{Conclusion}, \code{Comment}
\item Sort by: \code{Family} and \code{Missing}
}
\item \code{PM}:
\itemize{
\item Column order: \code{Sample}, \code{Missing}, \code{Family}, \code{LR}, \code{GLR}, \code{Conclusion}, \code{Comment}
\item Sort by: \code{Sample}
}
}

Columns (in any of the data frames) other than these are simply ignored.
}
\examples{
u = findUndisputed(planecrash)
a = amDrivenDVI(u$dviReduced, threshold2 = 500)

u$summary
a$summary

formatSummary(list(u$summary, a$summary$AM))
formatSummary(list(u$summary, a$summary$PM), orientation = "PM", dvi = planecrash)

}
